% -------------------------------------------------------------------------- %
% ArrayRef Library
% Copyright (C) 2008 Matt Giuca
% 17/10/2008
%
% Provides functionality similar to Mercury's 'array' module, but accesses
% arrays through the store, so they are aliasable and mutable.
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %

:- module arrayref.

:- interface.

:- import_module store.
:- import_module array.
:- import_module int.

% An arrayref(T, S) is a reference to a mutable array, with elements of type
% T, in store type S.
% This behaves much like a regular array, but all accesses must go through the
% store, and aliased mutation is allowed.
% The array's length is fixed (it can be resized through copying, but not
% mutation). See vectorref for arrays whose lengths are mutable.
:- type arrayref(T, S).

% arrayref.init(Size, Init, ArrayRef, !S).
% Creates a new array in a store S, and returns a reference to it.
:- pred arrayref.init(int::in, T::in, arrayref(T, S)::out, S::di, S::uo)
    is det <= store(S).

% arrayref.unsafe_init(Size, ArrayRef, !S).
% Creates a new array in a store S, and returns a reference to it.
% The elements of the array are undefined, and must be set (using array_set)
% before being read.
:- pred arrayref.unsafe_init(int::in, arrayref(T, S)::out, S::di, S::uo)
    is det <= store(S).

% arrayref.extract(ArrayRef, Array, !S).
% Given an array reference, extracts a copy of the array from the store.
% The resulting array is unique, and does not interfere with the store.
:- pred arrayref.extract(arrayref(T, S)::in, array(T)::array_uo, S::di, S::uo)
    is det <= store(S).

% See array.min.
:- pred arrayref.min(arrayref(_T, S)::in, int::out, S::di, S::uo)
    is det <= store(S).

% See array.max.
:- pred arrayref.max(arrayref(_T, S)::in, int::out, S::di, S::uo)
    is det <= store(S).

% See array.size.
:- pred arrayref.size(arrayref(_T, S)::in, int::out, S::di, S::uo)
    is det <= store(S).

% See array.bounds.
:- pred arrayref.bounds(arrayref(_T, S)::in, int::out, int::out,
    S::di, S::uo) is det <= store(S).

% arrayref.lookup returns the Nth element of an array.
% Throws an exception if the index is out of bounds.
:- pred arrayref.lookup(arrayref(T, S)::in, int::in, T::out, S::di, S::uo)
    is det <= store(S).

% arrayref.set sets the Nth element of an array, destructively modifying the
% array through the store.
% Throws an exception if the index is out of bounds.
:- pred arrayref.set(arrayref(T, S)::in, int::in, T::in, S::di, S::uo)
    is det <= store(S).

% -------------------------------------------------------------------------- %

% arrayref.copy(ArrayRef0, ArrayRef, !Store).
% Makes a new unique copy of an array, and returns a reference to it.
% This makes a shallow copy.
:- pred arrayref.copy(arrayref(T, S)::in, arrayref(T, S)::out, S::di, S::uo)
    is det <= store(S).

% arrayref.resize(ArrayRef0, Size, Init, ArrayRef, !Store).
% Makes a new unique copy of an array, but with its length changed to `Size'.
% This makes a shallow copy.
% Returns a reference to the new array. Note that this does not mutate the
% existing array.
:- pred arrayref.resize(arrayref(T, S)::in, int::in, T::in,
    arrayref(T, S)::out, S::di, S::uo) is det <= store(S).

% arrayref.shrink(ArrayRef0, Size, ArrayRef, !Store).
% Makes a new unique copy of an array, but with its length reduced to `Size'.
% This makes a shallow copy. Throws an exception if `Size' is larger than the
% size of `ArrayRef0'.
:- pred arrayref.shrink(arrayref(T, S)::in, int::in,
    arrayref(T, S)::out, S::di, S::uo) is det <= store(S).

% arrayref.unsafe_resize(ArrayRef0, Size, ArrayRef, !Store).
% Makes a new unique copy of an array, but with its length changed to `Size'.
% This makes a shallow copy.
% Returns a reference to the new array. Note that this does not mutate the
% existing array.
% The new elements of the array are undefined, and must be set (using
% array_set) before being read.
:- pred arrayref.unsafe_resize(arrayref(T, S)::in, int::in,
    arrayref(T, S)::out, S::di, S::uo) is det <= store(S).

% arrayref.shrink_array(ArrayRef0, Size, Init, Array, !Store).
% Like arrayref.shrink, but returns it as a regular array, not an arrayref.
:- pred arrayref.shrink_array(arrayref(T, S)::in, int::in,
    array(T)::out, S::di, S::uo) is det <= store(S).

% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- implementation.

:- import_module string.
:- import_module exception.
:- import_module require.
:- import_module list.

:- type arrayref(T, S) ---> ar(array.array(T)).

arrayref.init(Size, Init, ar(Array), !S) :-
    array.init(Size, Init, Array).

arrayref.unsafe_init(Size, ar(Array), !S) :-
    ( Size < 0 ->
        error("arrayref.unsafe_init: negative size")
    ;
        unsafe_init_2(Size, Array)
    ).

arrayref.extract(ar(Array0), Array, !S) :-
    array.copy(Array0, Array).

arrayref.min(ar(Array), X, !S) :-
    array.min(Array, X).

arrayref.max(ar(Array), X, !S) :-
    array.max(Array, X).

arrayref.size(ar(Array), X, !S) :-
    array.size(Array, X).

arrayref.bounds(ar(Array), X, Y, !S) :-
    array.bounds(Array, X, Y).

arrayref.lookup(ar(Array), N, V, !S) :-
    array.lookup(Array, N, V).

:- pragma promise_pure(arrayref.set/5).
arrayref.set(ar(Array), N, V, !S) :-
    % Call impure code, but with a dummy store to make it semantically pure
    impure array_set(Array, N, V).

arrayref.copy(ar(Array0), ar(Array), !S) :-
    array.copy(Array0, Array).

arrayref.resize(ar(Array0), Size, Init, ar(Array), !S) :-
    array_pure_resize(Array0, Size, Init, Array).

arrayref.shrink(ar(Array0), Size, ar(Array), !S) :-
    OldSize = array.size(Array0),
    ( Size > OldSize ->
        error("arrayref.shrink: can't shrink to a larger size")
    ;
        % unsafe_resize is safe if shrinking
        array_pure_unsafe_resize(Array0, Size, Array)
    ).

arrayref.unsafe_resize(ar(Array0), Size, ar(Array), !S) :-
    array_pure_unsafe_resize(Array0, Size, Array).

arrayref.shrink_array(AR0, Size, Array, !S) :-
    arrayref.shrink(AR0, Size, ar(Array), !S).

%-----------------------------------------------------------------------------%
% The remaining code is copied, and slightly modified, from Mercury's 'array'
% module (we needed different modes to allow ourselves to call this impurely).
%-----------------------------------------------------------------------------%

:- pred unsafe_init_2(int::in, array(T)::array_uo) is det.

:- pragma foreign_proc("C",
    unsafe_init_2(Size::in, Array::array_uo),
    [will_not_call_mercury, promise_pure, thread_safe, will_not_modify_trail,
        does_not_affect_liveness
    ],
"
    ML_alloc_array(Array, Size + 1, MR_PROC_LABEL);
    Array->size = Size;
    /* Leave values in array->elements undefined. */
").

%-----------------------------------------------------------------------------%

% Impurely mutates an array object. (But does bounds checking).
% This is private. The public preds which call this function expose a pure
% interface.
:- impure pred array_set(array(T)::in, int::in, T::in) is det.
array_set(Array, Index, Item) :-
    ( bounds_checks, \+ array.in_bounds(Array, Index) ->
        out_of_bounds_error(Array, Index, "arrayref.set")
    ;
        impure array_unsafe_set(Array, Index, Item)
    ).

%-----------------------------------------------------------------------------%

% Unsafely / impurely mutates an array object.
% This is private. The public preds which call this function expose a
% safe / pure interface.
:- impure pred array_unsafe_set(array(T)::in, int::in, T::in) is det.

% Foreign language code based on Mercury's array code. Need foreign language
% code to guarantee destructive update and let us specify the "in" mode
% (rather than di).
:- pragma foreign_proc("C",
    array_unsafe_set(Array::in, Index::in, Item::in),
    [will_not_call_mercury, thread_safe, will_not_modify_trail],
"
    /* TEMP */
    Array->elements[Index] = Item; /* destructive update! */
").

:- pragma foreign_proc("C#",
    array_unsafe_set(Array::in, Index::in, Item::in),
    [will_not_call_mercury, thread_safe],
"{
    Array.SetValue(Item, Index);   /* destructive update! */
}").

:- pragma foreign_proc("Java",
    array_unsafe_set(Array::array_di, Index::in, Item::in),
    [will_not_call_mercury, thread_safe],
"
    /* destructive update! */
    java.lang.reflect.Array.set(Array, Index, Item);
").

%-----------------------------------------------------------------------------%

:- pred bounds_checks is semidet.
:- pragma inline(bounds_checks/0).

:- pragma foreign_proc("C",
    bounds_checks,
    [will_not_call_mercury, promise_pure, thread_safe, will_not_modify_trail],
"
#ifdef ML_OMIT_ARRAY_BOUNDS_CHECKS
    SUCCESS_INDICATOR = MR_FALSE;
#else
    SUCCESS_INDICATOR = MR_TRUE;
#endif
").

:- pragma foreign_proc("C#",
    bounds_checks,
    [will_not_call_mercury, promise_pure, thread_safe],
"
#if ML_OMIT_ARRAY_BOUNDS_CHECKS
    SUCCESS_INDICATOR = false;
#else
    SUCCESS_INDICATOR = true;
#endif
").

:- pragma foreign_proc("Java",
    bounds_checks,
    [will_not_call_mercury, promise_pure, thread_safe],
"
    // never do bounds checking for Java (throw exceptions instead)
    succeeded = false;
").

%-----------------------------------------------------------------------------%

    % throw an exception indicating an array bounds error
:- pred out_of_bounds_error(array(T), int, string).
:- mode out_of_bounds_error(array_ui, in, in) is erroneous.
%:- mode out_of_bounds_error(in, in, in) is erroneous.

out_of_bounds_error(Array, Index, PredName) :-
    % Note: we deliberately do not include the array element type name in the
    % error message here, for performance reasons: using the type name could
    % prevent the compiler from optimizing away the construction of the
    % type_info in the caller, because it would prevent unused argument
    % elimination. Performance is important here, because array.set and
    % array.lookup are likely to be used in the inner loops of
    % performance-critical applications.
    array.bounds(Array, Min, Max),
    throw(array.index_out_of_bounds(
        string.format("%s: index %d not in range [%d, %d]",
            [s(PredName), i(Index), i(Min), i(Max)]))).

%-----------------------------------------------------------------------------%

% Note: array.resize has been copied and modified with the following
% differences from the library version:
% * Not destructive on the old array (does not free it).
% * Always copies, even if size is the same.
% This function is pure. It does not modify the input array in any way.
:- pred array_pure_resize(array(T)::in, int::in, T::in, array(T)::array_uo)
    is det.

:- pragma foreign_decl("C", "
void MARS_pure_resize_array(MR_ArrayPtr new_array, MR_ArrayPtr old_array,
    MR_Integer array_size, MR_Word item);
").

:- pragma foreign_code("C", "
/*
** The caller is responsible for allocating the storage for the new array.
** This routine does the job of copying the old array elements to the
** new array, and initializing any additional elements in the new array.
*/
void
MARS_pure_resize_array(MR_ArrayPtr array, MR_ArrayPtr old_array,
    MR_Integer array_size, MR_Word item)
{
    MR_Integer i;
    MR_Integer elements_to_copy;

    elements_to_copy = old_array->size;
    if (elements_to_copy > array_size) {
        elements_to_copy = array_size;
    }

    array->size = array_size;
    for (i = 0; i < elements_to_copy; i++) {
        array->elements[i] = old_array->elements[i];
    }
    for (; i < array_size; i++) {
        array->elements[i] = item;
    }
}
").

:- pragma foreign_proc("C",
    array_pure_resize(Array0::in, Size::in, Item::in, Array::array_uo),
    [will_not_call_mercury, promise_pure, thread_safe, will_not_modify_trail,
        does_not_affect_liveness,
        sharing(yes(array(T), int, T, array(T)), [
            cel(Array0, []) - cel(Array, []),
            cel(Item, [])   - cel(Array, [T])
        ])
    ],
"
    ML_alloc_array(Array, Size + 1, MR_PROC_LABEL);
    MARS_pure_resize_array(Array, Array0, Size, Item);
").

% array_pure_unsafe_resize - same as pure_resize, but leaves elements of the
% array undefined.
:- pred array_pure_unsafe_resize(array(T)::in, int::in, array(T)::array_uo)
    is det.

:- pragma foreign_decl("C", "
void MARS_pure_unsafe_resize_array(MR_ArrayPtr new_array,
    MR_ArrayPtr old_array, MR_Integer array_size);
").

:- pragma foreign_code("C", "
/*
** The caller is responsible for allocating the storage for the new array.
** This routine does the job of copying the old array elements to the
** new array.
** It leaves the new elements of the array undefined.
*/
void
MARS_pure_unsafe_resize_array(MR_ArrayPtr array, MR_ArrayPtr old_array,
    MR_Integer array_size)
{
    MR_Integer i;
    MR_Integer elements_to_copy;

    elements_to_copy = old_array->size;
    if (elements_to_copy > array_size) {
        elements_to_copy = array_size;
    }

    array->size = array_size;
    for (i = 0; i < elements_to_copy; i++) {
        array->elements[i] = old_array->elements[i];
    }
}
").

:- pragma foreign_proc("C",
    array_pure_unsafe_resize(Array0::in, Size::in, Array::array_uo),
    [will_not_call_mercury, promise_pure, thread_safe, will_not_modify_trail,
        does_not_affect_liveness,
        sharing(yes(array(T), int, array(T)), [
            cel(Array0, []) - cel(Array, [])
        ])
    ],
"
    ML_alloc_array(Array, Size + 1, MR_PROC_LABEL);
    MARS_pure_unsafe_resize_array(Array, Array0, Size);
").

% XXX: Other backends not supported for arrayref.resize or unsafe_resize
% (due to effort involved in maintaining these huge functions).
