% -------------------------------------------------------------------------- %
% Mars/Venus Implementation
% Copyright (C) 2008 Matt Giuca
% 24/9/2008
%
% Conversion between AST and CFG code representation
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %
% This module converts code from AST (abstract syntax tree) representation
% (defined in ir.m) and CFG (control flow graph) representation (defined in
% cfg.m).
% -------------------------------------------------------------------------- %

:- module ast_cfg.

:- interface.

:- import_module ir.

:- import_module list.

% prog_to_cfg(!Program) converts all functions in a program to CFG form,
% calling func_to_cfg on each of them.
:- pred prog_to_cfg(program::in, program::out) is det.

% func_to_cfg(!Function) converts a function to CFG form if possible.
% Built-in functions and CFG functions remain the same.
% AST functions are converted to CFG form.
% This also performs error checking.
% Throws a context error if:
% - There exists a viable code path which does not return a value.
% - There is an attempt to read from a variable which may not have been
%   assigned on some or all code paths.
:- pred func_to_cfg(function::in, function::out) is det.

% expr_to_instrs(+Expr, -VarName, -Instrs).
% Converts a single expression into a sequence of low-level instructions
% which compute the expression and assign it to a variable (possibly a
% new temporary variable name). Returns the assigned variable name in VarName.
% (May clobber the temporary variable namespace, as there is no block or
% subscript information given). Useful for stand-alone statements.
:- pred expr_to_instrs(expr::in, varname::out, list(instr)::out) is det.

% basic_stmt_to_instrs(+BasicStmt, -Instrs).
% Converts a single basic statement into a sequence of low-level instructions
% which do the same thing. This will potentially assign many temporary
% variables.
% (May clobber the temporary variable namespace, as there is no block or
% subscript information given). Useful for stand-alone statements.
:- pred basic_stmt_to_instrs(basic_stmt::in, list(instr)::out) is det.

% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- implementation.

:- import_module pair.
:- import_module map.
:- import_module maybe.
:- import_module string.
:- import_module int.
:- import_module exception.

:- import_module cfg.
:- import_module context.

prog_to_cfg(program(Nodes0), program(Nodes)) :-
    prog_nodes_to_cfg(Nodes0, Nodes).

:- pred prog_nodes_to_cfg(list(program_node)::in, list(program_node)::out).
prog_nodes_to_cfg([], []).
prog_nodes_to_cfg([Node0|Nodes0], [Node|Nodes]) :-
    ( Node0 = pfunction(Func0) ->
        func_to_cfg(Func0, Func),
        Node = pfunction(Func)
    ;
        Node = Node0
    ),
    prog_nodes_to_cfg(Nodes0, Nodes).

:- pred new_def_map(def_map::out) is det.
new_def_map(Map) :-
    map.init(Map0),
    % Insert the special $RET variable, unbound.
    map.det_insert(Map0, retvname, unbound, Map).

% Inserts a parameter into a def_map. The name is bound to itself (since it is
% a parameter, it is assumed to already be bound).
:- pred def_map_insert_param({string, T, argmode}::in,
    def_map::in, def_map::out) is det.
def_map_insert_param({Name,_,_ArgMode}, !DefMap) :-
    map.det_insert(!.DefMap, svname(Name), bound(svname(Name)), !:DefMap).

% Inserts a local variable into a def_map. The name is unbound (since it is
% a parameter, it is assumed to be uninitialised).
:- pred def_map_insert_local(pair(varname, T)::in, def_map::in, def_map::out)
    is det.
def_map_insert_local(Name-_, !DefMap) :-
    map.det_insert(!.DefMap, Name, unbound, !:DefMap).

func_to_cfg(!Func) :-
    % Create a new defmap with all of the params.
    new_def_map(DefMap0),
    Params = !.Func ^ func_params,
    (
        % No parameters
        Params = no,
        DefMap = DefMap0
    ;
        Params = yes(YesParams),
        foldl(def_map_insert_param, YesParams, DefMap0, DefMap)
    ),
    func_body_to_cfg(DefMap, !.Func ^ func_context, !.Func ^ func_body,
        CFGBody),
    !:Func = !.Func ^ func_body := CFGBody.

% Note: def_map argument should have all of the function's parameters already
% bound in the map.
:- pred func_body_to_cfg(def_map::in, context::in,
    func_body::in, func_body::out) is det.
func_body_to_cfg(_DefMap, _Ctx, B@func_builtin, B).
func_body_to_cfg(_DefMap, _Ctx, B@func_body_cfg(_,_), B).
func_body_to_cfg(DefMap0, Ctx, func_body_ast(Decls,Stmts), NewFuncBody) :-
    NewFuncBody = 'new func_body_cfg'(Decls, CFG),
    % Insert all of the function's locals into the defmap, unbound
    foldl(def_map_insert_local, Decls, DefMap0, DefMap),
    ast_to_cfg(DefMap, Ctx, Stmts, CFG).

% ast_to_cfg(DefMap, Stmts, CFG) converts a statement block (AST form) to a
% control flow graph, translating the code.
% This also performs error checking (see func_to_cfg).
:- some [S] pred ast_to_cfg(def_map::in, context::in, stmt_block::in,
    cfg(S)::out) is det.
ast_to_cfg(DefMap, Ctx, Stmts, CFG) :-
    some [!CFG]
    (
        % Need to create the entry block and exit block so that
        % stmt_block_to_cfg has something to insert statements into, and to
        % branch to when it's done.
        !:CFG = cfg.new_cfg,
        Entry = cfg.get_entry(!.CFG),
        Exit = cfg.get_exit(!.CFG),
        stmt_block_to_cfg(Stmts, Entry, Exit, DefMap, map.init, AfterPredMap,
            map.init, ExitPredMap, !CFG),
        % The union of AfterPredMap and ExitPredMap contains information about
        % all predecessors to Exit. Use this to compute phi instructions for
        % Exit.
        map.overlay(ExitPredMap, AfterPredMap, UnionPredMap),
        ExitID = ref_id(Exit, !.CFG),
        reconcile_def_maps(ExitID, UnionPredMap, Phis, _, FinalDefMap),
        cfg.set_phis(Exit, Phis, !CFG),
        % Check that all code paths return a value
        retval_check(FinalDefMap, Ctx, FinalRetV),
        % Store the "final return value" name in the CFG so interpreter knows
        % where to look
        cfg.set_retv(FinalRetV, !CFG),
        CFG = !.CFG
    ).

% varname_binding
% Represents the binding state of an imperative variable at a particular point
% in the code (statically). There are three possibilities:
% - The variable may be bound to an SSA varname (which should be the qualified
%   name of the variable). This effectively says "at this point in the code,
%   the imperative variable is known as <given varname>".
% - The variable may be unbound. This means it has not yet been initialised,
%   and it would be an error to read from the variable at this point.
% - The variable may be "maybe_bound". This means that there are one or more
%   code paths reaching this point which result in the variable being bound,
%   but there are also one or more code paths resulting in the variable being
%   unbound. Because of the possibility of being unbound, it would also be an
%   error to read from the variable at this point, so this is effectively the
%   same as being unbound (and we don't even bother to keep track of the bound
%   name). We only distinguish the two cases for the sake of better error
%   messages.
% varname_binding is used as the values of def_maps.
:- type varname_binding
    --->    bound(varname)
    ;       maybe_bound
    ;       unbound.

% def_map
% Represents the binding state of all local variables in a function at a
% particular point in the code (statically).
% Each imperative variable (unqualified varname) is mapped onto a
% varname_binding. This means it is either mapped onto a qualified varname
% (representing the name of the SSA variable for this imperative variable at
% this point in the code), or flagged as being uninitialised at this point.
% All local variables must exist in the map (even if merely flagged as
% uninitialised). If a variable is referenced which is not in the map, it will
% be considered to be a global variable (and therefore be considered
% initialised).
:- type def_map == map(varname, varname_binding).

% pred_map(S)
% A binding state for each basic block. Allows a block to track the binding
% state for all of its predecessors (which may be folded into minimal phi
% instructions using reconcile_def_maps).
:- type pred_map(S) == map(bbref(S), def_map).

% subscript_map
% Specific to a single basic block. Represents the number of assignments which
% have been made to each imperative variable in the current block. This is
% used to generate unique subscripts per block.
% Note that any variables which have not been defined in this block will not
% have a key in the map.
:- type subscript_map == map(varname, int).

% new_ssa_variable(BlockID, VarName, QualVarName, !SubscriptMap, !DefMap).
% Given an unqualified variable name, produces a new qualified variable name
% which is unique, and updates the subscript map accordingly.
% Also updates the definition map, so that the QualVarName is now the
% "current" SSA name / qualified name for variable VarName.
% Note: Requires the block ID (integer), which will be used as part of the
% qualifier.
:- pred new_ssa_variable(int::in, varname::in, varname::out,
    subscript_map::in, subscript_map::out, def_map::in, def_map::out) is det.
new_ssa_variable(BlockID, VarName, QualVarName, !SubscriptMap, !DefMap) :-
    ( map.search(!.SubscriptMap, VarName, NewID) ->
        % Has been assigned already this block. Add a second qualification
        % level to disambiguate subsequent assignments per block.
        map.det_update(!.SubscriptMap, VarName, NewID+1, !:SubscriptMap),
        QualVarName = qvname(qvname(VarName,BlockID),NewID)
    ;
        % First assignment this block. Just qualify with this block's ID.
        map.det_insert(!.SubscriptMap, VarName, 1, !:SubscriptMap),
        QualVarName = qvname(VarName,BlockID)
    ),
    map.set(!.DefMap, VarName, bound(QualVarName), !:DefMap).

% new_temp_variable(BlockID, VarName, !SubscriptMap, !DefMap).
% Produces a new qualified variable name for a temporary variable, which is
% unique, and updates the subscript map accordingly.
% Note: Requires the block ID (integer), which will be used as part of the
% qualifier.
:- pred new_temp_variable(int::in, varname::out,
    subscript_map::in, subscript_map::out) is det.
new_temp_variable(BlockID, VarName, !SubscriptMap) :-
    % Just create a new SSA variable which qualifies the special "temp" name.
    % This guarantees a unique variable name within the procedure.
    % Do not update the DefMap (since we never need to know what the
    % currently-qualified name is for the "temp" variable name).
    new_ssa_variable(BlockID, tempvname, VarName, !SubscriptMap, map.init, _).

% apply_def_map_to_expr(+DefMap, !Expr).
% For each variable referenced in Expr, replaces it with its qualified name,
% as found in DefMap.
% Throws a context.error if the expression references a local variable which
% has not (or may not have) been initialised (but references to global
% variables are fine).
:- pred apply_def_map_to_expr(def_map::in, context::in, expr::in, expr::out)
    is det.
apply_def_map_to_expr(DefMap, Ctx, !Expr) :-
    E0 = !.Expr ^ expr_expr,
    apply_def_map_to_expr_(DefMap, Ctx, E0, E),
    !:Expr = !.Expr ^ expr_expr := E.

:- pred apply_def_map_to_expr_(def_map::in, context::in,
    expr_::in, expr_::out) is det.
apply_def_map_to_expr_(_DefMap, _Ctx, E@intlit(_), E).
apply_def_map_to_expr_(DefMap, Ctx, arraylit(Exprs0), arraylit(Exprs)) :-
    list.map(apply_def_map_to_expr(DefMap, Ctx), Exprs0, Exprs).
apply_def_map_to_expr_(DefMap, Ctx, E0@varref(VarName), E) :-
    ( map.search(DefMap, VarName, MaybeBinding) ->
        % Found in the map (so it is a local variable). But is it bound?
        (
            % Bound - rename the variable
            MaybeBinding = bound(QualVarName),
            E = varref(QualVarName)
        ;
            % MAY not yet be defined - error
            MaybeBinding = maybe_bound,
            context.throw_error(
                "Use of possibly-uninitialised variable: "
                ++ ir.varname_to_string_noescape(VarName), Ctx)
        ;
            % Not yet defined - error
            MaybeBinding = unbound,
            context.throw_error("Use of uninitialised variable: "
                ++ ir.varname_to_string_noescape(VarName), Ctx)
        )
    ; VarName = svname(Name) ->
        % Not found in the map; and simple variable name.
        % Assume this is a global variable, and therefore does not need to be
        % qualified. However, replace svname with gvname so we know it's
        % global for later.
        E = varref(gvname(Name))
    ; VarName = gvname(_Name) ->
        % Already global (for some reason).
        E = E0
    ;
        % Qualified, and not in the map - error
        throw("ast_cfg: Qualified variable not found in defmap: "
            ++ ir.varname_to_string_noescape(VarName))
    ).
apply_def_map_to_expr_(_DefMap, _Ctx, E@ctorref(_), E).
apply_def_map_to_expr_(DefMap, Ctx, fieldref(SE0,F), fieldref(SE,F)) :-
    apply_def_map_to_expr(DefMap, Ctx, SE0, SE).
apply_def_map_to_expr_(DefMap, Ctx, app(Func0,Args0), app(Func,Args)) :-
    apply_def_map_to_expr(DefMap, Ctx, Func0, Func),
    map(apply_def_map_to_expr(DefMap, Ctx), Args0, Args).
apply_def_map_to_expr_(DefMap, Ctx, parapp(Func0,Args0), parapp(Func,Args)) :-
    apply_def_map_to_expr(DefMap, Ctx, Func0, Func),
    map(apply_def_map_to_expr(DefMap, Ctx), Args0, Args).

% Check that $RET is bound (assigned on all code paths) in def map, which
% should be performed on the final def map of a function.
% Returns the varname of the final return value.
% Raises context errors if not assigned on all code paths.
:- pred retval_check(def_map::in, context::in, varname::out) is det.
retval_check(DefMap, Ctx, RetVarName) :-
    map.lookup(DefMap, retvname, MaybeBinding), % Must be in map (or error)
    (
        % Bound - okay
        MaybeBinding = bound(RetVarName)
    ;
        % MAY not yet be defined - error
        MaybeBinding = maybe_bound,
        context.throw_error(
            "Function may not return a value", Ctx)
    ;
        % Not yet defined - error
        MaybeBinding = unbound,
        context.throw_error("Function does not return a value", Ctx)
    ).

% add_to_pred_map(+Predecessor, +PredDefMap, +Target, !TargetPredMap, !CFG).
% Adds a predecessor / def-map pair to a pred-map. This should be done
% whenever any kind of branch instruction is added.
% This automatically adds BBRef as a predecessor, and the def-map for
% Predecessor is appended to TargetPredMap, where it is used
% (along with all other predecessors) to generate the phis and new defmap.
:- pred add_to_pred_map(bbref(S)::in, def_map::in, bbref(S)::in,
    pred_map(S)::in, pred_map(S)::out, cfg(S)::in, cfg(S)::out) is det.
add_to_pred_map(Pred, PredDefMap, Target, !TargetPredMap, !CFG) :-
    cfg.append_predecessor(Target, Pred, !CFG),
    map.set(!.TargetPredMap, Pred, PredDefMap, !:TargetPredMap).

% reconcile_def_maps(+BlockID, +PredMaps, -Phi, -SubscriptMap, -ResultDefMap).
% Detects any disagreements between members a set of variable mappings.
% Input is a mapping from basic blocks to def_maps. For each variable:
% - If all maps agree on the binding for the variable, nothing is done.
% - If there is any disagreement, a phi is created for the variable,
%   incorporating all blocks in the map as predecessors.
%   (Note that if a variable is unbound in any predecessor, it will not get a
%   phi block, and will be set to maybe_bound in the resulting def_map).
% The result is a list of the above phis, plus a subscript_map, with bindings
% for all the new phi'd variables, plus a def_map with the union of all the
% agreed-upon variables, and the phi'd variables (which is therefore all
% variables).
:- pred reconcile_def_maps(int::in, pred_map(S)::in,
    list(phi(S))::out, subscript_map::out, def_map::out) is det.
reconcile_def_maps(BlockID, PredMaps, Phis, SubscriptMap, DefMap) :-
    % Invert PredMaps, to produce a mapping from each variable to its phi.
    % (Though this includes redundant and illegal phis).
    invert_map_of_maps(PredMaps, VarMaps),
    map.foldl3(reconcile_def_map_var(BlockID), VarMaps, [], Phis,
        map.init, SubscriptMap, map.init, DefMap).

% Given an unqualified variable name, and a mapping which describes, for each
% predecessor block, which qualified variable name was used for that
% variable, determines what should be done (as per reconcile_def_maps), and
% updates subscript_map and def_map, and also potentially adds a phi to the
% list of phis.
:- pred reconcile_def_map_var(int::in, varname::in,
    map(bbref(S), varname_binding)::in, list(phi(S))::in, list(phi(S))::out,
    subscript_map::in, subscript_map::out, def_map::in, def_map::out) is det.
reconcile_def_map_var(BlockID, VarName, PredMap, !Phis,
    !SubscriptMap, !DefMap) :-
    VarBinds = map.values(PredMap),
    PredAssoc = map.to_assoc_list(PredMap),
    ( all_equal(VarBinds, MaybeVarBind) ->
        % All predecessors agree on the SSA name for VarName. Just store the
        % binding in DefMap and we're done (no need to phi).
        (
            % PredMap was empty (shouldn't happen, but we can handle it)
            MaybeVarBind = no,
            map.det_insert(!.DefMap, VarName, unbound, !:DefMap)
        ;
            MaybeVarBind = yes(VarBind),
            map.det_insert(!.DefMap, VarName, VarBind, !:DefMap)
        )
    ;
        % Disagreement. See if the variable is bound for all predecessors.
        ( foldl((pred(X-bound(V)::in, Map0::in, Map::out) is semidet :-
            map.set(Map0, X, V, Map)),
            PredAssoc, map.init, BoundAssoc) ->
            % All predecessors bind this variable, so we can make a phi
            % We need a new variable name
            new_ssa_variable(BlockID, VarName, Var, !SubscriptMap, !DefMap),
            Phi = phi(Var, BoundAssoc),
            !:Phis = [Phi|!.Phis]
        ; ifall(pred(_-unbound::in) is semidet, PredAssoc) ->
            % All predecessors are unbound, so the combined state is unbound
            map.det_insert(!.DefMap, VarName, unbound, !:DefMap)
        ;
            % Some are bound/maybe_bound, so the combined state is maybe_bound
            map.det_insert(!.DefMap, VarName, maybe_bound, !:DefMap)
        )
    ).

% Determines whether all elements of a list are equal.
% As a bonus, if they are, outputs the value found in the list (or 'no' if the
% list was empty).
:- pred all_equal(list(T)::in, maybe(T)::out) is semidet.
all_equal([], no).
all_equal([X], yes(X)).
all_equal([X|Ys@[Y|_]], Val) :- X = Y, all_equal(Ys, Val).

% Succeeds if predicate succeeds on any element of the list.
:- pred ifall(pred(T), list(T)).
:- mode ifall(pred(in) is semidet, in) is semidet.
ifall(_Pred, []).
ifall(Pred, [X|Xs]) :- Pred(X), ifall(Pred, Xs).

% invert_map_of_maps(Map0, Map).
% Take a map of maps, and invert it. If a map-of-maps is seen as a 2D matrix,
% this swaps the columns and rows.
:- pred invert_map_of_maps(map(T, map(U, V))::in, map(U, map(T, V))::out)
    is det.
invert_map_of_maps(Map0, Map) :-
    map.foldl(invert_map_of_maps2, Map0, map.init, Map).

:- pred invert_map_of_maps2(T::in, map(U, V)::in,
    map(U, map(T, V))::in, map(U, map(T, V))::out) is det.
invert_map_of_maps2(T, UVMap, !Map) :-
    map.foldl(invert_map_of_maps3(T), UVMap, !Map).

:- pred invert_map_of_maps3(T::in, U::in, V::in,
    map(U, map(T, V))::in, map(U, map(T, V))::out) is det.
invert_map_of_maps3(T, U, V, !Map) :-
    ( map.search(!.Map, U, TVMap0) ->
        map.set(TVMap0, T, V, TVMap),
        map.det_update(!.Map, U, TVMap, !:Map)
    ;
        map.det_insert(map.init, T, V, TVMap),
        map.det_insert(!.Map, U, TVMap, !:Map)
    ).

% stmt_block_to_cfg(+Stmts, +BBCurrent, +BBAfter, +DefMap,
%   !AfterPredMap, !ExitPredMap, !CFG).
% Compiles a statement block into basic blocks. Begins inserting statements
% into BBCurrent, and may create more blocks as necessary.
% BBCurrent must be incomplete (have a null terminator). Will modify CFG such
% that BBCurrent is complete, as are all newly-created branches.
% All code paths will end up branching into BBAfter unless explicitly branched
% elsewhere (eg. via a return statement).
% Will not modify BBAfter (just uses it as a branch target).
% AfterPredMap contains information about predecessors to the BBAfter block --
% this may append to the map information about blocks which branch to BBAfter,
% for the purpose of computing phis.
% Throws a context error if there is an attempt to read from a variable which
% may not have been assigned on some or all code paths.
:- pred stmt_block_to_cfg(stmt_block::in, bbref(S)::in, bbref(S)::in,
    def_map::in, pred_map(S)::in, pred_map(S)::out,
    pred_map(S)::in, pred_map(S)::out, cfg(S)::in, cfg(S)::out)
    is det.
stmt_block_to_cfg(Stmts, BBCurrent, BBAfter, DefMap, !AfterPredMap,
    !ExitPredMap, !CFG) :-
    map.init(SubscriptMap),
    stmt_block_to_cfg(Stmts, BBCurrent, BBAfter, SubscriptMap, DefMap,
        !AfterPredMap, !ExitPredMap, !CFG).

:- pred stmt_block_to_cfg(stmt_block::in, bbref(S)::in, bbref(S)::in,
    subscript_map::in, def_map::in, pred_map(S)::in, pred_map(S)::out,
    pred_map(S)::in, pred_map(S)::out, cfg(S)::in, cfg(S)::out) is det.
stmt_block_to_cfg([], BBCurrent, BBAfter, _SubscriptMap, DefMap,
    !AfterPredMap, !ExitPredMap, !CFG) :-
    % Special case for empty block - just branch to BBAfter.
    add_to_pred_map(BBCurrent, DefMap, BBAfter, !AfterPredMap, !CFG),
    cfg.set_terminator(BBCurrent, branch(BBAfter), !CFG).
stmt_block_to_cfg([S], BBCurrent, BBAfter, SubscriptMap, DefMap,
    !AfterPredMap, !ExitPredMap, !CFG) :-
    % Base-case for the last statement in a block - compile statement into
    % BBCurrent, but end up by branching to BBAfter.
    stmt_to_cfg_last(S, BBCurrent, BBAfter, SubscriptMap, DefMap,
        !AfterPredMap, !ExitPredMap, !CFG).
stmt_block_to_cfg([S|Ss@[_|_]], BBCurrent, BBAfter, SubscriptMap0, DefMap0,
    !AfterPredMap, !ExitPredMap, !CFG) :-
    % Compile a single statement into BBCurrent. BBContinue is the block to
    % compile subsequent statements into (may be equal to BBCurrent).
    stmt_to_cfg(S, BBCurrent, BBContinue, SubscriptMap0, SubscriptMap,
        DefMap0, DefMap, !ExitPredMap, !CFG),
    % Compile subsequent statements into BBContinue.
    stmt_block_to_cfg(Ss, BBContinue, BBAfter, SubscriptMap, DefMap,
        !AfterPredMap, !ExitPredMap, !CFG).

%%% HIGH-LEVEL STATEMENT-COMPILATION PREDICATES %%%

% stmt_to_cfg(+Stmt, +BBCurrent, -BBAfter, !SubscriptMap, !DefMap,
%     !ExitPredMap, !CFG).
% Compiles a single statement into basic blocks. Assumes this statement is NOT
% the last statement in its containing block (if it is, no harm, but may
% generate redundant basic blocks).
% Inserts the statement into BBCurrent (and may generate new basic blocks for
% compound statements). BBCurrent must be incomplete.
% There are two possible results:
% - May bind BBAfter to BBCurrent, leaving it incomplete, and preserve
%   SubscriptMap (potentially updating it with new definitions).
% - May create a new block and bind BBAfter to it. In this case, BBCurrent
%   will be completed (have a terminator set), and SubscriptMap will be reset
%   to refer to the new block (and typically be empty).
% May also update def_map with any new variable definitions.
:- pred stmt_to_cfg(stmt::in, bbref(S)::in, bbref(S)::out,
        subscript_map::in, subscript_map::out, def_map::in, def_map::out,
        pred_map(S)::in, pred_map(S)::out, cfg(S)::in, cfg(S)::out) is det.
stmt_to_cfg(basic_stmt(S,Ctx), !BBRef, !SubscriptMap, !DefMap, !ExitPredMap,
    !CFG) :-
    basic_stmt_to_cfg(S, Ctx, !.BBRef, !SubscriptMap, !DefMap, !CFG).
stmt_to_cfg(compound_stmt(S,Ctx), BBCurrent, BBAfter, !SubscriptMap,
    !DefMap, !ExitPredMap, !CFG) :-
    % Generate a target block for the compound statement to branch to
    cfg.new_basic_block(BBAfter, !CFG),
    compound_stmt_to_cfg(S, Ctx, BBCurrent, BBAfter, !.SubscriptMap, !.DefMap,
        map.init, AfterPredMap, !ExitPredMap, !CFG),
    % AfterPredMap contains information about all predecessors to BBAfter.
    % Use this to compute phi instructions for BBAfter and update DefMap.
    AfterID = ref_id(BBAfter, !.CFG),
    reconcile_def_maps(AfterID, AfterPredMap, Phis, !:SubscriptMap, !:DefMap),
    cfg.set_phis(BBAfter, Phis, !CFG).

% stmt_to_cfg_last(+Stmt, +BBCurrent, +BBAfter, +SubscriptMap, +DefMap,
%     !AfterPredMap, !ExitPredMap, !CFG).
% Special version of stmt_to_cfg, used on the last statement in a statement
% block to avoid generating redundant empty basic blocks.
% Inserts the statement into BBCurrent, and ensures BBCurrent is complete.
% As with stmt_block_to_cfg, ensures all code paths end up branching to
% BBAfter unless explicitly branched elsewhere.
% AfterPredMap contains information about predecessors to the BBAfter block --
% this may append to the map information about blocks which branch to BBAfter,
% for the purpose of computing phis.
% May update DefMap with new bindings, but SubscriptMap is read-only (since it
% will definitely branch to a new block).
:- pred stmt_to_cfg_last(stmt::in, bbref(S)::in, bbref(S)::in,
    subscript_map::in, def_map::in, pred_map(S)::in, pred_map(S)::out,
    pred_map(S)::in, pred_map(S)::out, cfg(S)::in, cfg(S)::out) is det.
stmt_to_cfg_last(basic_stmt(S,Ctx), BBCurrent, BBAfter, SubscriptMap,
    DefMap, !AfterPredMap, !ExitPredMap, !CFG) :-
    basic_stmt_to_cfg(S, Ctx, BBCurrent, SubscriptMap, _, DefMap, NewDefMap,
        !CFG),
    % Branch to the provided BBAfter
    add_to_pred_map(BBCurrent, NewDefMap, BBAfter, !AfterPredMap, !CFG),
    cfg.set_terminator(BBCurrent, branch(BBAfter), !CFG).
stmt_to_cfg_last(compound_stmt(S,Ctx), BBCurrent, BBAfter, SubscriptMap,
    DefMap, !AfterPredMap, !ExitPredMap, !CFG) :-
    compound_stmt_to_cfg(S, Ctx, BBCurrent, BBAfter, SubscriptMap,
        DefMap, !AfterPredMap, !ExitPredMap, !CFG).

%%% LOW-LEVEL STATEMENT-COMPILATION PREDICATES %%%
% Either of the above two statement compilation predicates may call either of
% the two low-level predicates, which handle the compilation of each statement
% kind.

% exprs_to_instrs(+BlockID, +Ctx, +Exprs, -VarNames, -Instrs, !SubscriptMap).
:- pred exprs_to_instrs(int::in, context::in, list(expr)::in,
    list(varname)::out, list(instr)::out,
    subscript_map::in, subscript_map::out) is det.
exprs_to_instrs(BlockID, Ctx, Es, VarNames, Instrs, !SubscriptMap) :-
    exprs_to_instrs_(BlockID, Ctx, Es, [], VarNamesRev, [], LLInstrsRev,
        !SubscriptMap),
    list.reverse(VarNamesRev, VarNames),
    list.reverse(LLInstrsRev, LLInstrs),
    % "Condense" the list-of-lists into a flat list of instructions
    list.condense(LLInstrs, Instrs).

% exprs_to_instrs_ takes accumulators for varnames and instrs, and prepends
% (so it ends up reversed).
% Instrs is a list of lists (to be flattened). The outer list is reversed, the
% inner list is not -- this is to allow prepending.
:- pred exprs_to_instrs_(int::in, context::in, list(expr)::in,
    list(varname)::in, list(varname)::out,
    list(list(instr))::in, list(list(instr))::out,
    subscript_map::in, subscript_map::out) is det.
exprs_to_instrs_(_BlockID, _Ctx, [], !VarNames, !Instrs, !SubscriptMap).
exprs_to_instrs_(BlockID, Ctx, [E|Es], !VarNames, !Instrs, !SubscriptMap) :-
    % Create a new variable name and generate instructions for each expr
    expr_to_instrs(BlockID, Ctx, E, V, Is, !SubscriptMap),
    !:VarNames = [V | !.VarNames],
    !:Instrs = [Is | !.Instrs],     % Do not flatten Is; keep as list of lists
    exprs_to_instrs_(BlockID, Ctx, Es, !VarNames, !Instrs, !SubscriptMap).

% expr_to_instrs(+BlockID, +Ctx, +Expr, -VarName, -Instrs, !SubscriptMap).
% Converts a single expression into a sequence of low-level instructions
% which compute the expression and assign it to a variable (possibly a
% new temporary variable name). Returns the assigned variable name in VarName.
:- pred expr_to_instrs(int::in, context::in, expr::in, varname::out,
    list(instr)::out, subscript_map::in, subscript_map::out) is det.
expr_to_instrs(BlockID, Ctx, E, V, Is, !SubscriptMap) :-
    (E = expr(varref(Var), _) ->
        % Special case -- expr_to_instrs_as will generate a mov.
        % We can avoid this by not generating a temp var, and just returning
        % the variable name (with no instructions at all).
        V = Var,
        Is = []
    ;
        new_temp_variable(BlockID, V, !SubscriptMap),
        expr_to_instrs_as(BlockID, Ctx, E, V, Is, !SubscriptMap)
    ).

expr_to_instrs(E, V, Is) :-
    expr_to_instrs(0, context.blank_context, E, V, Is, map.init, _).

% expr_to_instrs_as(+BlockID, +Ctx, +Expr, +VarName, -Instrs, !SubscriptMap).
% Like expr_to_instrs, but it assigns to a specific VarName as given.
% (This is less efficient as it may generate a mov instruction).
:- pred expr_to_instrs_as(int::in, context::in, expr::in, varname::in,
    list(instr)::out, subscript_map::in, subscript_map::out) is det.
expr_to_instrs_as(_BlockID, Ctx, expr(intlit(Val),_), V, Is, !SubscriptMap) :-
    Is = [instr(ld_intlit(V, Val), Ctx)].
expr_to_instrs_as(BlockID, Ctx, expr(arraylit(Elems0),_), V, Is,
    !SubscriptMap) :-
    exprs_to_instrs(BlockID, Ctx, Elems0, Elems, Is0, !SubscriptMap),
    Is = Is0 ++ [instr(ld_arraylit(V, Elems), Ctx)].
expr_to_instrs_as(_BlockID, Ctx, expr(varref(Var),_), V, Is, !SubscriptMap) :-
    Is = [instr(mov(V, Var), Ctx)].
expr_to_instrs_as(_BlockID, Ctx, expr(ctorref(Name),_), V, Is,
    !SubscriptMap) :-
    Is = [instr(ld_ctor(V, Name), Ctx)].
expr_to_instrs_as(BlockID, Ctx, expr(fieldref(Obj0, Field),_), V, Is,
    !SubscriptMap) :-
    expr_to_instrs(BlockID, Ctx, Obj0, Obj, Is0, !SubscriptMap),
    Is = Is0 ++ [instr(ld_field(V, Obj, Field), Ctx)].
expr_to_instrs_as(BlockID, Ctx, expr(app(Func0, Args0),_), V, Is,
    !SubscriptMap) :-
    expr_to_instrs(BlockID, Ctx, Func0, Func, Is0, !SubscriptMap),
    exprs_to_instrs(BlockID, Ctx, Args0, Args, Is1, !SubscriptMap),
    Is2 = [instr(call(V, Func, Args), Ctx)],
    Is = Is0 ++ Is1 ++ Is2.
expr_to_instrs_as(BlockID, Ctx, expr(parapp(Func0, Args0),_), V, Is,
    !SubscriptMap) :-
    expr_to_instrs(BlockID, Ctx, Func0, Func, Is0, !SubscriptMap),
    exprs_to_instrs(BlockID, Ctx, Args0, Args, Is1, !SubscriptMap),
    Is2 = [instr(parcall(V, Func, Args), Ctx)],
    Is = Is0 ++ Is1 ++ Is2.

basic_stmt_to_instrs(BasicStmt, Instrs) :-
    basic_stmt_to_instrs(0, BasicStmt, context.blank_context, Instrs,
        map.init, _).

% basic_stmt_to_instrs(+BlockID, +BasicStmt, +Ctx, -Instrs, !SubscriptMap).
% Converts a single basic statement into a sequence of low-level instructions
% which do the same thing. This will potentially assign many temporary
% variables.
:- pred basic_stmt_to_instrs(int::in, basic_stmt::in, context::in,
    list(instr)::out, subscript_map::in, subscript_map::out) is det.
basic_stmt_to_instrs(_BlockID, pass, _Ctx, [], !SubscriptMap).
basic_stmt_to_instrs(BlockID, assign(V, E), Ctx, Is, !SubscriptMap) :-
    expr_to_instrs_as(BlockID, Ctx, E, V, Is, !SubscriptMap).
basic_stmt_to_instrs(BlockID, fieldset(Obj0, Field, Val0), Ctx, Is,
    !SubscriptMap) :-
    expr_to_instrs(BlockID, Ctx, Obj0, Obj, Is0, !SubscriptMap),
    expr_to_instrs(BlockID, Ctx, Val0, Val, Is1, !SubscriptMap),
    Is2 = [instr(i_fieldset(Obj, Field, Val), Ctx)],
    Is = Is0 ++ Is1 ++ Is2.
basic_stmt_to_instrs(BlockID, fieldreplace(Obj0, Field, Val0), Ctx, Is,
    !SubscriptMap) :-
    % XXX fieldreplace statement is very weird (should be an expr).
    % Currently re-assign to the same variable.
    expr_to_instrs(BlockID, Ctx, Obj0, Obj, Is0, !SubscriptMap),
    expr_to_instrs(BlockID, Ctx, Val0, Val, Is1, !SubscriptMap),
    Is2 = [instr(i_fieldreplace(Obj, Obj, Field, Val), Ctx)],
    Is = Is0 ++ Is1 ++ Is2.
basic_stmt_to_instrs(BlockID, eval(E), Ctx, Is, !SubscriptMap) :-
    % The variable _T is assigned to, then thrown away.
    expr_to_instrs(BlockID, Ctx, E, _T, Is, !SubscriptMap).

% basic_stmt_to_cfg(+BasicStmt, +Ctx, +BBCurrent, !SubscriptMap, !DefMap,
%   !CFG).
% Compiles a single basic statement into a basic block. Appends the statement
% to the end of BBCurrent. May also update SubscriptMap and DefMap with any
% newly-defined variables.
:- pred basic_stmt_to_cfg(basic_stmt::in, context::in, bbref(S)::in,
    subscript_map::in, subscript_map::out, def_map::in, def_map::out,
    cfg(S)::in, cfg(S)::out) is det.
% "pass" statements disappear at this point
basic_stmt_to_cfg(pass, _Ctx, _BBCurrent, !SubscriptMap, !DefMap, !CFG).
basic_stmt_to_cfg(assign(Var0, E0), Ctx, BBCurrent, !SubscriptMap, !DefMap,
    !CFG) :-
    apply_def_map_to_expr(!.DefMap, Ctx, E0, E),
    BlockID = ref_id(BBCurrent, !.CFG),
    new_ssa_variable(BlockID, Var0, Var, !SubscriptMap, !DefMap),
    basic_stmt_to_instrs(BlockID, assign(Var, E), Ctx, Instrs, !SubscriptMap),
    cfg.append_instrs(BBCurrent, Instrs, !CFG).
basic_stmt_to_cfg(fieldset(Target0,FieldName,Expr0), Ctx, BBCurrent,
    !SubscriptMap, !DefMap, !CFG) :-
    apply_def_map_to_expr(!.DefMap, Ctx, Target0, Target),
    apply_def_map_to_expr(!.DefMap, Ctx, Expr0, Expr),
    BlockID = ref_id(BBCurrent, !.CFG),
    basic_stmt_to_instrs(BlockID, fieldset(Target, FieldName, Expr), Ctx,
        Instrs, !SubscriptMap),
    cfg.append_instrs(BBCurrent, Instrs, !CFG).
basic_stmt_to_cfg(fieldreplace(Target0,FieldName,Expr0), Ctx, BBCurrent,
    !SubscriptMap, !DefMap, !CFG) :-
    apply_def_map_to_expr(!.DefMap, Ctx, Target0, Target),
    apply_def_map_to_expr(!.DefMap, Ctx, Expr0, Expr),
    BlockID = ref_id(BBCurrent, !.CFG),
    basic_stmt_to_instrs(BlockID, fieldreplace(Target, FieldName, Expr), Ctx,
        Instrs, !SubscriptMap),
    cfg.append_instrs(BBCurrent, Instrs, !CFG).
basic_stmt_to_cfg(eval(E0), Ctx, BBCurrent, !SubscriptMap, !DefMap, !CFG) :-
    apply_def_map_to_expr(!.DefMap, Ctx, E0, E),
    BlockID = ref_id(BBCurrent, !.CFG),
    basic_stmt_to_instrs(BlockID, eval(E), Ctx, Instrs, !SubscriptMap),
    cfg.append_instrs(BBCurrent, Instrs, !CFG).

% compound_stmt_to_cfg(+CompoundStmt, +Ctx, +BBCurrent, +BBAfter,
%   +SubscriptMap, +DefMap, !AfterPredMap, !ExitPredMap, !CFG).
% Compiles a single compound statement into a basic block. Appends statements
% to the end of BBCurrent and completes that block (with a terminator). May
% generate new basic blocks, and ensures they are complete as well.
% As with stmt_block_to_cfg, ensures all code paths end up branching to
% BBAfter unless explicitly branched elsewhere.
% AfterPredMap contains information about predecessors to the BBAfter block --
% this may append to the map information about blocks which branch to BBAfter,
% for the purpose of computing phis.
% SubscriptMap is read-only (since it will definitely branch to a new block).
% DefMap is read-only (the resulting def-maps should be appended to
% AfterPredMap).
:- pred compound_stmt_to_cfg(compound_stmt::in, context::in, bbref(S)::in,
    bbref(S)::in, subscript_map::in, def_map::in,
    pred_map(S)::in, pred_map(S)::out, pred_map(S)::in, pred_map(S)::out,
    cfg(S)::in, cfg(S)::out) is det.
compound_stmt_to_cfg(return(Expr), Ctx, BBCurrent, _BBAfter, SubscriptMap,
    DefMap0, !AfterPredMap, !ExitPredMap, !CFG) :-
    % Return is expressed in CFG as $RET = expr, followed by a branch to the
    % exit block.
    basic_stmt_to_cfg(assign(retvname, Expr), Ctx, BBCurrent,
        SubscriptMap, _, DefMap0, DefMap, !CFG),
    ExitBlock = cfg.get_exit(!.CFG),
    add_to_pred_map(BBCurrent, DefMap, ExitBlock, !ExitPredMap, !CFG),
    cfg.set_terminator(BBCurrent, branch(ExitBlock), !CFG).
compound_stmt_to_cfg(switch(Ctrl0, Cases0), Ctx, BBCurrent, BBAfter,
    SubscriptMap, DefMap, !AfterPredMap, !ExitPredMap, !CFG) :-
    % Calculate the condition expression at the end of BBCurrent
    apply_def_map_to_expr(DefMap, Ctx, Ctrl0, Ctrl),
    BlockID = ref_id(BBCurrent, !.CFG),
    expr_to_instrs(BlockID, Ctx, Ctrl, CtrlVar, CtrlInstrs, SubscriptMap, _),
    cfg.append_instrs(BBCurrent, CtrlInstrs, !CFG),
    list.map_foldl3(case_stmt_to_cfg(BBCurrent,BBAfter,SubscriptMap,DefMap),
        Cases0, Cases, !AfterPredMap, !ExitPredMap, !CFG),
    cfg.set_terminator(BBCurrent, t_switch(CtrlVar, Cases), !CFG).
compound_stmt_to_cfg(if_then_else(Cond0,ThenPart,ElsePart), Ctx, BBCurrent,
    BBAfter, SubscriptMap, DefMap, !AfterPredMap, !ExitPredMap, !CFG) :-
    % Calculate the condition expression at the end of BBCurrent
    apply_def_map_to_expr(DefMap, Ctx, Cond0, Cond),
    BlockID = ref_id(BBCurrent, !.CFG),
    expr_to_instrs(BlockID, Ctx, Cond, CondVar, CondInstrs, SubscriptMap, _),
    cfg.append_instrs(BBCurrent, CondInstrs, !CFG),
    % Create a new basic block for the "then" part.
    cfg.new_basic_block(ThenEnter, !CFG),
    cfg.append_predecessor(ThenEnter, BBCurrent, !CFG),
    stmt_block_to_cfg(ThenPart, ThenEnter, BBAfter, DefMap, !AfterPredMap,
        !ExitPredMap, !CFG),
    ( list.is_empty(ElsePart) ->
        % Branch straight to the "after" block if the condition fails
        add_to_pred_map(BBCurrent, DefMap, BBAfter, !AfterPredMap, !CFG),
        cfg.set_terminator(BBCurrent,
            cond_branch(CondVar,ThenEnter,BBAfter), !CFG)
    ;
        % Create a new basic block for the "else" part.
        cfg.new_basic_block(ElseEnter, !CFG),
        cfg.append_predecessor(ElseEnter, BBCurrent, !CFG),
        stmt_block_to_cfg(ElsePart, ElseEnter, BBAfter, DefMap, !AfterPredMap,
            !ExitPredMap, !CFG),
        % Have the current block terminate by branching into the then/else
        % blocks
        cfg.set_terminator(BBCurrent,
            cond_branch(CondVar,ThenEnter,ElseEnter), !CFG)
    ).
compound_stmt_to_cfg(while(Cond0, Body), Ctx, BBCurrent, BBAfter,
    SubscriptMap, DefMap, !AfterPredMap, !ExitPredMap, !CFG) :-
    % Create a new basic block for the condition and body
    cfg.new_basic_block(CondBlock, !CFG),
    cfg.new_basic_block(BodyEnter, !CFG),
    % BBCurrent unconditionally jumps to CondBlock
    add_to_pred_map(BBCurrent, DefMap, CondBlock, map.init, CondBlockPredMap,
        !CFG),
    cfg.set_terminator(BBCurrent, branch(CondBlock), !CFG),

    % XXX: The hardest part of the whole ast->cfg process - while loops
    % (chicken and egg problem here - we need to know the phi variables for
    % CondBlock before constructing BodyEnter, but can't because it might
    % depend upon the bindings in the body).
    % For now (very inefficient) we should be able to compute this by doing a
    % "dummy run" of the body, just for the purpose of determining CondBlock's
    % pred map. Note that this breaks linearity of CFG.

    % Do the whole fixpoint. Updates CFG by setting CondBlock's phis, and
    % compiling Body.
    while_fixpoint(CondBlock, BodyEnter, Body, CondBlockPredMap,
        !ExitPredMap, DefMap, CondBlockDefMap, !CFG),

    % CondBlock branches to BodyBlock or BBAfter, based on the condition
    % Note: Use append_predecessor (not add_to_pred_map) because CondBlock is
    % BodyEnter's only predecessor (so no phis need to be computed).
    cfg.append_predecessor(BodyEnter, CondBlock, !CFG),
    % Note: Use add_to_pred_map because After needs to compute phis.
    add_to_pred_map(CondBlock, CondBlockDefMap, BBAfter, !AfterPredMap, !CFG),
    % Only now can we apply the defmap to Cond, to subscript its variables.
    % Calculate the condition expression at the end of CondBlock
    apply_def_map_to_expr(CondBlockDefMap, Ctx, Cond0, Cond),
    BlockID = ref_id(CondBlock, !.CFG),
    expr_to_instrs(BlockID, Ctx, Cond, CondVar, CondInstrs, SubscriptMap, _),
    cfg.append_instrs(CondBlock, CondInstrs, !CFG),
    cfg.set_terminator(CondBlock, cond_branch(CondVar, BodyEnter, BBAfter),
                       !CFG).

% while_fixpoint(CondBlock, BodyEnter, Body, CondPredMap, !ExitPredMap,
%   !DefMap, !CFG).
% Does fixpoint iteration over the condition and body block of the while loop
% until they agree on the phi nodes for the condition block.
% Updates the CFG as follows:
%  - Sets the phis of CondBlock to the computed phi nodes.
%  - Compiles the body into BodyEnter, branching to CondBlock.
:- pred while_fixpoint(bbref(S)::in, bbref(S)::in, stmt_block::in,
    pred_map(S)::in, pred_map(S)::in, pred_map(S)::out,
    def_map::in, def_map::out, cfg(S)::in, cfg(S)::out) is det.
while_fixpoint(CondBlock, BodyEnter, Body, CondPredMap, !ExitPredMap,
    !DefMap, !CFG) :-
    % First iteration, assume there are no phis
    while_fixpoint_(CondBlock, BodyEnter, Body, CondPredMap, [], !ExitPredMap,
        !DefMap, !CFG).

:- pred while_fixpoint_(bbref(S)::in, bbref(S)::in, stmt_block::in,
    pred_map(S)::in, list(phi(S))::in, pred_map(S)::in, pred_map(S)::out,
    def_map::in, def_map::out, cfg(S)::in, cfg(S)::out) is det.
while_fixpoint_(CondBlock, BodyEnter, Body, CondPredMap0, Phis0,
    !ExitPredMap, !CondDefMap, !CFG) :-
    % Compute CondPredMap (after another iteration of the body)
    stmt_block_to_cfg(Body, BodyEnter, CondBlock, !.CondDefMap,
        CondPredMap0, CondPredMap, !ExitPredMap, !.CFG, PotentialFinalCFG),
    % Now we can compute the phis for the cond block
    CondBlockID = ref_id(CondBlock, !.CFG),
    reconcile_def_maps(CondBlockID, CondPredMap, Phis, _, CondDefMap1),
    % Determine whether phis changed (if not, we reached fixpoint)
    ( Phis0 = Phis ->
        !:CFG = PotentialFinalCFG,
        % Write the phi into the cond block
        cfg.set_phis(CondBlock, Phis, !CFG),
        !:CondDefMap = CondDefMap1
    ;
        % Now iterate till fixpoint
        while_fixpoint_(CondBlock, BodyEnter, Body, CondPredMap, Phis,
            !ExitPredMap, CondDefMap1, !:CondDefMap, !CFG)
    ).

% case_stmt_to_cfg(+BBAfter, +SubscriptMap, +DefMap, +CaseStmt, -CFGCaseStmt,
%   !AfterPredMap, !ExitPredMap, !CFG).
% Compiles a case statement into a CFG case statement (which is
% branch-oriented rather than sub-statement-block-oriented).
% Compiles the sub-statement-block in the case statement to new basic blocks,
% which will ultimately branch to BBAfter.
% BBCurrent is not modified, just used as a predecessor.
% AfterPredMap contains information about predecessors to the BBAfter block --
% this may append to the map information about blocks which branch to BBAfter,
% for the purpose of computing phis.
% SubscriptMap is read-only (since it will definitely branch to a new block).
% DefMap is read-only (the resulting def-maps should be appended to
% AfterPredMap).
:- pred case_stmt_to_cfg(bbref(S)::in, bbref(S)::in,
    subscript_map::in, def_map::in, case_stmt::in, cfg_case_stmt(S)::out,
    pred_map(S)::in, pred_map(S)::out, pred_map(S)::in, pred_map(S)::out,
    cfg(S)::in, cfg(S)::out) is det.
case_stmt_to_cfg(BBCurrent, BBAfter, SubscriptMap, DefMap,
    case_stmt(Pattern0, Stmts, Ctx), cfg_case_stmt(Pattern, BBFirst, Ctx),
    !AfterPredMap, !ExitPredMap, !CFG) :-
    % Traverse the pattern and uniquely label each newly-introduced
    % variable, as well as adding them to DefMap.
    BlockID = ref_id(BBCurrent, !.CFG),
    pattern_to_ssa(BlockID, Pattern0, Pattern, SubscriptMap, _, DefMap,
        InnerScopeDefMap),

    cfg.new_basic_block(BBFirst, !CFG),
    cfg.append_predecessor(BBFirst, BBCurrent, !CFG),
    stmt_block_to_cfg(Stmts, BBFirst, BBAfter, InnerScopeDefMap,
        !AfterPredMap, !ExitPredMap, !CFG).

% pattern_to_ssa(BlockID, !Pattern, !SubscriptMap, !DefMap).
% Converts a pattern to SSA form by replacing all of its introduced variables
% with subscripted SSA variables.
% BlockID: An int identifying the block the pattern is in (for computing
% subscripts).
% SubscriptMap and DefMap are updated with the new variable definitions.
:- pred pattern_to_ssa(int::in, pattern::in, pattern::out,
    subscript_map::in, subscript_map::out, def_map::in, def_map::out) is det.
pattern_to_ssa(BlockID, pattern(Pat0, PatType), pattern(Pat, PatType),
    !SubscriptMap, !DefMap) :-
    (
        Pat0 = pat_any,
        Pat = Pat0
    ;
        Pat0 = pat_var(Var0),
        new_ssa_variable(BlockID, Var0, Var, !SubscriptMap, !DefMap),
        Pat = pat_var(Var)
    ;
        Pat0 = pat_ctor(CtorName, SubPatterns0),
        (
            SubPatterns0 = yes(YesSubPatterns0),
            map_foldl2(pattern_to_ssa(BlockID), YesSubPatterns0, SubPatterns,
                !SubscriptMap, !DefMap),
            Pat = pat_ctor(CtorName, yes(SubPatterns))
        ;
            SubPatterns0 = no,
            Pat = pat_ctor(CtorName, no)
        )
    ;
        Pat0 = pat_intlit(_),
        Pat = Pat0
    ).
