% -------------------------------------------------------------------------- %
% Mars/Venus builtin types / functions
% Copyright (C) 2008 Matt Giuca
% 3/6/2008
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- module builtins.

:- interface.

:- import_module tables.

% builtins_progtable is a progtable containing definitions of all the builtin
% functions and types offered by the language.
:- func builtins_progtable = tables.progtable.

% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- implementation.

:- import_module bool.
:- import_module stack.
:- import_module list.
:- import_module pair.
:- import_module maybe.
:- import_module varset.

:- import_module ir.
:- import_module types.
:- import_module context.

% This function just throws all the builtins into a table.
% See below for the declarations of each builtin.
builtins_progtable = Table :-
    some [!Table] (
        !:Table = tables.empty_progtable,
        % Builtin types %
        % Note: There is no table entry for the special built-in type
        % "function". That is handled by special cases.
        update_progtable_typedef(type_int,   !Table),
        update_progtable_typedef(type_array, !Table),

        % Builtin functions %
        % Comparison
        update_progtable_function(func_eq,  !Table),
        update_progtable_function(func_is,  !Table),
        update_progtable_function(func_cmp, !Table),
        % Arithmetic
        update_progtable_function(func_add, !Table),
        update_progtable_function(func_sub, !Table),
        update_progtable_function(func_mul, !Table),
        update_progtable_function(func_div, !Table),
        update_progtable_function(func_mod, !Table),
        % Array
        update_progtable_function(func_array,         !Table),
        update_progtable_function(func_array_ref,     !Table),
        update_progtable_function(func_impure_array_set,     !Table),
        update_progtable_function(func_array_replace, !Table),
        update_progtable_function(func_array_length,  !Table),
        update_progtable_function(func_impure_array_append,  !Table),
        update_progtable_function(func_impure_array_extend,  !Table),
        update_progtable_function(func_array_concat,  !Table),
        % IO
        update_progtable_function(func_put_char, !Table),
        update_progtable_function(func_get_char, !Table),
        update_progtable_function(func_get_env,  !Table),
        % Miscellaneous
        update_progtable_function(func_show,  !Table),
        update_progtable_function(func_error, !Table),
        Table = !.Table
    ).

:- func ctx = context.context.
ctx = context("builtins", 1, 1, stack.init).

% -------------------------------------------------------------------------- %
% Builtin Types %

:- func type_int = typedef.
type_int = typedef("Int", no, [], varset.init, ctx).

:- func type_array = typedef.
type_array = typedef("Array", yes([A]), [], Varset, ctx) :-
    varset.new_named_var(varset.init, "a", A, Varset).

% -------------------------------------------------------------------------- %
% Builtin Functions %

% Comparison %

:- func func_eq = function.
func_eq = function("eq", yes([{"x", types.variable(A), def_argmode},
                              {"y", types.variable(A), def_argmode}]),
                           types.const("Int"),
                           Varset,
                           func_builtin,
                           ctx) :-
    varset.new_named_var(varset.init, "a", A, Varset).

:- func func_is = function.
func_is = function("is", yes([{"x", types.variable(A), def_argmode},
                              {"y", types.variable(A), def_argmode}]),
                           types.const("Int"),
                           Varset,
                           func_builtin,
                           ctx) :-
    varset.new_named_var(varset.init, "a", A, Varset).

:- func func_cmp = function.
func_cmp = function("cmp", yes([{"x", types.const("Int"), def_argmode},
                                {"y", types.const("Int"), def_argmode}]),
                            types.const("Int"),
                            varset.init,
                            func_builtin,
                            ctx).

% Arithmetic %

:- func func_add = function.
func_add = function("add", yes([{"x", types.const("Int"), def_argmode},
                                {"y", types.const("Int"), def_argmode}]),
                           types.const("Int"),
                           varset.init,
                           func_builtin,
                           ctx).

:- func func_sub = function.
func_sub = function("sub", yes([{"x", types.const("Int"), def_argmode},
                                {"y", types.const("Int"), def_argmode}]),
                           types.const("Int"),
                           varset.init,
                           func_builtin,
                           ctx).

:- func func_mul = function.
func_mul = function("mul", yes([{"x", types.const("Int"), def_argmode},
                                {"y", types.const("Int"), def_argmode}]),
                           types.const("Int"),
                           varset.init,
                           func_builtin,
                           ctx).

:- func func_div = function.
func_div = function("div", yes([{"x", types.const("Int"), def_argmode},
                                {"y", types.const("Int"), def_argmode}]),
                           types.const("Int"),
                           varset.init,
                           func_builtin,
                           ctx).

:- func func_mod = function.
func_mod = function("mod", yes([{"x", types.const("Int"), def_argmode},
                                {"y", types.const("Int"), def_argmode}]),
                           types.const("Int"),
                           varset.init,
                           func_builtin,
                           ctx).

% Array %

:- func func_array = function.
func_array = function("array",
                           yes([{"length", types.const("Int"), def_argmode},
                                {"default", types.variable(A), def_argmode}]),
                           types.app(types.const("Array"),
                               [types.variable(A)]),
                           Varset,
                           func_builtin,
                           ctx) :-
    varset.new_named_var(varset.init, "a", A, Varset).

:- func func_array_ref = function.
func_array_ref = function("array_ref",
                           yes([{"array", types.app(
                                        types.const("Array"),
                                        [types.variable(A)]), def_argmode},
                                {"index", types.const("Int"), def_argmode}]),
                           types.variable(A),
                           Varset,
                           func_builtin,
                           ctx) :-
    varset.new_named_var(varset.init, "a", A, Varset).

:- func func_impure_array_set = function.
func_impure_array_set = function("__impure_array_set",
                           yes([{"array", types.app(
                                        types.const("Array"),
                                        [types.variable(A)]), def_argmode},
                                {"index", types.const("Int"), def_argmode},
                                {"value", types.variable(A), def_argmode}]),
                           types.app(types.const("Array"),
                               [types.variable(A)]),
                           Varset,
                           func_builtin,
                           ctx) :-
    varset.new_named_var(varset.init, "a", A, Varset).

:- func func_array_replace = function.
func_array_replace = function("array_replace",
                           yes([{"array", types.app(
                                        types.const("Array"),
                                        [types.variable(A)]), def_argmode},
                                {"index", types.const("Int"), def_argmode},
                                {"value", types.variable(A), def_argmode}]),
                           types.app(types.const("Array"),
                               [types.variable(A)]),
                           Varset,
                           func_builtin,
                           ctx) :-
    varset.new_named_var(varset.init, "a", A, Varset).

:- func func_array_length = function.
func_array_length = function("array_length",
                           yes([{"array", types.app(
                                        types.const("Array"),
                                        [types.variable(A)]), def_argmode}]),
                           types.const("Int"),
                           Varset,
                           func_builtin,
                           ctx) :-
    varset.new_named_var(varset.init, "a", A, Varset).

:- func func_impure_array_append = function.
func_impure_array_append = function("__impure_array_append",
                           yes([{"array", types.app(
                                        types.const("Array"),
                                        [types.variable(A)]), def_argmode},
                            {"value", types.variable(A), def_argmode}]),
                           types.app(types.const("Array"),
                               [types.variable(A)]),
                           Varset,
                           func_builtin,
                           ctx) :-
    varset.new_named_var(varset.init, "a", A, Varset).

:- func func_impure_array_extend = function.
func_impure_array_extend = function("__impure_array_extend",
                           yes([{"array", types.app(
                                        types.const("Array"),
                                        [types.variable(A)]), def_argmode},
                                {"values", types.app(
                                        types.const("Array"),
                                        [types.variable(A)]), def_argmode}]),
                           types.app(types.const("Array"),
                               [types.variable(A)]),
                           Varset,
                           func_builtin,
                           ctx) :-
    varset.new_named_var(varset.init, "a", A, Varset).

:- func func_array_concat = function.
func_array_concat = function("array_concat",
                           yes([{"array1", types.app(
                                        types.const("Array"),
                                        [types.variable(A)]), def_argmode},
                                {"array2", types.app(
                                        types.const("Array"),
                                        [types.variable(A)]), def_argmode}]),
                           types.app(types.const("Array"), [types.variable(A)]),
                           Varset,
                           func_builtin,
                           ctx) :-
    varset.new_named_var(varset.init, "a", A, Varset).

% IO %

:- func func_put_char = function.
func_put_char = function("put_char",
                           yes([{"c", types.const("Int"), def_argmode}]),
                           types.const("Int"),
                           varset.init,
                           func_builtin,
                           ctx).

:- func func_get_char = function.
func_get_char = function("get_char",
                           yes([]),
                           types.const("Int"),
                           varset.init,
                           func_builtin,
                           ctx).

:- func func_get_env = function.
func_get_env = function("get_env",
                           yes([{"name", types.app(types.const("Array"),
                                                   [types.const("Int")]),
                                                   def_argmode}]),
                           types.app(types.const("Array"),
                                     [types.const("Int")]),
                           varset.init,
                           func_builtin,
                           ctx).

% Miscellaneous %

:- func func_show = function.
func_show = function("show", yes([{"x", types.variable(A), def_argmode}]),
                           types.app(types.const("Array"),
                               [types.const("Int")]),
                           Varset,
                           func_builtin,
                           ctx) :-
    varset.new_named_var(varset.init, "a", A, Varset).

:- func func_error = function.
func_error = function("error",
                           yes([{"msg", types.app(types.const("Array"),
                                              [types.const("Int")]),
                                              def_argmode}]),
                           types.variable(A),
                           Varset,
                           func_builtin,
                           ctx) :-
    varset.new_named_var(varset.init, "a", A, Varset).
