% -------------------------------------------------------------------------- %
% Mars/Venus Implementation
% Copyright (C) 2008 Matt Giuca
% 23/9/2008
%
% CFG (Control Flow Graph)
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %
% The CFG represents the code to a single function as a series of basic
% blocks, which contain basic statements and link to each other via terminator
% statements.
% This representation is ideal for analysis and makes certain problems
% trivially easy to diagnose (such as not returning a value or using a
% variable without definition).
% -------------------------------------------------------------------------- %

:- module cfg.

:- interface.

:- import_module ir.

:- import_module list.
:- import_module map.

% Note: cfg and bbref types have an S parameter, which is the store type.
:- type cfg(S).

% A bbref (Basic Block Reference) is an opaque handle on a basic block within
% the cfg with store S.
:- type bbref(S).

% A phi is a phi statement in Static Single Assignment notation.
% It maps a variable name V onto a mapping from bbrefs to varnames.
% The bbref is a predecessor block P, and the varname is V0.
% That is to say, "if you just came from block P, set V to the value of V0".
:- type phi(S)
    --->    phi(
                phi_varname :: varname,
                phi_preds   :: map(bbref(S), varname)
            ).

% A basic_block can be looked up in the CFG given a bbref.
% This data structure is very much designed to be built sequentially by
% appending to various components.
% Its constructor (new_basic_block) just sets bb_id and all empty lists. The
% block is then to be created by populating all lists except bb_phis, then a
% final pass is done to compute bb_phis.
:- type basic_block(S)
    --->    basic_block(
                % bb_id is unique to this CFG
                bb_id           :: int,
                bb_phis         :: list(phi(S)),
                % bb_body is stored in reverse. Should be traversed with
                % foldl_body, which traverses in the correct order.
                bb_body         :: list(instr),
                bb_terminator   :: terminator_instr(S),
                bb_predecessors :: list(bbref(S)),
                % bb_defs - all vars written-to in this block
                bb_defs         :: list(varname),
                % bb_refs - all vars read-from in this block
                bb_refs         :: list(varname)
            ).

% new_cfg creates a new cfg(S) object.
:- some [S] func new_cfg = cfg(S).

% Fold over all the basic blocks in the CFG, in order.
% Passes a basic block reference and the basic block.
:- pred foldl_cfg(pred(bbref(S), basic_block(S), A, A), cfg(S), A, A).
:- mode foldl_cfg(pred(in, in, in, out) is det, in, in, out) is det.
:- mode foldl_cfg(pred(in, in, in, out) is semidet, in, in, out) is semidet.
:- mode foldl_cfg(pred(in, in, di, uo) is det, in, di, uo) is det.

:- pred foldl2_cfg(pred(bbref(S), basic_block(S), A, A, B, B), cfg(S),
                    A, A, B, B).
:- mode foldl2_cfg(pred(in, in, in, out, di, uo) is det, in, in, out, di, uo)
    is det.

% Set the name of the return value in the CFG.
% (Typically will be some qualified form of retvname).
:- pred set_retv(varname::in, cfg(S)::in, cfg(S)::out) is det.
% Get the name of the return value in the CFG.
:- func get_retv(cfg(S)) = varname.

% new_basic_block appends a basic block to an existing CFG, and returns a
% bbref which references the newly-created block.
% The new basic block has a unique id.
:- pred new_basic_block(bbref(S)::out, cfg(S)::in, cfg(S)::out) is det.

% Get a basic block from the CFG.
:- pred get_basic_block(bbref(S)::in, basic_block(S)::out, cfg(S)::in) is det.

% Overwrite a basic block in the CFG.
:- pred set_basic_block(bbref(S)::in, basic_block(S)::in,
    cfg(S)::in, cfg(S)::out) is det.

% Append a phi to a basic block in the CFG.
:- pred append_phi(bbref(S)::in, phi(S)::in, cfg(S)::in, cfg(S)::out) is det.
% Overwrite the phis in a basic block.
:- pred set_phis(bbref(S)::in, list(phi(S))::in, cfg(S)::in, cfg(S)::out)
    is det.
% Append an instr to a basic block in the CFG.
:- pred append_instr(bbref(S)::in, instr::in,
    cfg(S)::in, cfg(S)::out) is det.
% Append a sequence of instrs to a basic block in the CFG.
:- pred append_instrs(bbref(S)::in, list(instr)::in,
    cfg(S)::in, cfg(S)::out) is det.
% Replace the terminator statement in a basic block with the given one.
:- pred set_terminator(bbref(S)::in, terminator_instr(S)::in,
    cfg(S)::in, cfg(S)::out) is det.
% Append a predecessor to a basic block in the CFG.
:- pred append_predecessor(bbref(S)::in, bbref(S)::in,
    cfg(S)::in, cfg(S)::out) is det.
% Append a def variable to a basic block in the CFG.
:- pred append_def(bbref(S)::in, varname::in, cfg(S)::in, cfg(S)::out) is det.
% Append a ref variable to a basic block in the CFG.
:- pred append_ref(bbref(S)::in, varname::in, cfg(S)::in, cfg(S)::out) is det.

% Folds over the sequence of instrs in the basic block's body, in forwards
% order. (Since they are stored in reverse, this actually calls foldr).
:- pred foldl_body(pred(instr, T, T), basic_block(S), T, T) is det.
:- mode foldl_body(pred(in, in, out) is det, in, in, out) is det.
:- mode foldl_body(pred(in, in, out) is semidet, in, in, out) is semidet.
:- mode foldl_body(pred(in, in, out) is nondet, in, in, out) is nondet.
:- mode foldl_body(pred(in, di, uo) is det, in, di, uo) is det.

:- pred foldl2_body(pred(instr, T, T, U, U), basic_block(S), T, T, U, U)
    is det.
:- mode foldl2_body(pred(in, in, out, di, uo) is det, in, in, out, di, uo)
    is det.

:- pred foldl3_body(pred(instr, T, T, U, U, V, V), basic_block(S),
    T, T, U, U, V, V) is det.
:- mode foldl3_body(pred(in, in, out, in, out, in, out) is det,
    in, in, out, in, out, in, out) is det.
:- mode foldl3_body(pred(in, in, out, in, out, di, uo) is det,
    in, in, out, in, out, di, uo) is det.

% Folds over the sequence of instrs in the basic block's body, in backwards
% order. (Since they are stored in reverse, this actually calls foldl).
:- pred foldr_body(pred(instr, T, T), basic_block(S), T, T) is det.
:- mode foldr_body(pred(in, in, out) is det, in, in, out) is det.

% Map/folds over the sequence of instrs in the basic block's body, in
% backwards order.
:- pred map_foldr_body(pred(instr, instr, T, T),
    basic_block(S), basic_block(S), T, T).
:- mode map_foldr_body(pred(in, out, in, out) is det,
    in, out, in, out) is det.

% Succeeds if the block in the CFG's terminator is null_terminator.
:- pred is_null_terminated(bbref(S)::in, cfg(S)::in) is semidet.

% Retrieves the basic block ID associated with a particular basic block
% reference (for a given cfg).
:- func ref_id(bbref(S), cfg(S)) = int.

% Get the entry node of the CFG.
:- func get_entry(cfg(S)) = bbref(S).
% Get the exit node of the CFG.
:- func get_exit(cfg(S)) = bbref(S).

:- implementation.

:- import_module int.

:- import_module util.

% A CFG is a mapping from contiguous ints from 0 to basic blocks.
% CFGs always have an entry and exit node; the rest may be created by a code
% generator.
:- type cfg(S) ---> cfg(
                        cfg_map   :: map(int, basic_block(S)),
                        % The number of elements in the CFG (for generation of
                        % unique ids).
                        cfg_nel   :: int,
                        % Entry and exit nodes.
                        cfg_entry :: bbref(S),
                        cfg_exit  :: bbref(S),
                        % Name of the return value.
                        cfg_retv  :: varname,
                        % cfg_s is just a placeholder of type S. This allows
                        % new_cfg to initialise it so the existentially
                        % quantified type S gets bound.
                        cfg_s     :: S
                    ).

% A BBRef is a mutvar which references a basic block in a store S.
:- type bbref(S) ---> bbref(int).

% Bind the third field to 0, thereby giving it type int.
% It doesn't actually matter what value or type you give it, as long as it's
% something (to bind the existential type S).
new_cfg = cfg(Map, Nel, bbref(EnID), bbref(ExID), retvname, 0) :-
    map.init(Map0),
    EnID = 0,
    EnBlock = basic_block(EnID, [], [], null_terminator, [], [], []),
    map.det_insert(Map0, EnID, EnBlock, Map1),
    ExID = 1,
    ExBlock = basic_block(ExID, [], [], null_terminator, [], [], []),
    map.det_insert(Map1, ExID, ExBlock, Map),
    Nel = 2.


foldl_cfg(Pred, cfg(Blocks, Count, _En, _Ex, _, _), !A) :-
    foldl_cfg(Pred, Blocks, 0, Count, !A).

:- pred foldl_cfg(pred(bbref(S), basic_block(S), A, A),
    map(int, basic_block(S)), int, int, A, A).
:- mode foldl_cfg(pred(in, in, in, out) is det, in, in, in, in, out) is det.
:- mode foldl_cfg(pred(in, in, in, out) is semidet, in, in, in, in, out)
    is semidet.
:- mode foldl_cfg(pred(in, in, di, uo) is det, in, in, in, di, uo) is det.
foldl_cfg(Pred, Blocks, I, Count, !A) :-
    ( I < Count ->
        map.lookup(Blocks, I, Block),       % Should always find
        Pred(bbref(I), Block, !A),
        foldl_cfg(Pred, Blocks, I+1, Count, !A)
    ;
        true
    ).

foldl2_cfg(Pred, cfg(Blocks, Count, _En, _Ex, _, _), !A, !B) :-
    foldl2_cfg(Pred, Blocks, 0, Count, !A, !B).

:- pred foldl2_cfg(pred(bbref(S), basic_block(S), A, A, B, B),
    map(int, basic_block(S)), int, int, A, A, B, B).
:- mode foldl2_cfg(pred(in, in, in, out, di, uo) is det, in, in, in, in, out,
    di, uo) is det.
foldl2_cfg(Pred, Blocks, I, Count, !A, !B) :-
    ( I < Count ->
        map.lookup(Blocks, I, Block),       % Should always find
        Pred(bbref(I), Block, !A, !B),
        foldl2_cfg(Pred, Blocks, I+1, Count, !A, !B)
    ;
        true
    ).

set_retv(Name, !CFG) :-
    !:CFG = !.CFG ^ cfg_retv := Name.
get_retv(CFG) = CFG ^ cfg_retv.

new_basic_block(bbref(ID), cfg(Map0, ID, En, Ex, R, S),
    cfg(Map, ID+1, En, Ex, R, S)) :-
    NewBB = basic_block(ID, [], [], null_terminator, [], [], []),
    map.det_insert(Map0, ID, NewBB, Map).

% Note: Aborts if the bbref's ID is not found in the map. This should never be
% the case if this module's encapsulation holds (assuming bbrefs cannot be
% fabricated externally, which they shouldn't be able to be, given the
% existential type S).
get_basic_block(bbref(ID), BB, cfg(Map, _, _, _, _, _)) :-
    map.lookup(Map, ID, BB).

get_entry(CFG) = CFG ^ cfg_entry.
get_exit(CFG) = CFG ^ cfg_exit.

set_basic_block(bbref(ID), BB, cfg(Map0, N, En, Ex, R, S),
    cfg(Map, N, En, Ex, R, S)) :-
    map.det_update(Map0, ID, BB, Map).

append_phi(BBRef, Phi, !CFG) :-
    get_basic_block(BBRef, BB, !.CFG),
    BB1 = BB ^ bb_phis := [Phi | BB ^ bb_phis],
    set_basic_block(BBRef, BB1, !CFG).

set_phis(BBRef, Phis, !CFG) :-
    get_basic_block(BBRef, BB, !.CFG),
    BB1 = BB ^ bb_phis := Phis,
    set_basic_block(BBRef, BB1, !CFG).

append_instr(BBRef, Instr, !CFG) :-
    get_basic_block(BBRef, BB, !.CFG),
    BB1 = BB ^ bb_body := [Instr | BB ^ bb_body],
    set_basic_block(BBRef, BB1, !CFG).

append_instrs(BBRef, Instrs, !CFG) :-
    get_basic_block(BBRef, BB, !.CFG),
    % reverse Instr (since instructions are stored in reverse in the CFG)
    util.reverse_append(Instrs, BB ^ bb_body, NewBody),
    BB1 = BB ^ bb_body := NewBody,
    set_basic_block(BBRef, BB1, !CFG).

set_terminator(BBRef, Stmt, !CFG) :-
    get_basic_block(BBRef, BB, !.CFG),
    BB1 = BB ^ bb_terminator := Stmt,
    set_basic_block(BBRef, BB1, !CFG).

append_predecessor(BBRef, Pred, !CFG) :-
    get_basic_block(BBRef, BB, !.CFG),
    BB1 = BB ^ bb_predecessors := [Pred | BB ^ bb_predecessors],
    set_basic_block(BBRef, BB1, !CFG).

append_def(BBRef, Def, !CFG) :-
    get_basic_block(BBRef, BB, !.CFG),
    BB1 = BB ^ bb_defs := [Def | BB ^ bb_defs],
    set_basic_block(BBRef, BB1, !CFG).

append_ref(BBRef, Ref, !CFG) :-
    get_basic_block(BBRef, BB, !.CFG),
    BB1 = BB ^ bb_refs := [Ref | BB ^ bb_refs],
    set_basic_block(BBRef, BB1, !CFG).

foldl_body(Pred, Block, !Acc) :-
    foldr(Pred, Block ^ bb_body, !Acc).
foldl2_body(Pred, Block, !X, !Y) :-
    util.list_foldr2(Pred, Block ^ bb_body, !X, !Y).
foldl3_body(Pred, Block, !X, !Y, !Z) :-
    util.list_foldr3(Pred, Block ^ bb_body, !X, !Y, !Z).

foldr_body(Pred, Block, !Acc) :-
    foldl(Pred, Block ^ bb_body, !Acc).

map_foldr_body(Pred, !Block, !Acc) :-
    list.map_foldl(Pred, !.Block ^ bb_body, NewInstrs, !Acc),
    % Update the block
    !Block ^ bb_body := NewInstrs.

is_null_terminated(BBRef, CFG) :-
    get_basic_block(BBRef, Block, CFG),
    Block ^ bb_terminator = null_terminator.

ref_id(bbref(ID), _) = ID.
