% -------------------------------------------------------------------------- %
% Generic parser library - context / error types
% Copyright (C) 2008 Matt Giuca
% 1/6/2008
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- module context.

:- interface.

:- import_module stack.
:- import_module io.    % For print_error.

% Position information about the current state of a parse (line number).
:- type context
    --->    context(
                ctx_filename :: string,
                ctx_line :: int,
                ctx_column :: int,
                ctx_blocks :: stack(string)
            ).
    % ctx_blocks is a stack of names of typedefs or functions the
    % parser is in. This is used for error messages.

% An error object, which may be used as an exception type.
:- type context.error
    --->    context.error(
                error_context :: context,
                error_message :: string
            ).

% An empty context, which will cause error messages to omit the
% context entirely.
:- func context.blank_context = context.

% context.context_push_block(BlockName, !Context) pushes a new block on the
% stack (ie. when the parser is entering a new function or typedef).
:- pred context.context_push_block(string::in, context::in, context::out)
        is det.
% context.context_pop_block(BlockName, !Context) pops a block off the
% stack (ie. when the parser is exiting a function or typedef).
% Error if the block is already empty (this should not happen during normal
% parsing).
:- pred context.context_pop_block(context::in, context::out) is det.

% Throws a context.error.
:- pred context.throw_error(string::in, context::in) is erroneous.
% Throws a context.error, without a context.
:- pred context.throw_error(string::in) is erroneous.

% Prints an error to current output stream, with a newline.
:- pred context.print_error(context.error::in, io::di, io::uo) is det.

% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- implementation.

:- import_module string.
:- import_module exception.

blank_context = context("",1,1,stack.init).

context_push_block(Name, context(F, L, C, B0), context(F, L, C, B)) :-
    stack.push(B0, Name, B).

context_pop_block(context(F, L, C, B0), context(F, L, C, B)) :-
    stack.pop_det(B0, _, B).

throw_error(Msg, Ctx) :-
    throw(error(Ctx, Msg)).
throw_error(Msg) :-
    throw(error(blank_context, Msg)).

print_error(error(context(Filename, Line, _, BlockStack), Msg), !IO) :-
    % Note: We do nothing with Column, because some parsers may not
    % keep it updated.
    % Write to stderr.
    io.set_output_stream(io.stderr_stream, OldOutStream, !IO),
    ( Filename = "" ->
        true
    ;
        io.write_string(Filename ++ ":", !IO),
        io.write(Line, !IO),
        io.write_string(": ", !IO)
    ),
    % Get the name of the name of the innermost block we are in, if any
    ( stack.top(BlockStack, Block) ->
        io.write_string("(" ++ Block ++ ") ", !IO)
    ;
        true
    ),
    io.write_string(Msg ++ ".", !IO),
    io.nl(!IO),
    io.set_output_stream(OldOutStream, _, !IO).
