% -------------------------------------------------------------------------- %
% Mars/Venus interactive prompt
% Copyright (C) 2008 Matt Giuca
% 12/6/2008
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- module interactive.

:- interface.

:- import_module io.

:- import_module tables.

% prompts(ProgTable, !IO).
% asks the user for either:
%   - A statement, then typechecks and pretty-prints the statement
%   - :t an expression, then prints the type of the expression
%   - :k a type, then prints the kind of the type
%   - :b - prints a listing of all globals
% Handles the whole session. Loops until EOF.
% Normal errors (program or user errors) will be caught and printed.
% Only throws unexpected exceptions.
% Note: Expects that the module has already been compiled to CompiledFilename,
% using the Python backend. (Required, as it may use Python to query this
% file).
:- pred prompts(tables.progtable::in, io::di, io::uo) is cc_multi.

% do_user_command(CommandText, ProgTable, !IO)
% As for prompts, but non-interactive. Just executes the single user command
% CommandText.
:- pred do_user_command(string::in, tables.progtable::in, io::di, io::uo)
    is cc_multi.

:- implementation.

:- import_module bool.
:- import_module list.
:- import_module maybe.
:- import_module pair.
:- import_module int.
:- import_module char.
:- import_module string.
:- import_module univ.
:- import_module stack.
:- import_module varset.
:- import_module exception.
:- import_module map.

:- import_module ir.
:- import_module ast_cfg.
:- import_module token.
:- import_module parsem.
:- import_module context.
:- import_module types.
:- import_module typecheck.
:- import_module pretty.
:- import_module interpret.
:- import_module util.

prompts(ProgTable, !IO) :-
    % Call prompts with "" as the compiled filename. This is currently
    % OK, because CompiledFilename is not used within do_user_command.
    prompts("", ProgTable, !IO).

% Note: See do_user_command for a discussion on CompiledFilename.
:- pred prompts(string::in, tables.progtable::in, io::di, io::uo) is cc_multi.
prompts(CompiledFilename, ProgTable, !IO) :-
    % Compute the env to be used in all prompts, and a local table
    Env = interpret.progtable_env(ProgTable),
    LT = tables.empty_localtable,
    % Throw this env away when done
    prompts(CompiledFilename, ProgTable, LT, _NewLT, Env, _NewEnv, !IO).

:- pred prompts(string::in, tables.progtable::in,
                tables.localtable::in, tables.localtable::out,
                interpret.env::in, interpret.env::out,
                io::di, io::uo) is cc_multi.
prompts(CompiledFilename, ProgTable, !LT, !Env, !IO) :-
    util.try_io_thread2(prompt(CompiledFilename, ProgTable), Result,
        !LT, !Env, !IO),
    (
        Result = succeeded(no)
    ;
        Result = succeeded(yes),
        prompts(CompiledFilename, ProgTable, !LT, !Env, !IO)
    ;
        Result = exception(E),
        ( univ_to_type(E, PE : context.error) ->
            print_error(PE, !IO),
            % Ask again, since this is a user-error
            prompts(CompiledFilename, ProgTable, !LT, !Env, !IO)
        ;
            throw(E)
        )
    ).

do_user_command(CommandText, ProgTable, !IO) :-
    % Call do_user_command with "" as the compiled filename. This is currently
    % OK, because CompiledFilename is not used within do_user_command.
    do_user_command(CommandText, "", ProgTable, !IO).

% Note: CompiledFilename is vestigial - it was previously used when the model
% was to compile, then have "interactive" call compiled code in the given
% file. Our current model is to interpret directly, so this argument is
% ignored, and may be "". However, it is retained in case we return to the
% original model in the future.
:- pred do_user_command(string::in, string::in, tables.progtable::in,
                        io::di, io::uo) is cc_multi.
do_user_command(CommandText, CompiledFilename, ProgTable, !IO) :-
    % Compute a one-time env from ProgTable, and a local table
    Env = interpret.progtable_env(ProgTable),
    LT = tables.empty_localtable,
    % Throw the resulting environment and local table away
    util.try_io_thread2(handle_command(string.to_char_list(CommandText),
                            CompiledFilename, ProgTable), Result,
                            LT, _NewLT, Env, _NewEnv, !IO),
    (
        Result = succeeded(_)
    ;
        Result = exception(E),
        ( univ_to_type(E, PE : context.error) ->
            print_error(PE, !IO)
        ;
            throw(E)
        )
    ).

% prompt performs a single prompt.
% Throws context.error exceptions on program or user errors.
% Returns no on EOF, yes otherwise.
:- pred prompt(string::in, tables.progtable::in, bool::out,
                tables.localtable::in, tables.localtable::out,
                interpret.env::in, interpret.env::out,
                io::di, io::uo) is cc_multi.
prompt(CompiledFilename, ProgTable, Continue, !LT, !Env, !IO) :-
    % Ask the user for an expression, and we'll print the type
    io.write_string("?> ", !IO),
    io.read_line(Res, !IO),
    (
        Res = ok(Text),
        Continue = yes,
        handle_command(Text, CompiledFilename, ProgTable, !LT, !Env, !IO)
    ;
        Res = eof,
        Continue = no,
        io.nl(!IO)
    ;
        Res = error(Err),
        Continue = yes,
        io.write_string(io.error_message(Err), !IO)
    ).

% handle_command handles the command typed into a single prompt.
% Throws context.error exceptions on program or user errors.
% Returns no on EOF, yes otherwise.
:- pred handle_command(list(char)::in, string::in, tables.progtable::in,
                        bool::out,
                        tables.localtable::in, tables.localtable::out,
                        interpret.env::in, interpret.env::out,
                        io::di, io::uo) is cc_multi.
handle_command(Text, CompiledFilename, ProgTable, yes, !LT, !Env, !IO) :-
    handle_command(Text, CompiledFilename, ProgTable, !LT, !Env, !IO).
:- pred handle_command(list(char)::in, string::in, tables.progtable::in,
                        tables.localtable::in, tables.localtable::out,
                        interpret.env::in, interpret.env::out,
                        io::di, io::uo) is cc_multi.
handle_command(Text, _CompiledFilename, ProgTable, !LT, !Env, !IO) :-
    Context = context.context("", 1, 1, stack.init),
    % Check for ":" commands
    ( Text = [':', '?'] ->
        interact_help(!IO)
    ; Text = [':', '?', '\n'] ->
        interact_help(!IO)
    ; Text = [':','t','a'] ->
        % All globals
        all_types(ProgTable, !.LT, !IO)
    ; Text = [':','t','a','\n'] ->
        % All globals
        all_types(ProgTable, !.LT, !IO)
    ; Text = [':','t'|Text1] ->
        % Type of an expression
        token.tokenize(listchar_lstrip(Text1), Tokens),
        parsem.parse_expr(Tokens, UserExpr),
        % Get the type of this expression
        typecheck.expr_type(Context, ProgTable, !.LT,
            varset.init, Varset, UserExpr, _, Type),
        io.write_string(pretty.string_expr(UserExpr), !IO),
        io.write_string(" :: ", !IO),
        io.write_string(pretty.string_typeval(Varset, Type), !IO),
        io.nl(!IO)
    ; Text = [':','k'|Text1] ->
        % Kind of a type
        token.tokenize(listchar_lstrip(Text1), Tokens),
        parsem.parse_type(Tokens, Varset, UserType),
        % Get the kind of this expression
        typecheck.type_kind(Context, ProgTable, Varset, UserType, Kind),
        io.write_string(pretty.string_typeval(Varset, UserType), !IO),
        io.write_string(" :: ", !IO),
        io.write_string(pretty.string_kind(Kind), !IO),
        io.nl(!IO)
    ; Text = [':','b'] ->
        % All globals
        browse(ProgTable, !IO)
    ; Text = [':','b','\n'] ->
        % All globals
        browse(ProgTable, !IO)
    ; Text = [':','e'] ->
        % Print environment
        interpret.print_env(!.Env, !IO)
    ; Text = [':','e','\n'] ->
        % Print environment
        interpret.print_env(!.Env, !IO)
    ; Text = [':'|_] ->
        % Error
        io.write_string("Unrecognised command. Type \":?\" for help.\n", !IO)
    ; string.strip(string.from_char_list(Text)) = "" ->
        % Empty line -- do nothing
        true
    ;
        % Typecheck, then evaluate a statement
        token.tokenize(Text, Tokens),
        parsem.parse_stmt(Tokens, UserStmt),
        (
            UserStmt = basic_stmt(BasicStmt, _),
            % Typecheck statement, and update locals if a new variable is
            % being assigned to
            assignment_update_locals(BasicStmt, Context, ProgTable, !LT),
            exec_stmt_internal(BasicStmt, !Env, !IO)
        ;
            UserStmt = compound_stmt(_, _),
            io.write_string(
                "Compound statements not allowed in interactive mode\n", !IO)
        )
    ).

:- pred interact_help(io::di, io::uo) is det.
interact_help(!IO) :-
    io.write_string("Interactive Commands\n\
:b          Browse - Print all type and function definitions\n\
:e          Environment - Print the current value of all variables\n\
:t <expr>   Type - Print the type of an expression\n\
:ta         All types - Print the type of all global variables\n\
:k <type>   Kind - Print the kind of a type expression\n\
:?          Help - This help message\n\
<expr>      Evaluate an expression, and print out the result\n\
<statement> Execute a basic statement\n\
", !IO).

% basic_stmt_update_locals(BasicStmt, ProgTable, !LocalTable).
% If BasicStmt is an assignment statement, typechecks the expression and
% updates locals with a new variable being assigned to.
% Otherwise, typechecks the statement itself, and does not modify the local
% table.
:- pred assignment_update_locals(basic_stmt::in, context.context::in,
    tables.progtable::in, tables.localtable::in, tables.localtable::out)
    is det.
assignment_update_locals(BasicStmt, Ctx, ProgTable, !LT) :-
    ( BasicStmt = assign(Target, Expr) ->
        typecheck.expr_type(Ctx, ProgTable, !.LT, varset.init, _Varset,
        Expr, _NewExpr, Type),
        tables.update_localtable(Target, Type, !LT)
    ;
        typecheck.check_basic_stmt(Ctx, ProgTable, !.LT,
            varset.init, _Varset, BasicStmt, _)
    ).

% exec_stmt_internal(BasicStmt, !Env, !IO).
% Executes a basic statement, using the internal interpreter, and prints the
% result to stdout (with a newline) if it was an expression statement.
% Expects Stmt is type-checked and valid for this program.
:- pred exec_stmt_internal(basic_stmt::in,
            interpret.env::in, interpret.env::out, io::di, io::uo)
            is cc_multi.
exec_stmt_internal(Stmt, !Env, !IO) :-
    try_io_thread1(exec_stmt_internal_(Stmt), TryResult, !Env, !IO),
    (
        TryResult = succeeded(MResult),
        (
            MResult = yes(Result),
            interpret.show_val(Result, ResultStr, !IO),
            io.write_string(ResultStr, !IO),
            io.nl(!IO)
        ;
            MResult = no
        )
    ;
        TryResult = exception(E),
        ( univ_to_type(E, mars_error(Msg)) ->
            io.write_string(io.stderr_stream, "Runtime Error: ", !IO),
            io.write_string(io.stderr_stream, Msg, !IO),
            io.nl(io.stderr_stream, !IO)
        ;
            throw(E)
        )
    ).

% exec_stmt_internal_(BasicStmt, Result, !Env, !IO).
% Executes a basic statement, using the internal interpreter, and returns the
% result if it was an expression statement.
:- pred exec_stmt_internal_(basic_stmt::in, maybe(interpret.value)::out,
    interpret.env::in, interpret.env::out, io::di, io::uo) is det.
exec_stmt_internal_(Stmt, Result, !Env, !IO) :-
    (Stmt = eval(Expr) ->
        % Special case - print out the expression's value
        ast_cfg.expr_to_instrs(Expr, VarName, Instrs),
        foldl2(interpret.exec_instr, Instrs, !Env, !IO),
        % Get the value from the temporary
        interpret.read_var(!.Env, VarName, Val, !IO),
        Result = yes(Val)
    ;
        % Convert the statement into executable (low-level) instructions
        ast_cfg.basic_stmt_to_instrs(Stmt, Instrs),
        foldl2(interpret.exec_instr, Instrs, !Env, !IO),
        Result = no
    ).

% all_types(ProgTable, LocalTable, !IO)
% Implements the :ta command.
% Prints out all variables (local, global and ctors) and their types.
:- pred all_types(tables.progtable::in, tables.localtable::in, io::di,
    io::uo) is det.
all_types(ProgTable, LocalTable, !IO) :-
    % First, build a map from name strings to type strings
    some [!TypeMap] (
        map.init(!:TypeMap),
        % Add the global functions
        map.foldl((pred(Name::in, Function::in, M0::in, M::out) is det :-
            function_type(Function, Varset, Type),
            map.set(M0, Name, pretty.string_typeval(Varset, Type), M)),
            ProgTable ^ prog_funcs, !TypeMap),
        % Add the ctors
        map.foldl((pred(Name::in, TypeDef::in, M0::in, M::out) is det :-
            ( ctor_type(TypeDef, Name, Varset, Type) ->
                map.set(M0, Name, pretty.string_typeval(Varset, Type), M)
            ;
                throw("browse: ctor_type failed on known ctor")
            )),
            ProgTable ^ prog_ctors, !TypeMap),
        % Add the local variables (shadow any globals)
        map.foldl((pred(Name::in, Type::in, M0::in, M::out) is det :-
            map.set(M0, ir.varname_to_string_noescape(Name),
                pretty.string_typeval(varset.init, Type), M)),
            LocalTable, !TypeMap),
        % Print out all the variables
        map.foldl((pred(Name::in, Type::in, I::di, O::uo) is det :-
            io.write_string(Name ++ " :: " ++ Type ++ "\n", I, O)),
            !.TypeMap, !IO)
    ).

% browse(ProgTable, LocalTable, !IO)
% Implements the :b command.
% Prints out all global functions and types in header form
% (Gives a bit more information than :ta, because it prints them flat, before
% transforming them to singleton types).
:- pred browse(tables.progtable::in, io::di, io::uo) is det.
browse(ProgTable, !IO) :-
    % Print the types
    io.write_string("Types:\n", !IO),
    map.foldl((pred(_Name::in, TypeDef-_::in, IO0::di, IO::uo) is det :-
        print_typedef(0, TypeDef, IO0, IO)),
        ProgTable ^ prog_types, !IO),
    % Print the global functions
    io.write_string("Functions:\n", !IO),
    map.foldl((pred(_Name::in, Function::in, IO0::di, IO::uo) is det :-
        pretty.print_function_head(Function, IO0, IO1),
        io.nl(IO1, IO)), ProgTable ^ prog_funcs, !IO).

:- func listchar_lstrip(list(char)) = list(char).
listchar_lstrip([]) = [].
listchar_lstrip(Chars@[C|Cs]) = Done :-
    ( char.is_whitespace(C) ->
        Done = listchar_lstrip(Cs)
    ;
        Done = Chars
    ).
