% -------------------------------------------------------------------------- %
% Mars/Venus Implementation
% Copyright (C) 2008 Matt Giuca
% 14/10/2008
%
% Interpreter
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %
% Directly interprets Mars code in CFG representation.
% -------------------------------------------------------------------------- %

:- module interpret.

:- interface.

:- import_module ir.
:- import_module tables.

:- import_module integer.
:- import_module bool.
:- import_module unit.
:- import_module list.
:- import_module vectorref.
:- import_module io.

% A value (a first-class "thing" at runtime).
:- type value
    --->    val_int(integer)
    ;       val_term(string, list(value))
    ;       val_array(vectorref(value, io))
    ;       val_func(runtime_func).

% A "function object" at runtime.
% Currying is implicit. Function objects accept exactly one argument.
% Note that the function is self-sufficient (carries a program reference), so
% can be evaluated without requiring a context. This may change ...
% (For one thing, it means functions aren't sensitive to changes to the
% program after they are created, which may be a good thing, but may tie up
% memory).
:- type runtime_func.

% Holds the state of a program. This is a mapping from varnames to values for
% locals (in a particular context), globals and builtins.
% Note that the "local" state may just be the variables local to the
% interactive environment.
:- type env.

:- pred print_env(env::in, io::di, io::uo) is det.

% Read a variable from the environment. Throw an exception if not found.
% Threads IO (computable global constants can have side-effects).
% Throws an interpreter error if the variable is not in the local or global
% namespace.
:- pred read_var(env::in, varname::in, value::out, io::di, io::uo) is det.

% env_assign_local(VarName, Value, !Env).
% Updates or sets the value of a local variable in an environment.
:- pred env_assign_local(varname::in, value::in, env::in, env::out) is det.

% progtable_env(ProgTable) builds an environment from the global definitions
% in the ProgTable, as well as the builtins, with an empty local state.
% An env is required to execute code.
% This transforms the static definitions into runtime values, but does not
% directly execute any code (nullary functions will be transformed into
% thunks, NOT immediately evaluated). So this is guaranteed to terminate.
:- func progtable_env(tables.progtable) = env.

% Converts a value to it's canonical Mars literal representation (as a
% string). Note that some values (such as functions and thunks) can't be shown
% as literals; these are described inside angle brackets.
% Note: IO is threaded so this has access to the store (needed to read mutable
% variables).
:- pred show_val(value::in, string::out, io::di, io::uo) is det.

% Succeeds if two values are equal, by the Mars definition of equality.
% This means they have the same type, and does a deep equality on arrays or
% terms.
% Functions raise an error if compared for equality.
% Threads IO and returns a bool, because it needs to look in the mutable
% state.
:- pred eq_vals(value::in, value::in, bool::out, io::di, io::uo) is det.

% Succeeds if two values are aliases (semantically, that they are equal, and
% that modifying one will cause the other to be updated to the new value).
% For Ints, this is the same as standard equality.
% Functions raise an error if compared for reference equality.
:- pred ref_eq_vals(value::in, value::in, bool::out, io::di, io::uo) is det.

% Executes an instruction, updating the environment as a result.
% Also threads IO, because computations (in Mars) may modify state or have
% other side-effects.
:- pred exec_instr(instr::in, env::in, env::out, io::di, io::uo) is det.
% Version of exec_instr which returns unit, for use with util.try_io_thread1.
:- pred exec_instr(instr::in, unit::out, env::in, env::out, io::di, io::uo)
    is det.

:- type mars_error ---> mars_error(string).

% -------------------------------------------------------------------------- %
:- implementation.

:- import_module cfg.
:- import_module context.
:- import_module util.

:- import_module int.
:- import_module string.
:- import_module char.
:- import_module maybe.
:- import_module map.
:- import_module svmap.
:- import_module pair.
:- import_module array.
:- import_module exception.
:- import_module store.

:- type runtime_func
    --->    runtime_func(
                list(value),        % Curried arguments (so far)
                bool,               % True iff thunk (should auto-eval)
                runtime_func_data   % More info, specific to this type
            ).

:- type runtime_func_data
    % Will call a user-defined function when enough arguments are passed.
    --->    func_user(
                tables.progtable,
                function
            )
    % Will become a term when enough arguments are passed.
    ;       func_ctor(
                string              % Name
            )
    % Will call a built-in function when enough arguments are passed.
    ;       func_builtin(
                string,             % Name
                builtin_pred_type
            ).

% Note: Built-ins may have side-effects (and the mode system prevents us from
% threading IO through these). So all built-ins are declared impure, and may
% do whatever side-effects they want.
:- type builtin_pred_type == (impure func(list(value)) = value).

:- type varmap == map.map(varname, value).

:- type env
    --->    env(
                % Built-ins, functions and constructors
                prog_globals    :: varmap,
                % Current local table (for locals in a function, or the
                % interactive environment)
                prog_locals     :: varmap
            ).

print_env(env(Globals, Locals), !IO) :-
    io.write_string("Globals:\n", !IO),
    map.foldl(print_var, Globals, !IO),
    io.write_string("Locals:\n", !IO),
    map.foldl(print_var, Locals, !IO).

:- pred print_var(varname::in, value::in, io::di, io::uo) is det.
print_var(Name, Val, !IO) :-
    io.write_string("    ", !IO),
    io.write_string(ir.varname_to_string_noescape(Name), !IO),
    io.write_string(" = ", !IO),
    show_val(Val, ValStr, !IO),
    io.write_string(ValStr, !IO),
    io.nl(!IO).

% env_search(Env, VarName, Value)
% Searches an environment for a variable of the given name, and returns its
% value.
% Fails if the variable is not in the local or global namespace.
:- pred env_search(env::in, varname::in, value::out) is semidet.
env_search(env(Globals, Locals), VarName, Value) :-
    ( varmap_search(Locals, VarName, V) ->
        Value = V
    ; varmap_search(Globals, VarName, V) ->
        Value = V
    ;
        fail
    ).

env_assign_local(VarName, Value, env(Globals, Locals0),
    env(Globals, Locals)) :-
    varmap_set(VarName, Value, Locals0, Locals).

progtable_env(ProgTable) = env(progtable_varmap(ProgTable), map.init).

% Transform a ProgTable (and builtins) into a varmap.
:- func progtable_varmap(tables.progtable) = varmap.

progtable_varmap(ProgTable) = Map :-
    some [!Map] (
        map.init(!:Map),
        map.foldl(func_varmap(ProgTable), ProgTable ^ prog_funcs, !Map),
        map.foldl(ctor_varmap(ProgTable), ProgTable ^ prog_ctors, !Map),
        Map = !.Map
    ).

:- pred func_varmap(tables.progtable::in, string::in, function::in,
    varmap::in, varmap::out) is det.
func_varmap(ProgTable, Name, Func, !Map) :-
    ( Func ^ func_body = func_builtin ->
        % The value must be derived from the builtins
        ( map.search(all_builtins, Name, BFunc) ->
            RFunc = BFunc,
            svmap.det_insert(svname(Name), val_func(RFunc), !Map)    % TEMP
        ;
            throw("Error: Built-in function \"" ++ Name
                ++ "\" not defined by interpreter")
        )
    ;
        Params = Func ^ func_params,
        (
            Params = no,
            IsThunk = yes
        ;
            Params = yes(_),
            IsThunk = no
        ),
        RFunc = runtime_func([], IsThunk, func_user(ProgTable, Func)),
        svmap.det_insert(svname(Name), val_func(RFunc), !Map)       % TEMP
    ).
    %svmap.det_insert(svname(Name), val_func(RFunc), !Map).

:- pred ctor_varmap(tables.progtable::in, string::in, typedef::in,
    varmap::in, varmap::out) is det.
ctor_varmap(ProgTable, Name, _TypeDef, !Map) :-
    ( tables.lookup_ctor(ProgTable, Name, _TypeDefAgain, Ctor) ->
        Params = Ctor ^ ctor_args,
        (
            Params = no,
            % Param-less constructor. Replace it with its term immediately.
            done_ctor(Name, [], Val)
        ;
            Params = yes(_),
            RFunc = runtime_func([], no, func_ctor(Name)),
            Val = val_func(RFunc)
        ),
        svmap.det_insert(svname(Name), Val, !Map)
    ;
        throw("Error: Constructor lookup failed; should not have")
    ).

:- func varmap_init = varmap.
varmap_init = map.init.

:- pred varmap_search(varmap::in, varname::in, value::out) is semidet.
varmap_search(Map, Name, Value) :- map.search(Map, Name, Value).

:- pred varmap_set(varname::in, value::in, varmap::in, varmap::out) is det.
varmap_set(Name, Value, !VarMap) :- map.set(!.VarMap, Name, Value, !:VarMap).

% show_arg_list(Values).
% Returns "" if Values is [], otherwise a comma-separated list of values,
% surrounded by ().
:- pred show_arg_list(list(value)::in, string::out, io::di, io::uo) is det.
show_arg_list(Args, S, !IO) :-
    (
        Args = [],
        S = ""
    ;
        Args = [_|_],
        show_values(Args, ArgsString, !IO),
        S = "(" ++ ArgsString ++ ")"
    ).

% show_values(Values).
% Returns a string with a comma-separated list of values.
:- pred show_values(list(value)::in, string::out, io::di, io::uo) is det.
show_values(Values, S, !IO) :-
    util.list_foldr2((pred(Arg::in, L::in, [ArgStr|L]::out, I::di, O::uo)
        is det :- show_val(Arg, ArgStr, I, O)),
        Values, [], ValueStrings, !IO),
    S = string.join_list(", ", ValueStrings).

show_val(val_int(I), integer.to_string(I), !IO).
show_val(val_term(Ctor, Args), Ctor ++ ArgList, !IO) :-
    show_arg_list(Args, ArgList, !IO).
show_val(val_array(VectorRef), "[" ++ ValsString ++ "]", !IO) :-
    % Convert the vector to a list of values
    vectorref.extract(VectorRef, Array, !IO),
    array.to_list(Array, List),
    show_values(List, ValsString, !IO).
show_val(val_func(Func), "<" ++ IsCurried ++ WhatIs ++ More ++ ">", !IO) :-
    Func = runtime_func(Args, IsThunk, FuncData),
    ( IsThunk = yes ->
        (
            FuncData = func_user(_, F),
            Name = F ^ func_name
        ;
            FuncData = func_ctor(Name)
        ;
            FuncData = func_builtin(Name, _)
        ),
        WhatIs = "thunk " ++ Name
    ;
        (
            FuncData = func_user(_, F),
            WhatIs = "function " ++ F ^ func_name
        ;
            FuncData = func_ctor(Name),
            WhatIs = "constructor function " ++ Name
        ;
            FuncData = func_builtin(Name, _),
            WhatIs = "built-in function " ++ Name
        )
    ),
    (
        Args = [],
        IsCurried = ""
    ;
        Args = [_|_],
        IsCurried = "curried "
    ),
    show_arg_list(Args, More, !IO).

eq_vals(val_int(X), Y, Eq, !IO) :-
    ( Y = val_int(Yi), X = Yi ->
        Eq = yes
    ;
        Eq = no
    ).
eq_vals(val_term(Ctor1, Args1), Term2, Eq, !IO) :-
    ( Term2 = val_term(Ctor2, Args2), Ctor1 = Ctor2 ->
        eq_val_lists(Args1, Args2, Eq, !IO)
    ;
        Eq = no
    ).
eq_vals(val_array(V1), Term2, Eq, !IO) :-
    ( Term2 = val_array(V2) ->
        % Check if the size is equal
        vectorref.size(V1, S1, !IO),
        vectorref.size(V2, S2, !IO),
        ( S1 = S2 ->
            % Convert the vectors to lists of values
            vectorref.extract(V1, A1, !IO),
            array.to_list(A1, L1),
            vectorref.extract(V2, A2, !IO),
            array.to_list(A2, L2),
            % Guaranteed same size lists
            % Apply eq_vals to each corresponding pair of values
            list.map_corresponding_foldl(eq_vals, L1, L2, Eqs, !IO),
            % Eqs : list(bool). True if all yes.
            ( list.all_true(pred(yes::in) is semidet, Eqs) ->
                Eq = yes
            ;
                Eq = no
            )
        ;
            Eq = no
        )
    ;
        Eq = no
    ).
eq_vals(val_func(_), _, _, !IO) :-
    throw_mars_error("Functions cannot be compared for equality").

:- pred eq_val_lists(list(value)::in, list(value)::in, bool::out,
    io::di, io::uo) is det.
eq_val_lists([], [], yes, !IO).
eq_val_lists([], [_|_], no, !IO).
eq_val_lists([_|_], [], no, !IO).
eq_val_lists([X|Xs], [Y|Ys], Eq, !IO) :-
    eq_vals(X, Y, Eq0, !IO),
    (
        Eq0 = no,
        Eq = no
    ;
        Eq0 = yes,
        eq_val_lists(Xs, Ys, Eq, !IO)
    ).

ref_eq_vals(val_int(X), Y, Eq, !IO) :-
    ( Y = val_int(Yi), X = Yi ->
        Eq = yes
    ;
        Eq = no
    ).
ref_eq_vals(val_term(Ctor1, _Args1), Term2, Eq, !IO) :-
    ( Term2 = val_term(Ctor2, _Args2), Ctor1 = Ctor2 ->
        % XXX: Reference equality on terms (doesn't yet work because terms
        % aren't mutable). For now, consider all terms to be unaliased.
        Eq = no
    ;
        Eq = no
    ).
ref_eq_vals(val_array(V1), Term2, Eq, !IO) :-
    ( Term2 = val_array(V2), V1 = V2 ->
        Eq = yes
    ;
        Eq = no
    ).
ref_eq_vals(val_func(_), _, _, !IO) :-
    throw_mars_error("Functions cannot be compared for reference equality").

read_var(Env, Name, Result, !IO) :-
    % Replace gvname with svname (global vars are stored as svname inside the
    % interpreter).
    ( Name = gvname(N) ->
        Name1 = svname(N)
    ;
        Name1 = Name
    ),
    ( env_search(Env, Name1, Result0) ->
        eval_thunk(Env, Result0, Result, !IO)
    ;
        throw("Interpreter Error: Undefined variable " ++
            ir.varname_to_string(Name1))
    ).

% read_var_into_vector(Env, VectorRef, VarName, !IO).
% Reads a variable, and appends its value onto the end of the vector.
% Uses IO as the vector's store.
:- pred read_var_into_vector(env::in, vectorref(value, io)::in, varname::in,
    io::di, io::uo) is det.
read_var_into_vector(Env, VectorRef, VarName, !IO) :-
    read_var(Env, VarName, Value, !IO),
    vectorref.add(VectorRef, Value, !IO).

exec_instr(Instr, !Env, !IO) :-
    exec_instr_(Instr ^ instr_instr, Instr ^ instr_context, !Env, !IO).

exec_instr(I, unit, !Env, !IO) :-
    exec_instr(I, !Env, !IO).

:- pred exec_instr_(instr_::in, context::in, env::in, env::out,
    io::di, io::uo) is det.
exec_instr_(nop, _Ctx, !Env, !IO).
exec_instr_(mov(Dst, Src), _Ctx, !Env, !IO) :-
    read_var(!.Env, Src, Result, !IO),
    env_assign_local(Dst, Result, !Env).
exec_instr_(i_fieldset(_Var, _FieldName, _Src), _Ctx, !Env, !IO) :-
    throw("Not Implemented: Field-set instruction").
exec_instr_(i_fieldreplace(_Dst, _Var, _FieldName, _Src), _Ctx, !Env, !IO) :-
    throw("Not Implemented: Field-replace instruction").
exec_instr_(ld_intlit(Dst, I), _Ctx, !Env, !IO) :-
    env_assign_local(Dst, val_int(I), !Env).
exec_instr_(ld_arraylit(Dst, Elems), _Ctx, !Env, !IO) :-
    % Create a new, empty vector of values
    Capacity = good_vector_capacity(list.length(Elems)),
    vectorref.init(Capacity, VectorRef, !IO),
    % Evaluate each expression, populating the vector
    list.foldl(read_var_into_vector(!.Env, VectorRef), Elems, !IO),
    env_assign_local(Dst, val_array(VectorRef), !Env).
exec_instr_(ld_ctor(Dst, Name), _Ctx, !Env, !IO) :-
    ( env_search(!.Env, svname(Name), Result0) ->
        eval_thunk(!.Env, Result0, Result, !IO)
    ;
        throw("Interpreter Error: Undefined constructor " ++ Name)
    ),
    env_assign_local(Dst, Result, !Env).
exec_instr_(ld_field(_Dst, _Val, _FieldName), _Ctx, !Env, !IO) :-
    throw("Not Implemented: Field reference").
exec_instr_(call(Dst, Func, Args), _Ctx, !Env, !IO) :-
    % Evaluate Func and Arg (eagerly)
    read_var(!.Env, Func, FuncVal, !IO),
    list.map_foldl(read_var(!.Env), Args, ArgVals, !IO),
    % Apply Arg to Func
    apply_vals(!.Env, FuncVal, ArgVals, Result, !IO),
    env_assign_local(Dst, Result, !Env).
exec_instr_(parcall(Dst, Func, Args), _Ctx, !Env, !IO) :-
    % Evaluate Func and Arg (eagerly)
    read_var(!.Env, Func, FuncVal, !IO),
    list.map_foldl(read_var(!.Env), Args, ArgVals, !IO),
    % Curry Args in Func, but do not evaluate
    partial_apply_vals(FuncVal, ArgVals, Result),
    env_assign_local(Dst, Result, !Env).

% good_vector_capacity(Min) finds a good initial vector capacity >= Min.
:- func good_vector_capacity(int) = int.
good_vector_capacity(Min) = good_vector_capacity(Min, 8).
:- func good_vector_capacity(int, int) = int.
good_vector_capacity(Min, Try) = X :-
    ( Try >= Min ->
        X = Try
    ;
        X = good_vector_capacity(Min, Try*2)
    ).

% Given a list of formals and a list of actuals (values), appends into a
% varmap.
:- pred bind_args(list({string, A, B})::in, list(value)::in,
    varmap::in, varmap::out) is det.
bind_args([], [], !Map).
bind_args([], [_|_], !Map) :-
    throw("Interpreter Error: More actuals than formals").
bind_args([_|_], [], !Map) :-
    throw("Interpreter Error: More formals than actuals").
bind_args([{F,_,_}|Fs], [A|As], !Map) :-
    svmap.set(svname(F), A, !Map),
    bind_args(Fs, As, !Map).

% Evaluate a function (with all arguments).
:- pred done_func_data(env::in, runtime_func_data::in, list(value)::in,
    value::out, io::di, io::uo) is det.
done_func_data(env(Globals, _Locals), func_user(Prog, Func), Args, Result,
    !IO) :-
    Body = Func ^ func_body,
    ( Body = func_body_cfg(_Decls, CFG) ->
        % Create a new local varmap, with the arguments
        Params = Func ^ func_params,
        (
            Params = no,
            InitVarMap = varmap_init
        ;
            Params = yes(YesParams),
            bind_args(YesParams, Args, varmap_init, InitVarMap)
        ),
        exec_cfg(CFG, Prog, env(Globals, InitVarMap), env(_, FinalVarMap),
            !IO),
        % Now FinalVarMap[RetVName] should hold the return value
        RetVName = cfg.get_retv(CFG),
        ( varmap_search(FinalVarMap, RetVName, Res) ->
            Result = Res
        ;
            % retvame was not set (code error)
            throw("Function did not assign a return value")
        )
    ;
        throw("Function not represented as CFG")
    ).
done_func_data(_Env, func_ctor(Name), Args, Result, !IO) :-
    done_ctor(Name, Args, Result).
done_func_data(_Env, func_builtin(_Name, Func), Args, Result, !IO) :-
    % Note: Function may cause side-effects (hence it is labeled "impure").
    % The goal itself is promised pure, which is legitimate because we are
    % threading IO through here.

    % equivalent of: Result = Func(Args)
    promise_pure impure Result = impure_apply(Func, Args).

% Evaluate a ctor (with all arguments). First argument is the name.
:- pred done_ctor(string::in, list(value)::in, value::out) is det.
done_ctor(Name, Args, val_term(Name, Args)).

% Given a value, check if it is a thunk which needs to be evaluated.
% Return the value, evaluated if necessary.
:- pred eval_thunk(env::in, value::in, value::out, io::di, io::uo) is det.
eval_thunk(Env, Val0, Val, !IO) :-
    % If a thunk, evaluate
    ( Val0 = val_func(runtime_func(Args, yes, FuncData)) ->
        done_func_data(Env, FuncData, Args, Val, !IO)
    ;
        Val = Val0
    ).

:- pred apply_func(env::in, runtime_func::in, list(value)::in, value::out,
    io::di, io::uo) is det.
apply_func(Env, runtime_func(Args, _Thk, FuncData), NewArgs, Result, !IO) :-
    % Curry the remaining arguments, then immediately evaluate the function,
    % and return its result.
    done_func_data(Env, FuncData, Args ++ NewArgs, Result, !IO).

:- pred apply_vals(env::in, value::in, list(value)::in, value::out,
    io::di, io::uo) is det.
apply_vals(Env, Val, Args, Result, !IO) :-
    ( Val = val_func(Func) ->
        apply_func(Env, Func, Args, Result, !IO)
    ;
        throw("Interpreter Error: Apply to something not a function")
    ).

:- pred partial_apply_func(runtime_func::in, list(value)::in, value::out)
    is det.
partial_apply_func(runtime_func(Args, IsThunk, FuncData), NewArgs, Result) :-
    % Curry and return a new closure object.
    Result = val_func(runtime_func(Args ++ NewArgs, IsThunk, FuncData)).

:- pred partial_apply_vals(value::in, list(value)::in, value::out) is det.
partial_apply_vals(Val, Args, Result) :-
    ( Val = val_func(Func) ->
        partial_apply_func(Func, Args, Result)
    ;
        throw("Interpreter Error: Partial apply to something not a function")
    ).

% -------------------------------------------------------------------------- %
% CFG EXECUTION
% -------------------------------------------------------------------------- %

% Executes a CFG (begins at entry, runs to exit).
% Updates varmap with the new state of the locals, including retvname (the
% return value).
:- pred exec_cfg(cfg(S)::in, progtable::in, env::in, env::out, io::di, io::uo)
    is det.
exec_cfg(CFG, _Prog, !Env, !IO) :-
    % Execute the body of the first block
    Entry = cfg.get_entry(CFG),
    exec_block_body(CFG, Entry, !Env, !IO).

% exec_block(CFG, Predecessor, Target, !Env, !IO).
% Executes block Target, coming from block Pred. (Pred is used to determine
% phis).
:- pred exec_block(cfg(S)::in, bbref(S)::in, bbref(S)::in, env::in, env::out,
    io::di, io::uo) is det.
exec_block(CFG, Pred, Target, !Env, !IO) :-
    get_basic_block(Target, TargetBlock, CFG),
    Phis = TargetBlock ^ bb_phis,
    % For each phi, make a binding in the environment
    list.foldl(bind_phi_env(Pred), Phis, !Env),
    exec_block_body(CFG, Target, !Env, !IO).

% bind_phi_env(Predecessor, Phi, !Env).
% Given a predecessor block reference and a phi, make a new variable in the
% environment by copying from an existing variable, based on the predecessor.
:- pred bind_phi_env(bbref(S)::in, phi(S)::in, env::in, env::out) is det.
bind_phi_env(Pred, phi(NewVarName, Map), !Env) :-
    % Get the varname corresponding to the block we just came from
    ( map.search(Map, Pred, FromVarName) ->
        ( env_search(!.Env, FromVarName, Value) ->
            env_assign_local(NewVarName, Value, !Env)
        ;
            throw("Interpreter Error: Undefined variable " ++
                ir.varname_to_string(FromVarName))
        )
    ;
        throw("Interpreter Error: Phi missing prececessor in CFG")
    ).

% Executes the body of a basic block, and executes the terminator, branching
% to the following block.
% Updates the environment and causes IO side-effects.
:- pred exec_block_body(cfg(S)::in, bbref(S)::in, env::in, env::out,
    io::di, io::uo) is det.
exec_block_body(CFG, BBRef, !Env, !IO) :-
    get_basic_block(BBRef, Block, CFG),
    Body = Block ^ bb_body,
    % Execute backwards (remember the list is in reverse, for efficient
    % construction).
    util.list_foldr2(exec_instr, Body, !Env, !IO),
    % Execute the terminator
    Term = Block ^ bb_terminator,
    (
        Term = t_switch(Control, Cases),
        read_var(!.Env, Control, ControlVal, !IO),
        exec_cases(CFG, BBRef, !Env, ControlVal, Cases, !IO)
    ;
        Term = cond_branch(Cond, Then, Else),
        read_var(!.Env, Cond, CondVal, !IO),
        ( CondVal = val_int(integer(0)) ->
            % False
            exec_block(CFG, BBRef, Else, !Env, !IO)
        ; CondVal = val_int(_) ->
            % True
            exec_block(CFG, BBRef, Then, !Env, !IO)
        ;
            % Not an int ... should have been a type error
            throw("Interpreter Error: Non-int in conditional branch")
        )
    ;
        Term = branch(Target),
        exec_block(CFG, BBRef, Target, !Env, !IO)
    ;
        Term = null_terminator
        % Done - do nothing (and execution will fall out of the CFG)
    ).

% exec_cases(CFG, Predecessor, !Env, ControlValue, Cases, !IO).
% Performs a switch over Cases, and executes the first one which matches.
:- pred exec_cases(cfg(S)::in, bbref(S)::in, env::in, env::out,
    value::in, list(cfg_case_stmt(S))::in, io::di, io::uo) is det.
exec_cases(_CFG, _Predecessor, !Env, _ControlValue, [], !IO) :-
    % No pattern matched
    % XXX Should be checked statically so this never occurs, and therefore, we
    % raise an interpreter error at this point.
    % Currently, no check is made, and ast_cfg goes nuts if a switch
    % statement isn't handled by default. Therefore, we need to raise a
    % friendly runtime error for the time being.
    throw_mars_error("Pattern match failure").
    %throw("Interpreter Error: Pattern match failure").
exec_cases(CFG, Predecessor, !Env, ControlValue, [Case|Cases], !IO) :-
    Case = cfg_case_stmt(Pattern, Target, _Ctx),
    ( pattern_match(Pattern, ControlValue, !Env) ->
        % Succeeded. Branch to Target (with the new bindings) and ignore the
        % remaining Cases.
        exec_block(CFG, Predecessor, Target, !Env, !IO)
    ;
        % That match failed. Try the rest.
        exec_cases(CFG, Predecessor, !Env, ControlValue, Cases, !IO)
    ).

% pattern_match(Pattern, Value, !Env).
% Matches Value against Pattern, making bindings in !Env if successful.
% Fails if the pattern match fails.
:- pred pattern_match(pattern::in, value::in, env::in, env::out) is semidet.
pattern_match(pattern(Pattern, _PatType), Value, !Env) :-
    (
        Pattern = pat_any
        % Done, succeed with no bindings
    ;
        Pattern = pat_var(VarName),
        % Succeed, and bind
        env_assign_local(VarName, Value, !Env)
    ;
        Pattern = pat_ctor(CtorName, SubPatterns),
        Value = val_term(CtorName, SubValues),              % Can fail
        % Recursively match all sub-terms with sub-patterns
        (
            SubPatterns = yes(YesSubPatterns),
            pattern_match_list(YesSubPatterns, SubValues, !Env)  % Can fail
        ;
            SubPatterns = no
        )
    ;
        Pattern = pat_intlit(Int),
        % Succeed if the value is a matching int
        Value = val_int(Int)    % Can fail
    ).

% pattern_match_list(Patterns, Values, !Env).
% Matches Values against Patterns, making bindings in !Env if successful.
% Succeeds only if each value matches its corresponding pattern, and the lists
% are exactly the same length.
:- pred pattern_match_list(list(pattern)::in, list(value)::in,
    env::in, env::out) is semidet.
pattern_match_list([], [], !Env).
pattern_match_list([P|Ps], [V|Vs], !Env) :-
    pattern_match(P, V, !Env),          % Can fail
    pattern_match_list(Ps, Vs, !Env).   % Can fail

% -------------------------------------------------------------------------- %
% BUILTINS
% -------------------------------------------------------------------------- %

% make_builtin(Name, NumArgs, Pred)
% Returns a runtime_func object describing a built-in function.
% Note: Does not allow thunk (parameter-less) built-ins. If any are required,
% this should be fixed, but none presently are.
:- func make_builtin(string, int, builtin_pred_type) = runtime_func.
make_builtin(Name, _NumArgs, Pred)
    = runtime_func([], no, func_builtin(Name, Pred)).

% throw_invalid throws an "Invalid arguments" error.
:- pred throw_invalid is erroneous.
throw_invalid :-
    throw("Invalid arguments").

:- pred throw_mars_error(string::in) is erroneous.
throw_mars_error(Msg) :-
    throw(mars_error(Msg)).

% integer_int(Integer, ProcName) converts an integer to an int, throwing a
% Mars error on overflow. The error message includes ProcName (the caller).
:- func integer_int(integer, string) = int.
integer_int(Integer, ProcName) = Int :-
    % Use promise_equivalent_solutions to cast cc_multi to det.
    promise_equivalent_solutions [Int] integer_int(Integer, ProcName, Int).
:- pred integer_int(integer::in, string::in, int::out) is cc_multi.
integer_int(Integer, ProcName, Int) :-
    try((pred(I::out) is det :- I = integer.int(Integer)), Res),
    (
        Res = succeeded(Int)
    ;
        Res = exception(_),
        throw_mars_error(ProcName ++ ": Integer value too large")
    ).

% string_to_mars_value(Str, Value, !IO).
% Converts a string into a Mars Array(Int) value.
:- pred string_to_mars_value(string::in, value::out, io::di, io::uo) is det.
string_to_mars_value(Str, Ret, !IO) :-
    string.to_char_list(Str, Chars),
    list.map((pred(Char::in, Val::out) is det :-
        Val = val_int(integer(char.to_int(Char)))),
        Chars, Vals),
    % Create a new, empty vector of values
    Capacity = good_vector_capacity(string.length(Str)),
    vectorref.init(Capacity, VectorRef, !IO),
    Ret = val_array(VectorRef),
    % Populate vector with values
    foldl(vectorref.add(VectorRef), Vals, !IO).

% mars_value_to_string(Value, ProcName, Str, !IO).
% Value must be of type Array(Int).
% Runtime error if any value is out of range.
% Uses ProcName in the error message.
:- pred mars_value_to_string(value::in, string::in, string::out,
    io::di, io::uo) is det.
mars_value_to_string(Value, ProcName, Str, !IO) :-
    ( Value = val_array(VectorRef) ->
        % Pull the int values out of the vector
        vectorref.extract(VectorRef, Array, !IO),
        array.to_list(Array, Vals),
        list.map((pred(Val::in, Char::out) is det :-
            ( Val = val_int(Integer) ->
                ( Integer >= integer(0), Integer < integer(256) ->
                    Char = char.det_from_int(integer.int(Integer))
                ;
                    throw_mars_error(ProcName ++
                        ": Character value out of range [0,255]")
                )
            ;
                throw_invalid
            )),
            Vals, Chars),
        Str = string.from_char_list(Chars)
    ;
        throw_invalid
    ).

% COMPARISON %

% eq(x :: a, y :: a) :: Int
:- func builtin_eq = runtime_func.
builtin_eq = make_builtin("eq", 2, impure func(Args) = Ret :-
    ( Args = [X, Y] ->
        impure invoke_io1(eq_vals(X, Y), Eq),
        (
            Eq = yes,
            R = 1
        ;
            Eq = no,
            R = 0
        ),
        Ret = val_int(integer(R))
    ;
        throw_invalid
    )
).

% is(x :: a, y :: a) :: Int
:- func builtin_is = runtime_func.
builtin_is = make_builtin("is", 2, impure func(Args) = Ret :-
    ( Args = [X, Y] ->
        impure invoke_io1(ref_eq_vals(X, Y), Eq),
        (
            Eq = yes,
            R = 1
        ;
            Eq = no,
            R = 0
        ),
        Ret = val_int(integer(R))
    ;
        throw_invalid
    )
).

% cmp(x :: Int, y :: Int) :: Int
:- func builtin_cmp = runtime_func.
builtin_cmp = make_builtin("cmp", 2, impure func(Args) = Ret :-
    ( Args = [val_int(X), val_int(Y)] ->
        ( X > Y ->
            R = 1
        ; X < Y ->
            R = -1
        ;
            R = 0
        ),
        Ret = val_int(integer(R))
    ;
        throw_invalid
    )
).

% ARITHMETIC %

% add(x :: Int, y :: Int) :: Int
:- func builtin_add = runtime_func.
builtin_add = make_builtin("add", 2, impure func(Args) = Ret :-
    ( Args = [val_int(X), val_int(Y)] ->
        Ret = val_int(X + Y)
    ;
        throw_invalid
    )
).

% sub(x :: Int, y :: Int) :: Int
:- func builtin_sub = runtime_func.
builtin_sub = make_builtin("sub", 2, impure func(Args) = Ret :-
    ( Args = [val_int(X), val_int(Y)] ->
        Ret = val_int(X - Y)
    ;
        throw_invalid
    )
).

% mul(x :: Int, y :: Int) :: Int
:- func builtin_mul = runtime_func.
builtin_mul = make_builtin("mul", 2, impure func(Args) = Ret :-
    ( Args = [val_int(X), val_int(Y)] ->
        Ret = val_int(X * Y)
    ;
        throw_invalid
    )
).

% div(x :: Int, y :: Int) :: Int
:- func builtin_div = runtime_func.
builtin_div = make_builtin("div", 2, impure func(Args) = Ret :-
    ( Args = [val_int(X), val_int(Y)] ->
        ( Y = integer(0) ->
            throw_mars_error("div: Division by zero")
        ;
            Ret = val_int(X div Y)   % Flooring integer division
        )
    ;
        throw_invalid
    )
).

% mod(x :: Int, y :: Int) :: Int
:- func builtin_mod = runtime_func.
builtin_mod = make_builtin("mod", 2, impure func(Args) = Ret :-
    ( Args = [val_int(X), val_int(Y)] ->
        ( Y = integer(0) ->
            throw_mars_error("mod: Division by zero")
        ;
            Ret = val_int(X mod Y)   % Flooring integer mod
        )
    ;
        throw_invalid
    )
).

% ARRAY %

% array(length :: Int, default :: a) :: Array(a)
:- func builtin_array = runtime_func.
builtin_array = make_builtin("array", 2, impure func(Args) = Ret :-
    ( Args = [val_int(LengthInteger), Default] ->
        Length = integer_int(LengthInteger, "array"),
        ( Length < 0 ->
            throw_mars_error("array: Negative size")
        ;
            impure invoke_io1(new_vector(Length, Default), VectorRef),
            Ret = val_array(VectorRef)
        )
    ;
        throw_invalid
    )
).

:- pred new_vector(int::in, T::in, vectorref(T, S)::out, S::di, S::uo)
    is det <= store(S).
new_vector(Length, Default, VectorRef, !S) :-
    % Create a new, empty vector of values
    Capacity = good_vector_capacity(Length),
    vectorref.init(Capacity, VectorRef, !S),
    % Populate the vector with copies of Default
    util.repeat((pred(S0::di, S::uo) is det :-
        vectorref.add(VectorRef, Default, S0, S)),
        Length, !S).

% array_bounds_check(VectorRef, Index, CallingFunction, !S)
% Checks an array access; throws a mars error
% (Nicer for the user than the array.index_out_of_bounds thrown otherwise).
:- pred array_bounds_check(vectorref(T, S)::in, int::in, string::in,
    S::di, S::uo) is det <= store(S).
array_bounds_check(VectorRef, Index, FuncName, !S) :-
    vectorref.size(VectorRef, Length, !S),
    ( (Index < 0 ; Index >= Length) ->
        throw_mars_error(
            string.format("%s: Index %d out of bounds [0,%d]",
                          [s(FuncName), i(Index), i(Length-1)]))
    ;
        true
    ).

% array_ref(array :: Array(a), index :: Int) :: a
:- func builtin_array_ref = runtime_func.
builtin_array_ref = make_builtin("array_ref", 2, impure func(Args) = Ret :-
    ( Args = [val_array(VectorRef), val_int(IndexInteger)] ->
        Index = integer_int(IndexInteger, "array_ref"),
        impure invoke_io(array_bounds_check(VectorRef, Index, "array_ref")),
        impure invoke_io1(vectorref.lookup(VectorRef, Index), Ret)
    ;
        throw_invalid
    )
).

% __impure_array_set(array :: Array(a), index :: Int, value :: a) :: Array(a)
:- func builtin_impure_array_set = runtime_func.
builtin_impure_array_set
    = make_builtin("__impure_array_set", 3, impure func(Args) = Ret :-
    ( Args = [val_array(VectorRef), val_int(IndexInteger), Value] ->
        Index = integer_int(IndexInteger, "array_set"),
        impure invoke_io(array_bounds_check(VectorRef, Index, "array_set")),
        impure invoke_io(vectorref.set(VectorRef, Index, Value)),
        % For now, return the input array reference (may change to void)
        Ret = val_array(VectorRef)
    ;
        throw_invalid
    )
).

% array_replace(array :: Array(a), index :: Int, value :: a) :: Array(a)
:- func builtin_array_replace = runtime_func.
builtin_array_replace = make_builtin("array_replace", 3,
    impure func(Args) = Ret :-
    ( Args = [val_array(VectorRef), val_int(IndexInteger), Value] ->
        Index = integer_int(IndexInteger, "array_replace"),
        impure invoke_io(array_bounds_check(VectorRef, Index,
                                            "array_replace")),
        % Shallow-copy the array
        impure invoke_io1(vectorref.copy(VectorRef), NewVectorRef),
        impure invoke_io(vectorref.set(NewVectorRef, Index, Value)),
        % Return the new array
        Ret = val_array(NewVectorRef)
    ;
        throw_invalid
    )
).

% __impure_array_append(array :: Array(a), value :: a) :: Array(a)
:- func builtin_impure_array_append = runtime_func.
builtin_impure_array_append = make_builtin("__impure_array_append", 2,
    impure func(Args) = Ret :-
    ( Args = [val_array(VectorRef), Value] ->
        impure invoke_io(vectorref.add(VectorRef, Value)),
        % For now, return the input array reference (may change to void)
        Ret = val_array(VectorRef)
    ;
        throw_invalid
    )
).

% __impure_array_extend(array :: Array(a), values :: Array(a)) :: Array(a)
:- func builtin_impure_array_extend = runtime_func.
builtin_impure_array_extend = make_builtin("__impure_array_extend", 2,
    impure func(Args) = Ret :-
    ( Args = [val_array(VectorRef), val_array(Values)] ->
        % Convert the vector to a list of values
        impure invoke_io1(vectorref.extract(Values), ValueArray),
        array.to_list(ValueArray, ValueList),
        impure invoke_io(foldl(vectorref.add(VectorRef), ValueList)),
        % For now, return the input array reference (may change to void)
        Ret = val_array(VectorRef)
    ;
        throw_invalid
    )
).

% array_concat(array1 :: Array(a), array2 :: Array(a)) :: Array(a)
:- func builtin_array_concat = runtime_func.
builtin_array_concat = make_builtin("array_concat", 2,
    impure func(Args) = Ret :-
    ( Args = [val_array(Vector1), val_array(Vector2)] ->
        impure invoke_io1(vectorref.size(Vector1), Length1),
        impure invoke_io1(vectorref.size(Vector2), Length2),
        % Create a new, empty vector of values
        Capacity = good_vector_capacity(Length1 + Length2),
        impure invoke_io1(vectorref.init(Capacity), VectorRef),
        Ret = val_array(VectorRef),
        % Populate vector as with array_extend, but for both arrays
        impure invoke_io1(vectorref.extract(Vector1), Array1),
        array.to_list(Array1, List1),
        impure invoke_io(foldl(vectorref.add(VectorRef), List1)),
        impure invoke_io1(vectorref.extract(Vector2), Array2),
        array.to_list(Array2, List2),
        impure invoke_io(foldl(vectorref.add(VectorRef), List2))
    ;
        throw_invalid
    )
).

% array_length(array :: Array(a)) :: Int
:- func builtin_array_length = runtime_func.
builtin_array_length = make_builtin("array_length", 1,
    impure func(Args) = Ret :-
    ( Args = [val_array(VectorRef)] ->
        impure invoke_io1(vectorref.size(VectorRef), Length),
        Ret = val_int(integer(Length))
    ;
        throw_invalid
    )
).

% IO %

% put_char(c :: Int) :: Int
:- func builtin_put_char = runtime_func.
builtin_put_char = make_builtin("put_char", 1,
    impure func(Args) = Ret :-
    ( Args = [val_int(X)] ->
        ( X >= integer(0), X < integer(256) ->
            Char = char.det_from_int(integer.int(X)),
            impure util.invoke_io(io.write_char(Char)),
            Ret = val_int(integer(0))
        ;
            throw_mars_error("put_char: Argument out of range [0,255]")
        )
    ;
        throw_invalid
    )
).

% get_char() :: Int
:- func builtin_get_char = runtime_func.
builtin_get_char = make_builtin("get_char", 0,
    impure func(Args) = Ret :-
    ( Args = [] ->
        impure util.invoke_io1(io.read_char, IORes),
        (
            IORes = ok(Char),
            R = char.to_int(Char)
        ;
            IORes = eof,
            R = -1
        ;
            IORes = error(IOError),
            Msg = io.error_message(IOError),
            throw_mars_error("get_char: " ++ Msg)
        ),
        Ret = val_int(integer(R))
    ;
        throw_invalid
    )
).

% get_env(name :: Array(Int)) :: Array(Int)
% Returns "" if not found.
:- func builtin_get_env = runtime_func.
builtin_get_env = make_builtin("get_env", 1,
    impure func(Args) = Ret :-
    ( Args = [Name] ->
        impure invoke_io1(mars_value_to_string(Name, "get_env"), NameStr),
        impure invoke_io1(io.get_environment_var(NameStr), MaybeVal),
        (
            MaybeVal = yes(ValStr)
        ;
            MaybeVal = no,
            ValStr = ""
        ),
        impure invoke_io1(string_to_mars_value(ValStr), Ret)
    ;
        throw_invalid
    )
).

% MISC %

% show(x :: a) :: Array(Int)
:- func builtin_show = runtime_func.
builtin_show = make_builtin("show", 1, impure func(Args) = Ret :-
    ( Args = [X] ->
        impure invoke_io1(show_val(X), Str),
        impure invoke_io1(string_to_mars_value(Str), Ret)
    ;
        throw_invalid
    )
).

% error(msg :: Array(Int)) :: a
:- func builtin_error = runtime_func.
builtin_error = make_builtin("error", 1, impure func(Args) = _Ret :-
    ( Args = [Msg] ->
        impure invoke_io1(mars_value_to_string(Msg, "error"), MsgStr),
        throw_mars_error(MsgStr)
    ;
        throw_invalid
    )
).

:- func all_builtins = map(string, runtime_func).
all_builtins = Map :-
    some [!Map] (
        map.init(!:Map),
        % Comparison
        svmap.det_insert("eq", builtin_eq, !Map),
        svmap.det_insert("is", builtin_is, !Map),
        svmap.det_insert("cmp", builtin_cmp, !Map),
        % Arithmetic
        svmap.det_insert("add", builtin_add, !Map),
        svmap.det_insert("sub", builtin_sub, !Map),
        svmap.det_insert("mul", builtin_mul, !Map),
        svmap.det_insert("div", builtin_div, !Map),
        svmap.det_insert("mod", builtin_mod, !Map),
        % Array
        svmap.det_insert("array", builtin_array, !Map),
        svmap.det_insert("array_ref", builtin_array_ref, !Map),
        svmap.det_insert("__impure_array_set", builtin_impure_array_set,
                         !Map),
        svmap.det_insert("array_replace", builtin_array_replace, !Map),
        svmap.det_insert("__impure_array_append", builtin_impure_array_append,
                         !Map),
        svmap.det_insert("__impure_array_extend", builtin_impure_array_extend,
                         !Map),
        svmap.det_insert("array_concat", builtin_array_concat, !Map),
        svmap.det_insert("array_length", builtin_array_length, !Map),
        % IO
        svmap.det_insert("put_char", builtin_put_char, !Map),
        svmap.det_insert("get_char", builtin_get_char, !Map),
        svmap.det_insert("get_env", builtin_get_env, !Map),
        % Misc
        svmap.det_insert("show", builtin_show, !Map),
        svmap.det_insert("error", builtin_error, !Map),
        Map = !.Map
    ).
