% -------------------------------------------------------------------------- %
% Mars/Venus Implementation
% Copyright (C) 2008 Matt Giuca
% 3/4/2008, 16/5/2008
%
% IR (Intermediate Representation)
% Note: This code is shared between Mars and Venus.
% Some features are part of just the Mars language (which is a
% nearly-superset of Venus). These are marked MARS-ONLY and should not be
% used by Venus code.
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %
% The IR represents an entire program. Each function may be represented in two
% distinct forms: an Abstract Syntax Tree (AST) or Control Flow Graph (CFG).
% In this structure, each function may be represented in either form, but in
% practice, the whole program will be converted from AST to CFG at some point.
% Some passes (analysis, code generation, interpretation) expect all functions
% to be in CFG form, while others (pretty printing) only work properly if in
% AST form (though pretty printing will still print CFG, it just won't be
% valid code).
%
% AST form is an in-memory representation of the Mars source code syntax. It
% features statement blocks with compound statements such as if and while,
% which can feature sub-statements, and statements feature expressions which
% can feature sub-expressions.
%
% CFG form is better suited to analysis and code generation. It features basic
% blocks with terminating instructions which branch between blocks, and simple
% (flattened) instructions with no sub-expressions.
% The cfg module supports the data structures used by this form.
% -------------------------------------------------------------------------- %

:- module ir.

:- interface.

:- import_module integer.
:- import_module string.
:- import_module bool.
:- import_module list.
:- import_module maybe.
:- import_module pair.

:- import_module term.
:- import_module varset.

:- import_module cfg.
:- import_module types.
:- import_module context.

% A variable name.
% retvname is the special variable name given to the return value (which may
% also be qualified).
% svname is a simple string variable name (local or global).
% gvname is a svname but indicates that this is a global variable
%   (only used in certain contexts).
% qvname is a qualified variable name (which is a variable name qualified by
% one or more integer subscripts).
% This is used chiefly to disambiguate variables in static single assignment
% (SSA) form which were spawned from the same variable in the original source.
:- type varname
    --->    retvname
    ;       tempvname
    ;       svname(string)
    ;       gvname(string)
    ;       qvname(varname, int).

% A "kind" (type of a type).
% Note that in a "full" kind system, kind_arrow would be
% kind_arrow(list(kind), bool, kind)
% (kind_arrow(ArgKinds, IsVariadic, ResultKind))
% This is a simplified system requiring all arguments and the result to be of
% kind *. Hence, we only need to keep track of the number of arguments, and
% the result kind is implicit.
:- type kind
    --->    kind_star
            % kind_arrow(Arity, IsVariadic)
    ;       kind_arrow(int, bool).

% typedef - Represents a user-defined type (may be a type constructor).
% An algebraic data type.
:- type typedef
    --->    typedef(
                typedef_name    :: string,
                typedef_params  :: maybe(list(var)),
                typedef_ctors   :: list(ctor),
                typedef_varset  :: varset,
                typedef_context :: context.context
            ).

% ctor - A constructor as part of a typedef.
% Vars contained within reference the type_varset of the parent typedef.
:- type ctor
    --->    ctor(
                ctor_name :: string,
                ctor_args :: maybe(list(pair(maybe(string), typeval))),
                ctor_context :: context.context
            ).

% expr - A term expression in the AST form of a program.
:- type expr
    --->    expr(
                expr_expr :: expr_,
                expr_type :: maybe(typeval)
            ).
:- type expr_
            % Integer literal. (Note arbitrary precision).
    --->    intlit(integer)
            % Array literal.
    ;       arraylit(list(expr))
            % Variable reference.
    ;       varref(varname)
            % Constructor function reference. String is the name of the ctor.
            % This is semantically the same as varref, but is distinguished
            % because it is separate both syntactically, and in the
            % implementation.
    ;       ctorref(string)
            % Field reference. String is the name of the field.
    ;       fieldref(expr, string)
            % Function application. First is a function. Applies this to the
            % second (list of expressions).
    ;       app(expr, list(expr))
            % Partial function application.
    ;       parapp(expr, list(expr)).

% stmt: Represents a statement in the AST form of a program.
:- type stmt
    --->    basic_stmt(basic_stmt, context.context)
    ;       compound_stmt(compound_stmt, context.context).

% basic_stmt: A statement with no control flow or sub-statements. Used by only
% the AST program representation (in CFG representation, these statements are
% replaced by the flattened 'instructions' without sub-expressions).
:- type basic_stmt
            % Pass statement
    --->    pass
            % Assignment statement. Evaluates expr and assigns it to variable
            % varname.
    ;       assign(varname, expr)
            % Field update (destructive). (Mars only).
    ;       fieldset(expr, string, expr)
            % Field replace (non-destructive).
    ;       fieldreplace(expr, string, expr)
            % Eval-then-throw-away statement. Evaluates expr, then throws it
            % away.
            % (This is used to cause side-effects).
    ;       eval(expr).

% compound_stmt: A statement with control flow or sub-statements. Used only by
% the AST program representation (in CFG representation, these statements are
% replaced by terminator_instrs).
:- type compound_stmt
            % Return statement. Evaluates expr, then ends the current function,
            % returning that value.
    --->    return(expr)
            % Switch statement. Performs pattern matching of algebraic data
            % types. Executes the first stmt_block to match the pattern
            % (binding variables from the pattern over the block's scope).
    ;       switch(expr, list(case_stmt))
            % If-then-else statement. Expr must be of type Int.
            % First list of statements is the "then part", second is the
            % "else part".
            % Evaluates Expr, then executes the "then part" if it is nonzero,
            % or the "else part" if it is zero.
    ;       if_then_else(expr,stmt_block,stmt_block)
            % While statement. Expr must be of type Int.
            % Evaluates Expr, then executes the "then part" if it is nonzero,
            % repeating until the condition is zero.
    ;       while(expr,stmt_block).

% instr: A low-level statement with no control flow, sub-statements or
% sub-expressions. Used by only the CFG program representation.
:- type instr
    --->    instr(
                instr_instr :: instr_,
                instr_context :: context.context
            ).
:- type instr_
            % NOP instruction (equivalent of pass statement)
    --->    nop
            % Copy instruction. Copies a value from one variable to another.
    ;       mov(varname, varname)
            % Field update (destructive). (Mars only). (S, F, V) <=> S.F := V
    ;       i_fieldset(varname, string, varname)
            % Field replace (non-destructive). (R, S, F, V) <=> R := S.F <- V
    ;       i_fieldreplace(varname, varname, string, varname)
            % Load an integer literal. (Note arbitrary precision).
    ;       ld_intlit(varname, integer)
            % Construct an array from a list of variables.
    ;       ld_arraylit(varname, list(varname))
            % Load a constructor function into a variable.
            % String is the name of the ctor.
    ;       ld_ctor(varname, string)
            % Load a field of a variable into a variable.
            % (R, S, F) <=> R := S.F
    ;       ld_field(varname, varname, string)
            % Function call. (Result, Function, Args).
    ;       call(varname, varname, list(varname))
            % Partial function application. (Result, Function, Args).
    ;       parcall(varname, varname, list(varname)).

% terminator_instr: An instruction which may only appear at the end of a basic
% block in the CFG representation. These roughly are the equivalents of the
% compound statements in the AST representation.
% Note: This has a type variable, S, which is the store type. It's a "dummy
% type" used to prevent terminators from referencing basic blocks in the wrong
% CFG.
:- type terminator_instr(S)
            % Switch instruction. Performs pattern matching of algebraic data
            % types. Branches to the first basic block to match the pattern
            % (binding variables from the pattern over the block's scope).
    --->    t_switch(varname, list(cfg_case_stmt(S)))
            % Conditional branch instruction. Expr must be of type Int.
            % Evaluates Expr, then branches to the first basic block if it is
            % nonzero, or the second basic block if it is zero.
    ;       cond_branch(varname, bbref(S), bbref(S))
            % Unconditionally branches to the given basic block.
    ;       branch(bbref(S))
            % "Null terminator". It is a semantic error if a block with a null
            % terminator is reachable (since all blocks must do *something* at
            % the end). However, it is still syntactically possible (and may
            % be a valid program; eg. if the block is unreachable).
    ;       null_terminator.

% A statement block.
:- type stmt_block == list(stmt).

% An AST representation's case statement.
:- type case_stmt
    --->    case_stmt(pattern, stmt_block, context.context).

% A CFG representation's case statement.
:- type cfg_case_stmt(S)
    --->    cfg_case_stmt(pattern, cfg.bbref(S), context.context).
% A pattern (part of a case statement).
:- type pattern
    --->    pattern(
                pat_pattern :: pattern_,
                pat_type    :: maybe(typeval)
            ).
:- type pattern_
            % "_" pattern (don't care / wildcard)
    --->    pat_any
            % Variable. Matches anything, and binds a new variable of the
            % given name.
    ;       pat_var(varname)
            % Constructor match. Matches only values with the given ctor
            % name, and then recursively matches all patterns.
    ;       pat_ctor(string, maybe(list(pattern)))
            % Integer literal. Matches exactly this value.
    ;       pat_intlit(integer).

% A uniqueness annotation for a particular argument of a function
:- type argmode
    --->    argmode(
                arg_kill        :: bool,
                arg_interest    :: bool
            ).

% Default argmode for most arguments
% def_argmode = argmode(no, no).
% (Not death, not interesting)
:- func def_argmode = argmode.

% Top-level definition of a function.
:- type function
    --->    function(
                func_name     :: string,
                func_params   :: maybe(list({string, typeval, argmode})),
                func_ret_type :: typeval,
                func_varset   :: varset,
                func_body     :: func_body,
                func_context  :: context.context
            ).

% Represents the implementation of a function
% (its declarations and statements).
:- type func_body
            % This function is builtin, so has no body.
    --->    func_builtin
            % The function is represented by an AST.
    ;       func_body_ast(
                list(pair(varname, typeval)),   % Decls
                stmt_block                      % Body
            )
            % The function is represented by a CFG.
    ;       some [S] func_body_cfg(
                list(pair(varname, typeval)),
                cfg(S)
            ).

:- type import == pair(string, context.context).

% Program: An ordered sequence of types and functions.
% (Kept in order for printing back - the render of a program preserves the
% order -- see ProgTable for a searchable version of Program).
:- type program ---> program(list(program_node)).
:- type program_node
    --->    ptypedef(typedef)
    ;       pfunction(function)
    ;       pimport(import).

% -------------------------------------------------------------------------- %

% typedef_kind(TypeDef) returns the kind of a given TypeDef.
:- func typedef_kind(typedef) = kind.

% typedef_type(Typedef, Varset, Type)
% returns the type which refers to Typedef, and a Varset which corresponds
% to the type's variables.
:- pred typedef_type(typedef::in, typeval::out) is det.

% ctor_type(Typedef, Ctor, Varset, Type)
% returns the type of Typedef.Ctor, and a Varset which corresponds to the
% type's variables.
% Ctor is a string, the name of the ctor for this typedef.
% Fails if Ctor is not found in the typedef.
:- pred ctor_type(typedef::in, string::in, varset::out, typeval::out)
                    is semidet.
% ctor_type(Typedef, CtorName, Varset, Type, Ctor)
:- pred ctor_type(typedef::in, string::in, varset::out, typeval::out,
                    ctor::out) is semidet.

% function_type(Func, Varset, Type)
% returns the type of Func, and a Varset which corresponds to the type's
% variables.
:- pred function_type(function::in, varset::out, typeval::out) is det.

% varname_to_string(Varname)
% Displays a variable name as a string. This function is injective (the string
% generated is unique to the given varname) and produces a valid identifier.
:- func varname_to_string(varname) = string.

% varname_to_string_noescape(Varname)
% Displays a variable name as a string. This function is injective.
% Uses ':' as a qualifier instead of '_'. Does not need to modify the string.
% While this is injective, and does not modify existing strings, it does not
% produce valid identifiers.
% Use for error messages (to avoid confusion with renaming variables), and
% perhaps pretty-printing, but not for code generation where variables must
% be valid identifiers.
:- func varname_to_string_noescape(varname) = string.

% varname_to_string_dodgy(Varname)
% Displays a variable name as a string. This function is dodgy (not injective,
% and therefore does not guarantee unique function names). It exists for
% backends like Python which need unqualified names to appear as-is, but also
% can't use any other escape characters (so it might break if vertain variable
% names are used - but then, the Python backend does that anyway).
:- func varname_to_string_dodgy(varname) = string.

% varname_is_global(Varname)
% Succeeds if Varname is a global variable name (gvname).
:- pred varname_is_global(varname::in) is semidet.

% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- implementation.

:- import_module int.
:- import_module map.
:- import_module require.

:- func tab_width = int.
tab_width = 4.

:- func tab(int) = string.
tab(N) = string.duplicate_char(' ', N*tab_width).

typedef_kind(typedef(_,no,_,_,_)) = kind_star.
typedef_kind(typedef(_,yes(Params),_,_,_)) = kind_arrow(length(Params), no).

%:- func typeval_kind(typeval) = kind.
% XXX: Needs to be passed a progtable in order to look up type_prim names.

typedef_type(typedef(Name, no, _, _, _), types.const(Name)).
typedef_type(typedef(Name, yes(Params), _, _, _), Type) :-
    map(pred(V::in, T::out) is det :- T = types.variable(V),
        Params, PVars),
    Type = types.app(types.const(Name), PVars).

def_argmode = argmode(no, no).

% The (implicit) argmode of all ctor arguments
:- func ctor_args_mode = argmode.
ctor_args_mode = argmode(no, no).

ctor_type(TypeDef@typedef(_, _, Ctors, Varset, _), CtorName, Varset, Type,
        Ctor) :-
    lookup_ctor(Ctors, CtorName, Ctor@ctor(_,Args,_)),     % Can fail
    typedef_type(TypeDef, TypedefType),
    (
        % Parameter-less constructor - type is the same as the defined type
        Args = no,
        Type = TypedefType
    ;
        % Proper constructor function
        Args = yes(YesArgs),
        list.map(pred(_-T::in, T::out) is det, YesArgs, ArgTypes),
        Type = types.app(types.functype, ArgTypes ++ [TypedefType])
    ).

ctor_type(TypeDef, CtorName, Varset, Type) :-
    ctor_type(TypeDef, CtorName, Varset, Type, _).

:- pred lookup_ctor(list(ctor)::in, string::in, ctor::out) is semidet.
lookup_ctor([C|Cs], Name, COut) :-
    ( C = ctor(Name,_,_) ->
        COut = C
    ;
        lookup_ctor(Cs, Name, COut)
    ).

function_type(Func, Func^func_varset, Type) :-
    (
        Func^func_params = no,
        Type = Func^func_ret_type
    ;
        Func^func_params = yes(Params),
        list.map(pred({_,T,_}::in, T::out) is det, Params, ParamTypes),
        Type = types.app(types.functype, ParamTypes ++ [Func^func_ret_type])
    ).

% Escape each '_' character as a "__", in order to guarantee injectivity (a
% single underscore always represents a qualifier separator).
varname_to_string(retvname) = "_RET_".
varname_to_string(tempvname) = "_T_".
varname_to_string(svname(Name)) = string.replace_all(Name, "_", "__").
varname_to_string(gvname(Name)) = Name.
varname_to_string(qvname(BaseName, Qual)) = varname_to_string(BaseName)
    ++ "_" ++ string.from_int(Qual).

varname_to_string_noescape(retvname) = "$RET".
varname_to_string_noescape(tempvname) = "$T".
varname_to_string_noescape(svname(Name)) = Name.
varname_to_string_noescape(gvname(Name)) = "@" ++ Name.
varname_to_string_noescape(qvname(BaseName, Qual)) =
    varname_to_string_noescape(BaseName) ++ ":" ++ string.from_int(Qual).

varname_to_string_dodgy(retvname) = "__RET__".
varname_to_string_dodgy(tempvname) = "__T__".
varname_to_string_dodgy(svname(Name)) = Name.
varname_to_string_dodgy(gvname(Name)) = Name.
varname_to_string_dodgy(qvname(BaseName, Qual)) =
    varname_to_string_dodgy(BaseName) ++ "_" ++ string.from_int(Qual).

varname_is_global(gvname(_)).
