% -------------------------------------------------------------------------- %
% Mars top-level program (interactive interpreter)
% Copyright (C) 2008-2009 Matt Giuca
% 6/1/2009
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- module mars.

:- interface.

:- import_module io.

:- pred main(io::di, io::uo) is cc_multi.

% -------------------------------------------------------------------------- %

:- implementation.

:- func version_string = string.
version_string = "mars 0.1.1~bzr".
:- func copyright_year = string.
copyright_year = "2008-2009".

:- import_module list.
:- import_module map.
:- import_module int.
:- import_module integer.
:- import_module char.
:- import_module string.
:- import_module exception.
:- import_module require.
:- import_module univ.
:- import_module unit.
:- import_module bool.
:- import_module maybe.
:- import_module getopt.
:- import_module varset.

:- import_module ir.
:- import_module interactive.
:- import_module interpret.
:- import_module util.
:- import_module tables.
:- import_module types.
:- import_module builtins.
:- import_module token.
:- import_module pretty.
:- import_module reader.
:- import_module context.
:- import_module typecheck.
:- import_module ast_cfg.

% exec_main_func(MainFuncName, ProgTable, !IO).
% Executes the "main" function of a program, displaying the results of any
% side-effects.
:- pred exec_main_func(string::in, tables.progtable::in, io::di, io::uo)
    is cc_multi.
exec_main_func(MainFuncName, ProgTable, !IO) :-
    % Make sure MainFuncName is a valid lower_identifier token
    ( token.is_lident(MainFuncName) ->
        true
    ;
        context.throw_error("Illegal function name: \"" ++ MainFuncName
            ++ "\"")
    ),
    % MainExpr is the expression "MainFuncName"
    MainExpr = expr(varref(svname(MainFuncName)),no),
    % Typecheck this expression (makes sure main exists, and get its type)
    typecheck.expr_type(blank_context, ProgTable, empty_localtable,
        varset.init, VarSet, MainExpr, _, MainType),
    % Ensure that MainType is () -> Int
    ExpType = app(functype, [const("Int")]),
    typecheck.det_unify(blank_context, MainFuncName, MainExpr, MainType,
                        ExpType, VarSet, _NewVarset),
    % MainInstr is the instruction "$T = MainFuncName()"
    MainInstr = instr(call(tempvname, svname(MainFuncName), []),
                      context.blank_context),
    % Evaluate the expression
    Env0 = interpret.progtable_env(ProgTable),
    try_io_thread1(interpret.exec_instr(MainInstr), TryResult, Env0,Env, !IO),
    (
        TryResult = succeeded(_),
        % Get the result out of the variable '$T'.
        interpret.read_var(Env, tempvname, MainResult, !IO),
        ( MainResult = val_int(ResultInteger) ->
            ResultInt = integer.int(ResultInteger mod integer(int.max_int)),
            % Set this as the exit status of `mars`
            io.set_exit_status(ResultInt, !IO)
        ;
            error("Result of main was not an int")
        )
    ;
        TryResult = exception(E),
        ( univ_to_type(E, mars_error(Msg)) ->
            io.write_string(io.stderr_stream, "Runtime Error: ", !IO),
            io.write_string(io.stderr_stream, Msg, !IO),
            io.nl(io.stderr_stream, !IO)
        ;
            throw(E)
        )
    ).

% Does the whole program, (taking a filename as input), but does not catch
% exceptions. Any context.error exceptions will be propagated upwards.
% Filename may be `no`, in which case it will use prelude by default.
:- pred do_everything(option_table::in, maybe(string)::in, unit::out,
                io::di, io::uo) is cc_multi.
do_everything(OptionTable, MaybeFilename, unit, !IO) :-
    % Get the options sorted out
    MainFunc = OptionTable ^ mainfunc,
    Command = OptionTable ^ command,

    % Get a definitive filename
    (
        MaybeFilename = yes(Filename),
        Interact = OptionTable ^ interact
    ;
        MaybeFilename = no,
        % No filename specified. Use prelude by default, and also
        % go into -i mode.
        reader.locate_module("prelude", MaybeFilename1, !IO),
        (
            MaybeFilename1 = yes(Filename)
        ;
            MaybeFilename1 = no,
            % Prelude not found
            context.throw_error("Can't find module named \"prelude\"")
        ),
        % Always -i mode (regargless of OptionTable ^ interact)
        Interact = yes
    ),

    % Open and parse the program
    reader.read_source_file(Filename, Prog, !IO),    % Can throw
    % Print the resulting token and pos.
    % Build a prog table
    % (Can throw)
    tables.build_progtable(Prog, builtins.builtins_progtable, ProgTable),
    % Print everything out
    %print_stuff(Prog, ProgTable, !IO),
    % Typecheck the program
    typecheck.check_program(ProgTable, Prog, _Prog_Checked),
    % Convert the program from AST representation into CFG representation
    ast_cfg.prog_to_cfg(Prog, ProgCFG),
    % Build another ProgTable
    tables.build_progtable(ProgCFG, builtins.builtins_progtable,
        ProgTableCFG),
    ( Command = no, Interact = no ->
        % Execute the main function, and terminate
        exec_main_func(MainFunc, ProgTableCFG, !IO)
    ;
        ( Command = yes(CommandText) ->
            % Execute the -c command before going into interactive mode
            interactive.do_user_command(CommandText, ProgTableCFG, !IO)
        ;
            true
        ),
        ( Interact = yes ->
            % Ask the user for an expression, and print its type.
            interactive.prompts(ProgTableCFG, !IO)
        ;
            % Just finish
            true
        )
    ).

%% OPTIONS PROCESSING %%

:- pred version_info(io::di, io::uo) is det.
version_info(!IO) :-
    io.write_string(version_string, !IO),
    io.write_string("\n\n", !IO),
    io.write_string(string.format("Mars Copyright (C) %s Matt Giuca\n\n",
        [s(copyright_year)]), !IO),
    io.write_string("This program comes with ABSOLUTELY NO WARRANTY\n",
        !IO),
    io.write_string("This is free software; you can redistribute it under"
        ++ " the terms of the GNU GPL\nversion 3 or later.\n", !IO),
    io.write_string("See <http://www.gnu.org/licenses/>.\n", !IO).

:- pred usage(io::di, io::uo) is det.
usage(!IO) :-
    usage(no, !IO).
:- pred usage(bool::in, io::di, io::uo) is det.
usage(_Help, !IO) :-
    io.write_string("Mars interactive interpreter\n", !IO),
    io.write_string("Usage: ./mars [OPTIONS] [mars-source-file]\n", !IO),
    io.write_string("Options:\n", !IO),
    io.write_string("    -i                     "
        ++ "Start an interactive session\n", !IO),
    io.write_string("    -c COMMAND             "
        ++ "Mars statement or interactive command to execute\n", !IO),
    io.write_string("    -m NAME                "
        ++ "Name of the main function (default: \"main\")\n", !IO),
    io.write_string("    -h, --help             "
        ++ "Show this help, and exit\n", !IO),
    io.write_string("    --version              "
        ++ "Show version information, and exit\n", !IO).

:- type option
    --->    interact        % -i
    ;       command         % -c, --command=
    ;       mainfunc        % -m, --main=
    ;       help            % -h, --help
    ;       version         % --version
    .

:- type option_table
    --->    option_table(
                interact :: bool,
                command  :: maybe(string),
                mainfunc :: string,
                help     :: bool,
                version  :: bool
            ).

:- pred short_options(char::in, option::out) is semidet.
short_options('i', interact).
short_options('c', command).
short_options('m', mainfunc).
short_options('h', help).

:- pred long_options(string::in, option::out) is semidet.
long_options("command", command).
long_options("main", mainfunc).
long_options("help", help).
long_options("version", version).

:- pred option_default(option::out, option_data::out) is multi.
option_default(interact, bool(no)).
option_default(command, maybe_string(no)).
option_default(mainfunc, string("main")).
option_default(help, bool(no)).
option_default(version, bool(no)).

:- func make_our_option_table(option_table(option)) = option_table.
make_our_option_table(OT)
    = option_table(Interact, Command, MainFunc, Help, Version) :-
    ( map.search(OT, interact, bool(Interact_)) ->
        Interact = Interact_
    ;
        error("make_our_option_table: Invalid interact field in option table")
    ),
    ( map.search(OT, command, maybe_string(Command_)) ->
        Command = Command_
    ;
        error("make_our_option_table: Invalid command field in option table")
    ),
    ( map.search(OT, mainfunc, string(MainFunc_)) ->
        MainFunc = MainFunc_
    ;
        error("make_our_option_table: Invalid mainfunc field in option table")
    ),
    ( map.search(OT, help, bool(Help_)) ->
        Help = Help_
    ;
        error("make_our_option_table: Invalid help field in option table")
    ),
    ( map.search(OT, version, bool(Version_)) ->
        Version = Version_
    ;
        error("make_our_option_table: Invalid version field in option table")
    ).

% option_default_nondet - same as option_default, but nondet instead of multi.
% The getopt documentation says it should be multi, but it's a compiler error
% unless it's nondet. This function (hack) tricks the compiler into thinking
% it's nondet.
:- pred option_default_nondet(option::out, option_data::out) is nondet.
option_default_nondet(Option, OD) :-
    always_succeeds_semidet,
    option_default(Option, OD).

% getopts(NonOptionArgs, OptionTable, !IO).
% Retrieves command-line arguments and parses them into an option table.
% On error, maybe_option_table is set to error.
% Sets SourceFilename to the single non-option argument (error if there is
% more than one).
:- pred getopts(list(string)::out, maybe_option_table(option)::out,
                io::di, io::uo) is det.
getopts(NonOptionArgs, OptionTable, !IO) :-
    io.command_line_arguments(Args, !IO),
    getopt.process_options(
        option_ops(short_options, long_options, option_default_nondet),
        Args, NonOptionArgs, OptionTable).

%% END OPTIONS PROCESSING %%

main(!IO) :-
    % Get options from command-line
    getopts(NonOptionArgs, MaybeOptionTable, !IO),
    (
        MaybeOptionTable = error(Message),
        ( Message \= "" ->
            io.write_string(Message, !IO),
            io.nl(!IO)
        ;
            true
        ),
        usage(!IO)
    ;
        MaybeOptionTable = ok(OptionTable0),
        OptionTable = make_our_option_table(OptionTable0),
        ( OptionTable ^ help = yes ->
            usage(yes, !IO)
        ; OptionTable ^ version = yes ->
            version_info(!IO)
        ;
            ( NonOptionArgs = [_,_|_] ->
                io.write_string("Too many arguments\n", !IO),
                usage(!IO)
            ;
                ( NonOptionArgs = [Filename_] ->
                    Filename = yes(Filename_)
                ;
                    Filename = no
                ),
                % Call do_everything, catching and printing
                % any context.error thrown.
                try_io(do_everything(OptionTable, Filename), Result, !IO),
                (
                    Result = succeeded(_)
                ;
                    Result = exception(E),
                    io.set_exit_status(1, !IO),
                    ( univ_to_type(E, PE : context.error) ->
                        print_error(PE, !IO)
                    ;
                        throw(E)
                    )
                )
            )
        )
    ).
