% -------------------------------------------------------------------------- %
% Marsc (Mars compiler) top-level program
% Copyright (C) 2008-2009 Matt Giuca
% 6/1/2009
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- module marsc.

:- interface.

:- import_module io.

:- pred main(io::di, io::uo) is cc_multi.

% -------------------------------------------------------------------------- %

:- implementation.

:- func version_string = string.
version_string = "marsc 0.1.1~bzr".
:- func copyright_year = string.
copyright_year = "2008-2009".

:- import_module list.
:- import_module map.
:- import_module int.
:- import_module char.
:- import_module string.
:- import_module exception.
:- import_module require.
:- import_module univ.
:- import_module unit.
:- import_module bool.
:- import_module maybe.
:- import_module getopt.

:- import_module ir.
:- import_module interactive.
:- import_module util.
:- import_module tables.
:- import_module types.
:- import_module varset.
:- import_module builtins.
:- import_module pretty.
:- import_module reader.
:- import_module context.
:- import_module typecheck.
:- import_module ast_cfg.

% use_backend(BackendName, ProgTable, Program, MaybeMainFunction, !IO).
% Executes the backend of the given name. Will throw a context_error if the
% backend isn't found.
% The backend will look at the given program, compile it, and print it out
% to the current output stream.
:- pred use_backend(string::in, progtable::in, program::in, maybe(string)::in,
                    io::di, io::uo) is det.
use_backend(BackendName, _PT, Program, _Main, !IO) :-
    ( BackendName = "pretty" ->
        pretty.print_program(Program, !IO)
    ; BackendName = "asm" ->
        % Differs from "pretty" in that it uses CFG representation
        pretty.print_program(Program, !IO)
    ;
        context.throw_error("Unknown backend: " ++ BackendName)
    ).

% backend_use_cfg(BackendName).
% Succeeds if the named backend expects CFG representation.
:- pred backend_use_cfg(string::in) is semidet.
backend_use_cfg("asm").

% backend_file_exts maps backend names on to default file extensions.
:- func backend_file_exts = map(string, string).
backend_file_exts = M :-
    some [!M] (
        map.init(!:M),
        map.set(!.M, "pretty", ".pretty.mar", !:M),
        map.set(!.M, "asm", ".mas", !:M),
        M = !.M
    ).

% default_output_name(BackendName, InputFilename) = OutputFilename.
% Will throw a context_error if the backend isn't found.
:- func default_output_name(string, string) = string.
default_output_name(BackendName, InputFilename) = Base ++ Ext :-
    ( util.split_ext(InputFilename, Basename, _) ->
        Base = Basename
    ;
        Base = InputFilename
    ),
    ( map.search(backend_file_exts, BackendName, Ext_) ->
        Ext = Ext_
    ;
        context.throw_error("Unknown backend: " ++ BackendName)
    ).

% check_main_func(ProgTable, MaybeMainFunc) succeeds iff:
%   The given main function is in the program and has type () -> Int, or
%   There is no given main function name.
% Otherwise throws an exception.
:- pred check_main_func(progtable::in, maybe(string)::in) is det.
check_main_func(_PT, no).
check_main_func(PT, yes(MainFunc)) :-
    % MainExpr is the expression "MainFuncName"
    MainExpr = expr(varref(svname(MainFunc)),no),
    % Typecheck this expression (makes sure main exists, and get its type)
    typecheck.expr_type(blank_context, PT, empty_localtable,
        varset.init, VarSet, MainExpr, _, MainType),
    % Ensure that MainType is () -> Int
    ExpType = app(functype, [const("Int")]),
    typecheck.det_unify(blank_context, MainFunc, MainExpr, MainType,
                        ExpType, VarSet, _NewVarset).

% Does the whole program, (taking a filename as input), but does not catch
% exceptions. Any context.error exceptions will be propagated upwards.
:- pred do_everything(option_table::in, string::in, unit::out,
                io::di, io::uo) is det.
do_everything(OptionTable, Filename, unit, !IO) :-
    % Get the options sorted out
    BackendName = OptionTable ^ backend,
    % Do this whether we use it or not; it throws an exception we want
    DefaultOutput = default_output_name(BackendName, Filename),
    ( OptionTable ^ output = yes(Output_) ->
        OutFilename = Output_
    ;
        OutFilename = DefaultOutput
    ),
    MainFunc = OptionTable ^ mainfunc,

    % Open and parse the program
    reader.read_source_file(Filename, Prog, !IO),    % Can throw
    % Print the resulting token and pos.
    % Build a prog table
    % (Can throw)
    tables.build_progtable(Prog, builtins.builtins_progtable, ProgTable),
    % Print everything out
    %print_stuff(Prog, ProgTable, !IO),
    % Typecheck the program
    typecheck.check_program(ProgTable, Prog, _Prog_Checked),
    % Check that MainFunc exists and has the correct type (() -> Int).
    check_main_func(ProgTable, MainFunc),
    ( backend_use_cfg(BackendName) ->
        % Convert the program from AST representation into CFG representation
        ast_cfg.prog_to_cfg(Prog, Prog2),
        % Build another ProgTable
        tables.build_progtable(Prog2, builtins.builtins_progtable,
            ProgTable2)
    ;
        % Just use regular AST representation for this backend
        Prog2 = Prog,
        ProgTable2 = ProgTable
    ),
    % Open the output filename (a Python program)
    io.open_output(OutFilename, OpenRes2, !IO),
    (
        OpenRes2 = ok(OutFile),
        % Set OutFile as the output stream
        io.set_output_stream(OutFile, _, !IO),
        % Write the program out
        % XXX Do not use the CFG yet (as backends don't support it)
        use_backend(BackendName, ProgTable2, Prog2, MainFunc, !IO),
        % Set output stream back to stdout
        io.set_output_stream(io.stdout_stream, _, !IO),
        io.close_output(OutFile, !IO)
    ;
        OpenRes2 = error(E),
        io.write_string(io.error_message(E), !IO),
        io.nl(!IO)
    ).

%% OPTIONS PROCESSING %%

:- pred version_info(io::di, io::uo) is det.
version_info(!IO) :-
    io.write_string(version_string, !IO),
    io.write_string("\n\n", !IO),
    io.write_string(string.format("Mars Copyright (C) %s Matt Giuca\n\n",
        [s(copyright_year)]), !IO),
    io.write_string("This program comes with ABSOLUTELY NO WARRANTY\n",
        !IO),
    io.write_string("This is free software; you can redistribute it under"
        ++ " the terms of the GNU GPL\nversion 3 or later.\n", !IO),
    io.write_string("See <http://www.gnu.org/licenses/>.\n", !IO).

:- pred usage(io::di, io::uo) is det.
usage(!IO) :-
    usage(no, !IO).
:- pred usage(bool::in, io::di, io::uo) is det.
usage(_Help, !IO) :-
    io.write_string("Mars compiler\n", !IO),
    io.write_string("Usage: ./marsc [OPTIONS] mars-source-file\n", !IO),
    io.write_string("Options:\n", !IO),
    io.write_string("    -o FILENAME            "
        ++ "Output filename\n", !IO),
    io.write_string("    --backend=BACKEND      "
        ++ "Specify the backend (default: asm)\n", !IO),
    io.write_string("                           "
        ++ "Backends: pretty, asm\n", !IO),
    io.write_string("    -m NAME                "
        ++ "Name of the main function (default: none)\n", !IO),
    io.write_string("    -h, --help             "
        ++ "Show this help, and exit\n", !IO),
    io.write_string("    --version              "
        ++ "Show version information, and exit\n", !IO).

:- type option
    --->    output          % -o
    ;       backend         % --backend=
    ;       mainfunc        % -m, --main=
    ;       help            % -h, --help
    ;       version         % --version
    .

:- type option_table
    --->    option_table(
                output   :: maybe(string),
                backend  :: string,
                mainfunc :: maybe(string),
                help     :: bool,
                version  :: bool
            ).

:- pred short_options(char::in, option::out) is semidet.
short_options('o', output).
short_options('m', mainfunc).
short_options('h', help).

:- pred long_options(string::in, option::out) is semidet.
long_options("backend", backend).
long_options("main", mainfunc).
long_options("help", help).
long_options("version", version).

:- pred option_default(option::out, option_data::out) is multi.
option_default(output, maybe_string(no)).
option_default(backend, string("asm")).
option_default(mainfunc, maybe_string(no)).
option_default(help, bool(no)).
option_default(version, bool(no)).

:- func make_our_option_table(option_table(option)) = option_table.
make_our_option_table(OT)
    = option_table(Output, Backend, MainFunc, Help, Version) :-
    ( map.search(OT, output, maybe_string(Output_)) ->
        Output = Output_
    ;
        error("make_our_option_table: Invalid output field in option table")
    ),
    ( map.search(OT, backend, string(Backend_)) ->
        Backend = Backend_
    ;
        error("make_our_option_table: Invalid backend field in option table")
    ),
    ( map.search(OT, mainfunc, maybe_string(MainFunc_)) ->
        MainFunc = MainFunc_
    ;
        error("make_our_option_table: Invalid mainfunc field in option table")
    ),
    ( map.search(OT, help, bool(Help_)) ->
        Help = Help_
    ;
        error("make_our_option_table: Invalid help field in option table")
    ),
    ( map.search(OT, version, bool(Version_)) ->
        Version = Version_
    ;
        error("make_our_option_table: Invalid version field in option table")
    ).

% option_default_nondet - same as option_default, but nondet instead of multi.
% The getopt documentation says it should be multi, but it's a compiler error
% unless it's nondet. This function (hack) tricks the compiler into thinking
% it's nondet.
:- pred option_default_nondet(option::out, option_data::out) is nondet.
option_default_nondet(Option, OD) :-
    always_succeeds_semidet,
    option_default(Option, OD).

% getopts(NonOptionArgs, OptionTable, !IO).
% Retrieves command-line arguments and parses them into an option table.
% On error, maybe_option_table is set to error.
% Sets SourceFilename to the single non-option argument (error if there is
% more than one).
:- pred getopts(list(string)::out, maybe_option_table(option)::out,
                io::di, io::uo) is det.
getopts(NonOptionArgs, OptionTable, !IO) :-
    io.command_line_arguments(Args, !IO),
    getopt.process_options(
        option_ops(short_options, long_options, option_default_nondet),
        Args, NonOptionArgs, OptionTable).

%% END OPTIONS PROCESSING %%

main(!IO) :-
    % Get options from command-line
    getopts(NonOptionArgs, MaybeOptionTable, !IO),
    (
        MaybeOptionTable = error(Message),
        ( Message \= "" ->
            io.write_string(Message, !IO),
            io.nl(!IO)
        ;
            true
        ),
        usage(!IO)
    ;
        MaybeOptionTable = ok(OptionTable0),
        OptionTable = make_our_option_table(OptionTable0),
        ( OptionTable ^ help = yes ->
            usage(yes, !IO)
        ; OptionTable ^ version = yes ->
            version_info(!IO)
        ;
            (
                NonOptionArgs = [],
                io.write_string("Missing source filename\n", !IO),
                usage(!IO)
            ;
                NonOptionArgs = [_,_|_],
                io.write_string("Too many arguments\n", !IO),
                usage(!IO)
            ;
                NonOptionArgs = [Filename],
                % Call do_everything, catching and printing
                % any context.error thrown.
                try_io(do_everything(OptionTable, Filename), Result, !IO),
                (
                    Result = succeeded(_)
                ;
                    Result = exception(E),
                    io.set_exit_status(1, !IO),
                    ( univ_to_type(E, PE : context.error) ->
                        print_error(PE, !IO)
                    ;
                        throw(E)
                    )
                )
            )
        )
    ).
