% -------------------------------------------------------------------------- %
% Generic parser library
% Copyright (C) 2008 Matt Giuca
% 22/5/2008
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- module parse.

:- interface.

:- import_module char.
:- import_module string.
:- import_module list.
:- import_module unit.

:- import_module context.

% -------------------------------------------------------------------------- %

% parser(TokenType, StateType).
% "parser" holds the parser state. parser objects are threaded through the
% parsing code. They hold both the stream, and also meta-information such
% as line numbers.
% TokenType is the type of tokens in the input stream (not the output).
% StateType is an optional type for any state you may wish to store along
% with the parser state.
% Using such an object as the DCG state type means you can't pull elements
% out of the stream in DCG [] notation. Instead, you must use predicates
% such as get.
:- type parser(T, S).
% parser(TokenType).
% Parser with no user state.
:- type parser(T) == parser(T, unit).

% -------------------------------------------------------------------------- %

% Note all of these are written in such a way that you can a) use them
% with state-variable or DCG notation, and b) they preserve uniqueness,
% by returning a new version of the input, even if they don't modify it.

% Creates a new parser from a list of things.
% parse.new_parser(Tokens, Filename, InitState, Parser).
% Filename does not have to be a real file (no IO is performed).
% It is just a string which is stored in the context and used for error
% messages.
:- pred parse.new_parser(list(T)::in, string::in, S::in, parser(T,S)::out)
        is det.
% parse.new_parser(Tokens, Filename, Parser).
% The state is unit.
:- pred parse.new_parser(list(T)::in, string::in, parser(T,unit)::out) is det.
% parse.new_parser(Tokens, Parser).
% The stored filename will be blank, and state is unit.
:- pred parse.new_parser(list(T)::in, parser(T,unit)::out) is det.

% Checks to see if the parser is empty.
:- pred parse.stream_empty(parser(T,S)::in, parser(T,S)::out) is semidet.

% Gets the context object for the current parser state.
:- pred parse.get_context(context::out, parser(T,S)::in, parser(T,S)::out)
        is det.

% Gets the state object for the current parser state. Does not update parser.
% (Has an out parser so it can be used in DCG syntax).
:- pred parse.get_state(S::out, parser(T,S)::in, parser(T,S)::out)
        is det.

% Sets the state object into the new parser state.
:- pred parse.set_state(S::in, parser(T,S)::in, parser(T,S)::out)
        is det.

% Gets a single token from the stream, updating the parser's state.
% As part of the input, accepts a function which updates the parser's position
% based on the current position and token.
% Fails iff the stream is empty.
:- pred parse.token_prim((pred(T,context,context)), T,
                            parser(T,S), parser(T,S)).
:- mode parse.token_prim((pred(in,in,out) is det), out, in, out) is semidet.

%% Handy utilities %%

% Applies a nullary parser 0 or more times. Stops once it fails.
:- pred parse.many(pred(parser(T,S), parser(T,S)), parser(T,S), parser(T,S)).
:- mode parse.many(pred(in, out) is semidet, in, out) is det.

% Applies a unary parser 0 or more times. Stops once it fails.
% Returns a list of all the values that were returned.
:- pred parse.many(pred(U, parser(T,S), parser(T,S)), list(U),
    parser(T,S), parser(T,S)).
:- mode parse.many(pred(out, in, out) is semidet, out, in, out) is det.

% Applies a nullary parser 1 or more times. Stops once it fails.
:- pred parse.many1(pred(parser(T,S), parser(T,S)), parser(T,S), parser(T,S)).
:- mode parse.many1(pred(in, out) is semidet, in, out) is semidet.

% Applies a unary parser 1 or more times. Stops once it fails.
% Returns a list of all the values that were returned.
:- pred parse.many1(pred(U, parser(T,S), parser(T,S)), list(U),
    parser(T,S), parser(T,S)).
:- mode parse.many1(pred(out, in, out) is semidet, out, in, out) is semidet.

% parse.sepby(Parser, Sep, ParserDescription, TokenToString, !P).
% Applies a nullary parser 0 or more times, ensuring each instance is
% separated by another nullary parser (such as comma). Stops once it fails.
% Needs a ParserDescription and TokenToString for error message. (Throws
% a parse error on a syntax error, if parser fails after a separator).
:- pred parse.sepby(pred(parser(T,S), parser(T,S)),
                    pred(parser(T,S), parser(T,S)),
                    string, func(T) = string,
                    parser(T,S), parser(T,S)).
:- mode parse.sepby(pred(in, out) is semidet, pred(in, out) is semidet,
                    in, in, in, out) is det.

% Applies a unary parser 0 or more times, ensuring each instance is
% separated by another nullary parser (such as comma). Stops once it fails.
% Returns a list of all the values that were returned.
:- pred parse.sepby(pred(U, parser(T,S), parser(T,S)),
                    pred(parser(T,S), parser(T,S)),
                    string, func(T) = string,
                    list(U), parser(T,S), parser(T,S)).
:- mode parse.sepby(pred(out, in, out) is semidet, pred(in, out) is semidet,
                    in, in, out, in, out) is det.

% Applies a nullary parser 1 or more times, ensuring each instance is
% separated by another nullary parser (such as comma). Stops once it fails.
:- pred parse.sepby1(pred(parser(T,S), parser(T,S)),
                    pred(parser(T,S), parser(T,S)),
                    string, func(T) = string,
                    parser(T,S), parser(T,S)).
:- mode parse.sepby1(pred(in, out) is semidet, pred(in, out) is semidet,
                    in, in, in, out) is semidet.

% Applies a unary parser 1 or more times, ensuring each instance is
% separated by another nullary parser (such as comma). Stops once it fails.
% Returns a list of all the values that were returned.
:- pred parse.sepby1(pred(U, parser(T,S), parser(T,S)),
                    pred(parser(T,S), parser(T,S)),
                    string, func(T) = string,
                    list(U), parser(T,S), parser(T,S)).
:- mode parse.sepby1(pred(out, in, out) is semidet, pred(in, out) is semidet,
                    in, in, out, in, out) is semidet.

% -------------------------------------------------------------------------- %

%% Specific for character parsers %%

% Character parser is the "default" for parser if no type specified.
:- type parser == parser(char, unit).

% Type cparser is also a character parser.
% The difference to parser is that cparser/1 lets you specify a state type.
:- type cparser(S) == parser(char, S).
:- type cparser == parser(char, unit).

% Gets a single character from the input stream.
% Fails iff the stream is empty.
:- pred parse.char(char, cparser(S), cparser(S)).
:- mode parse.char(out, in, out) is semidet.
% Get a single specific character from the input stream.
% Fails iff this character is not next in the stream.
% (Could be implied, but it's a very important mode, so it's explicit).
:- mode parse.char(in, in, out) is semidet.

% Gets a specific string from the input stream.
% Fails iff this string is not next in the stream.
:- pred parse.string(string::in, cparser(S)::in, cparser(S)::out) is semidet.

%% Special parsers for generic string productions %%

:- pred parse.alpha(char::out, cparser(S)::in, cparser(S)::out) is semidet.
:- pred parse.alpha_or_underscore(char::out, cparser(S)::in, cparser(S)::out)
    is semidet.
:- pred parse.alnum(char::out, cparser(S)::in, cparser(S)::out) is semidet.
:- pred parse.alnum_or_underscore(char::out, cparser(S)::in, cparser(S)::out)
    is semidet.
:- pred parse.digit(char::out, cparser(S)::in, cparser(S)::out) is semidet.
:- pred parse.hex_digit(char::out, cparser(S)::in, cparser(S)::out)
    is semidet.

% -------------------------------------------------------------------------- %

% Context and error handling

% parse.context_push_block(BlockName, !Parser) pushes a new block on the
% stack (ie. when the parser is entering a new function or typedef).
:- pred parse.context_push_block(string::in,
        parser(T,S)::in, parser(T,S)::out) is det.
% parse.context_pop_block(BlockName, !Parser) pops a block off the
% stack (ie. when the parser is exiting a function or typedef).
% Error if the block is already empty (this should not happen during normal
% parsing).
:- pred parse.context_pop_block(parser(T,S)::in, parser(T,S)::out) is det.

% Throws a parse.error.
% Signature like a normal parser pred so it can be used in DCG notation.
:- pred parse.throw_error(string::in, parser(T,S)::in, parser(T,S)::out)
        is erroneous.

% throw_unexpected(TokenToString, !Parser).
% Throws a parse_error exception "Unexpected <NextToken>", rendering the
% next token into a string with function TokenToString.
:- pred parse.throw_unexpected(func(T) = string::in,
    parser(T,S)::in, parser(T,S)::out) is erroneous.

% throw_expected(What, TokenToString, !Parser).
% Throws a parse_error exception "Unexpected <NextToken> (expected <What>)".
:- pred parse.throw_expected(string::in, func(T) = string::in,
    parser(T,S)::in, parser(T,S)::out) is erroneous.

% expect(Parser, What, TokenToString, !Parser).
% Run a nullary semidet parser production, and throw an "expected" parse error
% if it fails.
:- pred expect(pred(parser(T,S), parser(T,S)), string, func(T) = string,
                parser(T,S), parser(T,S)).
:- mode expect(pred(in, out) is semidet, in, in, in, out) is det.

% expect(Parser, What, TokenToString, Val, !Parser).
% Run a unary semidet parser production, and throw an "expected" parse error
% if it fails. Return its value on success.
:- pred expect(pred(U, parser(T,S), parser(T,S)), string, func(T) = string,
                U, parser(T,S), parser(T,S)).
:- mode expect(pred(out, in, out) is semidet, in, in, out, in, out) is det.

% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- implementation.

:- import_module int.
:- import_module stack.
:- import_module exception.

:- func tab_width = int.
tab_width = 8.

:- type parser(T, S)
    --->    parser(
                parser_stream :: list(T),
                parser_context :: context,
                parser_state :: S
            ).

new_parser(Stream, Filename, InitState, Parser) :-
    Parser = parser(Stream, context(Filename, 1,1, stack.init), InitState).
new_parser(Stream, Filename, Parser) :-
    new_parser(Stream, Filename, unit, Parser).
new_parser(Stream, Parser) :-
    new_parser(Stream, "", unit, Parser).

stream_empty(P@parser([],_,_), P).

get_context(P ^ parser_context, P, P).

get_state(P ^ parser_state, P, P).
set_state(S, parser(T,C,_), parser(T,C,S)).

token_prim(NextPos, Tok, parser([Tok|Rest], Ctx0, S), parser(Rest, Ctx, S)) :-
    % The head did all the work getting the token and updating the stream.
    % Just need to update Pos.
    NextPos(Tok, Ctx0, Ctx).

many(Parser) -->
    (Parser ->
        many(Parser)
    ;
        {true}
    ).

many(Parser, Vals) -->
    (Parser(V) ->
        many(Parser, Vs),
        {Vals = [V|Vs]}
    ;
        {Vals = []}
    ).

many1(Parser) -->
    Parser,             % semidet
    many(Parser).       % det

many1(Parser, Vals) -->
    Parser(V),          % semidet
    many(Parser, Vs),   % det
    {Vals = [V|Vs]}.

sepby(Parser, Sep, ParserDesc, TokenToString) -->
    ( sepby1(Parser, Sep, ParserDesc, TokenToString) ->
        {true}
    ;
        {true}
    ).

sepby1(Parser, Sep, ParserDesc, TokenToString) -->
    ( Parser ->
        ( Sep ->
            ( sepby1(Parser, Sep, ParserDesc, TokenToString) ->
                {true}
            ;
                throw_expected(ParserDesc, TokenToString)
            )
        ;
            {true}
        )
    ;
        {fail}
    ).

sepby(Parser, Sep, ParserDesc, TokenToString, Vals) -->
    ( sepby1(Parser, Sep, ParserDesc, TokenToString, Vals_) ->
        {Vals = Vals_}
    ;
        {Vals = []}
    ).

sepby1(Parser, Sep, ParserDesc, TokenToString, Vals) -->
    ( Parser(Val) ->
        ( Sep ->
            ( sepby1(Parser, Sep, ParserDesc, TokenToString, Vals_) ->
                {Vals = [Val|Vals_]}
            ;
                throw_expected(ParserDesc, TokenToString)
            )
        ;
            {Vals = [Val]}
        )
    ;
        {fail}
    ).

% -------------------------------------------------------------------------- %

% Specific to character parsers.

% NextPos function for char parsers. Reliably updates the source file position
% based on the character seen.
:- pred char_nextpos(char::in, context::in, context::out) is det.
char_nextpos(C, Ctx0, Ctx) :-
    ( char_nextpos_(C, Ctx0, Ctx1) ->
        Ctx = Ctx1
    ; char.to_int(C) >= 0x20 ->
        % All characters < 0x20 are 0, >= 0x20 are 1.
        Ctx0 = context(Fname,Line,Col,Blocks),
        Ctx = context(Fname,Line,Col+1,Blocks)
    ;
        Ctx = Ctx0
    ).
% char_nextpos handles all the special case characters
:- pred char_nextpos_(char::in, context::in, context::out) is semidet.
% backspace: -1 column
char_nextpos_('\b', context(F, L,C0, B), context(F, L, C, B)) :-
    (C0 > 1 ->
        C = C0 - 1
    ;
        C = C0
    ).
% tab: Up to tab_width columns
char_nextpos_('\t', context(F, L, C0, B), context(F, L, C, B)) :-
    C = C0 + (tab_width - (C0 mod tab_width)).
% newline: +1 line, 0 column
char_nextpos_('\n', context(F, L0, _, B), context(F, L0+1, 1, B)).
% vertical tab: +1 line
char_nextpos_('\v', context(F, L0, C, B), context(F, L0+1, C, B)).
% form feed: +1 line
char_nextpos_('\f', context(F, L0, C, B), context(F, L0+1, C, B)).
% carriage return: 0 column
char_nextpos_('\r', context(F, L, _, B), context(F, L, 1, B)).

char(Chr) -->
    token_prim(char_nextpos, Chr).

string(Str) -->
    {string.to_char_list(Str, Chrs)},
    {GetChar = (pred(C::in, P0::in, P::out) is semidet :- char(C, P0, P))},
    list.foldl(GetChar, Chrs).

%% Special parsers for generic string productions %%

alpha(Chr) -->
    parse.char(Chr),
    {char.is_alpha(Chr)}.

alpha_or_underscore(Chr) -->
    parse.char(Chr),
    {char.is_alpha_or_underscore(Chr)}.

alnum(Chr) -->
    parse.char(Chr),
    {char.is_alnum(Chr)}.

alnum_or_underscore(Chr) -->
    parse.char(Chr),
    {char.is_alnum_or_underscore(Chr)}.

digit(Chr) -->
    parse.char(Chr),
    {char.is_digit(Chr)}.

hex_digit(Chr) -->
    parse.char(Chr),
    {char.is_hex_digit(Chr)}.

% -------------------------------------------------------------------------- %

% Context and error handling

context_push_block(Name, parser(Stream, Context0, State),
                         parser(Stream, Context, State)) :-
    context.context_push_block(Name, Context0, Context).

context_pop_block(parser(Stream, Context0, State),
                         parser(Stream, Context, State)) :-
    context.context_pop_block(Context0, Context).

throw_error(Msg, Parser, _) :-
    context.throw_error("Syntax error: " ++ Msg, Parser ^ parser_context).

throw_unexpected(TokenToString, Parser, _) :-
    ( Parser = parser([NextToken|_],_,_) ->
        throw_error("Unexpected " ++ TokenToString(NextToken), Parser, _)
    ;
        % End of stream
        throw_error("Unexpected end of file", Parser, _)
    ).

throw_expected(What, TokenToString, Parser, _) :-
    ( Parser = parser([NextToken|_],_,_) ->
        throw_error("Unexpected " ++ TokenToString(NextToken) ++ " (expected "
                        ++ What ++ ")", Parser, _)
    ;
        % End of stream
        throw_error("Unexpected end of file (expected " ++ What ++ ")",
                        Parser, _)
    ).

expect(P, What, TokenToString) -->
    ( P ->
        {true}
    ;
        throw_expected(What, TokenToString)
    ).

expect(P, What, TokenToString, Val) -->
    ( P(Val_) ->
        {Val = Val_}
    ;
        throw_expected(What, TokenToString)
    ).
