% -------------------------------------------------------------------------- %
% Mars/Venus parser
% Copyright (C) 2008 Matt Giuca
% 28/5/2008
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- module parsem.

:- interface.

:- import_module string.
:- import_module list.
:- import_module varset.

:- import_module ir.
:- import_module types.
:- import_module token.

% Converts a token stream into a program.
% Throws a parse error exception on syntax error.
% First arg is Filename.
:- pred parsem.parse_prog(string::in, list(token)::in, program::out) is det.
:- pred parsem.parse_prog(list(token)::in, program::out) is det.

% Converts a token stream into a statement.
% Throws a parse error exception on syntax error.
% First arg is Filename.
:- pred parsem.parse_stmt(string::in, list(token)::in, stmt::out) is det.
:- pred parsem.parse_stmt(list(token)::in, stmt::out) is det.

% Converts a token stream into an expression.
% Throws a parse error exception on syntax error.
% First arg is Filename.
:- pred parsem.parse_expr(string::in, list(token)::in, expr::out) is det.
:- pred parsem.parse_expr(list(token)::in, expr::out) is det.

% Converts a token stream into a type value.
% Throws a parse error exception on syntax error.
% First arg is Filename.
:- pred parsem.parse_type(string::in, list(token)::in,
        varset::out, typeval::out) is det.
:- pred parsem.parse_type(list(token)::in,
        varset::out, typeval::out) is det.

% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- implementation.

:- import_module char.
:- import_module bool.
:- import_module integer.
:- import_module maybe.
:- import_module pair.
:- import_module map.
:- import_module term.

:- import_module parse.

% Consumes 1 or more newlines.
:- pred newline(tokenparser::in, tokenparser::out) is semidet.
newline -->
    parse.many1(token(newline)).

% Consumes any number of newlines, followed by an indent.
:- pred indent(tokenparser::in, tokenparser::out) is semidet.
indent -->
    parse.many(token(newline)),
    token(indent).

% Consumes any number of newlines, followed by a dedent.
:- pred dedent(tokenparser::in, tokenparser::out) is semidet.
dedent -->
    parse.many(token(newline)),
    token(dedent).

% Consumes any number of newlines, followed by a dedent.
% Throws a parse error if no dedent found.
:- pred expect_dedent(tokenparser::in, tokenparser::out) is det.
expect_dedent -->
    parse.many(token(newline)),
    expect_token(dedent).

% Wrapper around parse.throw_expected, but doesn't ask for a function to
% show a token (just uses showToken).
:- pred throw_expected(string::in, tokenparser::in, tokenparser::out)
        is erroneous.
throw_expected(What) -->
    parse.throw_expected(What, showToken).

% Wrapper around parse.expect.
% expect(Parser, What, !P).
:- pred expect(pred(tokenparser, tokenparser), string,
                tokenparser, tokenparser).
:- mode expect(pred(in, out) is semidet, in, in, out) is det.
expect(Parser, What) -->
    parse.expect(Parser, What, showToken).
% expect(Parser, What, Val, !P).
:- pred expect(pred(U, tokenparser, tokenparser), string,
                U, tokenparser, tokenparser).
:- mode expect(pred(out, in, out) is semidet, in, out, in, out) is det.
expect(Parser, What, Val) -->
    parse.expect(Parser, What, showToken, Val).

% expect_token(Token, !Parser).
% Expects a single nullary token. (Consumes and parse error if not found).
:- pred expect_token(token::in, tokenparser::in, tokenparser::out) is det.
expect_token(Token) -->
    expect(token.token(Token), token.showToken(Token)).

% Wrapper around parse.sepby, but specific to comma-separated parsing, for
% tokens.
:- pred sepbycomma(pred(U, tokenparser, tokenparser), string,
                    list(U), tokenparser, tokenparser).
:- mode sepbycomma(pred(out, in, out) is semidet, in, out, in, out) is det.
sepbycomma(Parser, ParserDesc, Vals) -->
    parse.sepby(Parser, token(comma), ParserDesc, showToken, Vals).

% Converts a list of strings into a list of vars.
% Updates the given varset with names matching the string.
% Ensures that the names are unique (syntax error if not), and throws a parse
% error iff they aren't (which is why this takes a parser).
% Warning: O(N^2) in the length of the list. (Should be very small).
:- pred names_to_vars(list(string)::in, list(var)::out,
                        varset::in, varset::out,
                        parser(T,S)::in, parser(T,S)::out) is det.
names_to_vars([], [], !Varset, !Parser).
names_to_vars([S|Ss], [V|Vs], !Varset, !Parser) :-
    ( list.member(S, Ss) ->
        % Not a unique name - error
        parse.throw_error("Duplicate argument name \"" ++ S
                            ++ "\" in type definition", !Parser)
    ;
        true
    ),
    varset.new_named_var(!.Varset, S, V, !:Varset),
    names_to_vars(Ss, Vs, !Varset, !Parser).

% Predicates in this module have a 3-valued result logic.
% * Success means they finished parsing the production, and return a result.
% * Failure means they did not find the first token they were looking for,
%   hence they could not parse the production. However, the caller is free
%   to continue looking for other alternatives.
% * Throwing a parse error means they started parsing this production (so
%   they are committed), and then found a syntax error. The error is reported
%   and should not be caught except for printing.
% The way this is structured, you can ignore the possibility of syntax errors
% when calling a production, and simply account for failure, which means you
% may try another alternative or report an error if there are no more
% alternatives.

% Consumes the entire input sequence, and returns a program object.
:- pred program(program::out, tokenparser::in, tokenparser::out) is det.
program(program(Nodes)) -->
    % Consume useless newlines
    parse.many(token(newline)),
    ( parse.stream_empty ->
        % Base case - done reading program
        {Nodes = []}
    ;
        % Read a program node (reports its own errors, doesn't fail).
        ( program_node(Node) ->
            program(program(Rest)),
            {Nodes = [Node|Rest]}
        ;
            throw_expected("function or type definition, or import")
        )
    ).

% Consumes a single function or typedef, and returns a program_node.
:- pred program_node(program_node::out, tokenparser::in, tokenparser::out)
    is semidet.
program_node(Node) -->
    % All program nodes have a varset (which is threaded through the parsers
    % called from this predicate, and stored in the node object).
    ( typedef(Type) ->
        {Node = ptypedef(Type)}
    ; function(Func) ->
        {Node = pfunction(Func)}
    ; import(Import) ->
        {Node = pimport(Import)}
    ;
        {fail}
    ).

% Consumes a single import statement.
:- pred import(import::out, tokenparser::in, tokenparser::out) is semidet.
import(ModuleName - Context) -->
    parse.get_context(Context),
    token(import),      % Can fail
    ( token(uident(ModuleName_)) ->
        {ModuleName = ModuleName_}
    ; token(lident(ModuleName_)) ->
        {ModuleName = ModuleName_}
    ;
        throw_expected("identifier")
    ),
    expect(newline, "newline").

% Consumes a single typedef.
:- pred typedef(typedef::out, tokenparser::in, tokenparser::out) is semidet.
typedef(typedef(TypeName, Params, Ctors, Varset, Context)) -->
    parse.get_context(Context),
    token(type_t),          % Can fail
    ( token(uident(TypeName_)) ->
        {TypeName = TypeName_},
        parse.context_push_block(TypeName),
        ( token(openparen) ->
            sepbycomma(lident, "lowercase identifier", ParamNames),
            names_to_vars(ParamNames, YesParams, varset.init, Varset),
            {Params = yes(YesParams)},
            expect_token(closeparen)
        ;
            {Params = no},
            {varset.init(Varset)}
        ),
        expect_token(colon),
        expect(indent, "indent"),
        % Get 1 or more ctors
        % (Pass it a name/var map for the variables of this type)
        ( parse.many1(ctor(varset.create_name_var_map(Varset)), Ctors_) ->
            {Ctors = Ctors_}
        ;
            throw_expected("data constructor")
        ),
        expect(dedent, "data constructor or dedent"),
        parse.context_pop_block
    ;
        throw_expected("uppercase identifier")
    ).

% Consumes a ctor line.
% Takes a name/var map for type variables.
:- pred ctor(map(string, var)::in, ctor::out,
            tokenparser::in, tokenparser::out) is semidet.
ctor(VarNames, ctor(Name, Args, Context)) -->
    parse.get_context(Context),
    token(uident(Name)),
    ( token(openparen) ->
        sepbycomma(ctor_param(VarNames),
                    "lowercase identifier or type expression", YesArgs),
        expect_token(closeparen),
        {Args = yes(YesArgs)}
    ;
        {Args = no}
    ),
    expect(newline, "newline").

:- pred ctor_param(map(string, var)::in, pair(maybe(string), typeval)::out,
            tokenparser::in, tokenparser::out) is semidet.
ctor_param(VarNames, MaybeName - Type) -->
    ( token(lident(Name)), token(coloncolon) ->
        {MaybeName = yes(Name)},
        expect(typeexpr(VarNames), "type expression", Type)
    ;
        {MaybeName = no},
        typeexpr(VarNames, Type) % May fail
    ).

% func_arrow(UniqueInfo, !Parser).
% Consumes an arrow (->) token.
:- pred func_arrow(tokenparser::in, tokenparser::out) is semidet.
func_arrow -->
    token(arrow).

% amstring_to_argmode(string).
% Parses a string of uniqueness annotations and returns a uniqueinfo.
% Raises a parse error if it's invalid.
:- pred amstring_to_argmode(string::in, argmode::out,
    tokenparser::in, tokenparser::out) is det.
amstring_to_argmode(AMString, AM, !Parser) :-
    ( foldl(amchar_to_argmode, string.to_char_list(AMString),
        def_argmode, AM_) ->
        AM = AM_
    ;
        parse.throw_error("Invalid argument mode: `" ++ AMString
            ++ "`", !Parser)
    ).
:- pred amchar_to_argmode(char::in, argmode::in, argmode::out) is semidet.
amchar_to_argmode('k', !AM) :-
    !:AM = !.AM ^ arg_kill := yes.
amchar_to_argmode('i', !AM) :-
    !:AM = !.AM ^ arg_interest := yes.

% typeexpr(VarNames, Type, !Parser).
% Consumes a type expression.
% Accepts a variable name mapping so as to bind the variable correctly.
% Syntax error if the variable name does not exist in the mapping.
:- pred typeexpr(map(string, var)::in, typeval::out,
                tokenparser::in, tokenparser::out) is semidet.
typeexpr(VarNames, Type) -->
    typeexpr(yes, Type, varset.init, _, VarNames, _).

% typeexpr(Type, !Varset, !VarNames, !Parser).
% Consumes a type expression.
% Accepts a variable name mapping so as to bind the variable correctly.
% Updates Varset and VarNames with a new variable if a variable name does
% not exist in the mapping.
:- pred typeexpr(typeval::out, varset::in, varset::out,
                map(string, var)::in, map(string, var)::out,
                tokenparser::in, tokenparser::out) is semidet.
typeexpr(Type, !Varset, !VarNames) -->
    typeexpr(no, Type, !Varset, !VarNames).

% typeexpr(ErrorUndefVar, Type, !Varset, !VarNames, !Parser).
% Consumes a type expression.
% Accepts a variable name mapping so as to bind the variable correctly.
% If the variable name does not exist in the mapping:
% Syntax error if ErrorUndefVar is yes.
% Updates Varset and VarNames with a new variable if ErrorUndefVar is no.
:- pred typeexpr(bool::in, typeval::out, varset::in, varset::out,
                map(string, var)::in, map(string, var)::out,
                tokenparser::in, tokenparser::out) is semidet.
typeexpr(ErrorUndefVar, Type, !Varset, !VarNames) -->
    ( token(openparen) ->
        typeexprs(ErrorUndefVar, ArgTypes, !Varset, !VarNames),
        expect_token(closeparen),
        expect(func_arrow, "->"),
        ( typeexpr(ErrorUndefVar, RetType, !Varset, !VarNames) ->
            {Type = app(functype, ArgTypes ++ [RetType])}
        ;
            parse.throw_expected("type expression", showToken)
        )
    ; token(lident(Name)) ->
        ( {map.search(!.VarNames, Name, Var)} ->
            {Type0 = variable(Var)}
        ; {ErrorUndefVar = yes} ->
            parse.throw_error("Undefined type variable \"" ++ Name ++ "\"")
        ;
            {varset.new_named_var(!.Varset, Name, Var, !:Varset)},
            {map.det_insert(!.VarNames, Name, Var, !:VarNames)},
            {Type0 = variable(Var)}
        ),
        typeexpr_prime(ErrorUndefVar, Type0, Type, !Varset, !VarNames)
    ; token(uident(Name)) ->
        {Type0 = const(Name)},
        typeexpr_prime(ErrorUndefVar, Type0, Type, !Varset, !VarNames)
    ; func_arrow ->
        % Special type (->), without arguments.
        {Type0 = functype},
        typeexpr_prime(ErrorUndefVar, Type0, Type, !Varset, !VarNames)
    ;
        {fail}
    ).

:- pred typeexpr_prime(bool::in, typeval::in, typeval::out,
                varset::in, varset::out,
                map(string, var)::in, map(string, var)::out,
                tokenparser::in, tokenparser::out) is det.
typeexpr_prime(ErrorUndefVar, !Type, !Varset, !VarNames) -->
    ( func_arrow ->
        % Special syntax for unary function (a -> b)
        % !.Type is the type left of the arrow
        ( typeexpr(ErrorUndefVar, Type1_, !Varset, !VarNames) ->
            {Type1 = Type1_}
        ;
            parse.throw_expected("type expression", showToken)
        ),
        % Type1 is the type right of the arrow
        % Now we want !:Type = !.Type -> Type1.
        {!:Type = app(functype, [!.Type, Type1])}
    ; token(openparen) ->
        typeexprs(ErrorUndefVar, Args, !Varset, !VarNames),
        expect_token(closeparen),
        {!:Type = app(!.Type, Args)},
        typeexpr_prime(ErrorUndefVar, !Type, !Varset, !VarNames)
    ;
        {true}
    ).

% typeexprs(ErrorUndefVar, Types, !Varset, !VarNames, !Parser).
% Consumes 0 or more types separated by commas.
% Updates Varset and VarNames with a new variable if a variable name does
% not exist in the mapping.
:- pred typeexprs(bool::in, list(typeval)::out, varset::in, varset::out,
                map(string, var)::in, map(string, var)::out,
                tokenparser::in, tokenparser::out) is det.
typeexprs(ErrorUndefVar, Types, !Varset, !VarNames) -->
    ( typeexprs1(ErrorUndefVar, Types_, !Varset, !VarNames) ->
        {Types = Types_}
    ;
        {Types = []}
    ).

:- pred typeexprs1(bool::in, list(typeval)::out, varset::in, varset::out,
                map(string, var)::in, map(string, var)::out,
                tokenparser::in, tokenparser::out) is semidet.
typeexprs1(ErrorUndefVar, Types, !Varset, !VarNames) -->
    ( typeexpr(ErrorUndefVar, Type, !Varset, !VarNames) ->
        ( token(comma) ->
            ( typeexprs1(ErrorUndefVar, Types_, !Varset, !VarNames) ->
                {Types = [Type|Types_]}
            ;
                throw_expected("type expression", showToken)
            )
        ;
            {Types = [Type]}
        )
    ;
        {fail}
    ).

% Consumes a single function.
:- pred function(function::out, tokenparser::in, tokenparser::out) is semidet.
function(
    function(FuncName, Params, RetType, Varset, Body, Context)) -->
    some [!Varset, !VarNames] (
        {varset.init(!:Varset)},
        {map.init(!:VarNames)},
        token(def),     % Can fail
        ( token(lident(FuncName_)) ->
            {FuncName = FuncName_},
            parse.context_push_block(FuncName),
            parse.get_context(Context),
            % Optional paren and args
            ( token(openparen) ->
                % Get params, as well as a varset and name/var mapping for all the
                % type variables in the function head.
                func_params(YesParams, !Varset, !VarNames),
                expect(token(closeparen), "parameter or \")\""),
                {Params = yes(YesParams)}
            ;
                {Params = no}
            ),
            % Get Return Type
            expect_token(coloncolon),
            ( typeexpr(RetType_, !Varset, !VarNames) ->
                {RetType = RetType_}
            ;
                parse.throw_expected("type expression", showToken)
            ),
            % Two different syntaxes for functions
            ( token(colon) ->
                % Full decls/statement block form
                expect(indent, "indent"),
                % Get Body
                parse.many(decl(!.VarNames), Decls),
                expect(parse.many1(stmt), "statement", Stmts),
                expect(dedent, "statement or dedent"),
                {Body = func_body_ast(Decls, Stmts)}
            ; token(eq) ->
                % Single expression form
                expect(expr, "expression", Expr),
                {Body = func_body_ast([],
                                      [compound_stmt(return(Expr), Context)])}
            ;
                parse.throw_expected("\":\" or \"=\"", showToken)
            ),
            parse.context_pop_block
        ;
            throw_expected("lowercase identifier")
        ),
        {Varset = !.Varset}
    ).

% func_param(Param, !Varset, !VarNames, !Parser).
% Consumes a function parameter (lident "::" typeval).
% Updates Varset and VarNames with a new variable if a variable name does
% not exist in the mapping.
:- pred func_param({string, typeval, argmode}::out,
            varset::in, varset::out,
            map(string, var)::in, map(string, var)::out,
            tokenparser::in, tokenparser::out) is semidet.
func_param({Name, Type, ArgMode}, !Varset, !VarNames) -->
    token(lident(Name)),        % Can fail
    expect_token(coloncolon),
    ( typeexpr(Type_, !Varset, !VarNames) ->
        {Type = Type_}
    ;
        parse.throw_expected("type expression", showToken)
    ),
    % Parse optional argmode annotation (eg. `d` for dead)
    ( token(backtick) ->
        expect(lident, "argument mode annotation", AMString),
        expect_token(backtick),
        amstring_to_argmode(AMString, ArgMode)
    ;
        {ArgMode = def_argmode}
    )
    .

% func_params(Params, !Varset, !VarNames, !Parser).
% Consumes 0 or more function parameters, separated by commas.
% Updates Varset and VarNames with a new variable if a variable name does
% not exist in the mapping.
:- pred func_params(list({string, typeval, argmode})::out,
                varset::in, varset::out,
                map(string, var)::in, map(string, var)::out,
                tokenparser::in, tokenparser::out) is det.
func_params(Params, !Varset, !VarNames) -->
    ( func_params1(Params_, !Varset, !VarNames) ->
        {Params = Params_}
    ;
        {Params = []}
    ).

:- pred func_params1(list({string, typeval, argmode})::out,
                varset::in, varset::out,
                map(string, var)::in, map(string, var)::out,
                tokenparser::in, tokenparser::out) is semidet.
func_params1(Params, !Varset, !VarNames) -->
    func_param(Param, !Varset, !VarNames),  % Can fail
    ( token(comma) ->
        ( func_params1(Params_, !Varset, !VarNames) ->
            {Params = [Param|Params_]}
        ;
            throw_expected("parameter", showToken)
        )
    ;
        {Params = [Param]}
    ).

% decl(VarNames, Decl, !Parser).
% Consumes a decl ("var" lident "::" typeval "nl").
% Syntax error if the variable name does not exist in the mapping.
:- pred decl(map(string, var)::in, pair(varname, typeval)::out,
            tokenparser::in, tokenparser::out) is semidet.
decl(VarNames, svname(Name) - Type) -->
    token(var),     % Can fail
    ( lident(Name_) ->
        {Name = Name_}
    ;
        parse.throw_expected("lowercase identifier", showToken)
    ),
    expect_token(coloncolon),
    ( typeexpr(VarNames, Type_) ->
        {Type = Type_}
    ;
        parse.throw_expected("type expression", showToken)
    ),
    expect(newline, "newline").

% stmt(Stmt, !Parser).
% Consumes a statement, including newline.
:- pred stmt(stmt::out, tokenparser::in, tokenparser::out) is semidet.
stmt(Stmt) -->
    parse.get_context(Context),
    ( basic_stmt(BStmt) ->
        {Stmt = basic_stmt(BStmt, Context)}
    ; compound_stmt(CStmt) ->
        {Stmt = compound_stmt(CStmt, Context)}
    ;
        {fail}
    ).

:- pred basic_stmt(basic_stmt::out, tokenparser::in, tokenparser::out)
    is semidet.
basic_stmt(Stmt) -->
    ( token(pass) ->
        {Stmt = pass},
        expect(newline, "newline")
    ; expr(expr(E,_)) ->
        ( {E = varref(Name)} ->
            ( token(eq) ->
                % assignment
                expect(expr, "expression", Expr),
                expect(newline, "newline"),
                {Stmt = assign(Name, Expr)}
            ;
                % Ordinary expression
                expect(newline, "\"=\" or newline"),
                {Stmt = eval(expr(E, no))}
            )
        ; {E = fieldref(BaseE, Field)} ->
            ( token(eq) ->
                % fieldset
                expect(expr, "expression", Expr),
                expect(newline, "newline"),
                {Stmt = fieldset(BaseE, Field, Expr)}
            ; token(coloneq) ->
                % fieldreplace
                expect(expr, "expression", Expr),
                expect(newline, "newline"),
                {Stmt = fieldreplace(BaseE, Field, Expr)}
            ;
                % Ordinary expression
                expect(newline, "\"=\", \":=\" or newline"),
                {Stmt = eval(expr(E, no))}
            )
        ;
            % Ordinary expression
            expect(newline, "newline"),
            {Stmt = eval(expr(E, no))}
        )
    ;
        {fail}
    ).

:- pred compound_stmt(compound_stmt::out, tokenparser::in, tokenparser::out)
    is semidet.
compound_stmt(Stmt) -->
    ( token(return) ->
        expect(expr, "expression", Expr),
        expect(newline, "newline"),
        {Stmt = return(Expr)}
    ; token(switch) ->
        expect(expr, "expression", Expr),
        expect_token(colon),
        expect(newline, "newline"),
        expect(indent, "indent"),
        expect(parse.many1(case), "case statement", Cases),
        expect(dedent, "case statement or dedent"),
        {Stmt = switch(Expr, Cases)}
    ; token(if_t) ->
        expect(expr, "expression", Cond),
        expect_token(colon),
        expect(newline, "newline"),
        expect(indent, "indent"),
        parse.many(stmt, ThenBlock),
        expect(dedent, "statement or dedent"),
        % Else is optional
        ( token(else_t) ->
            expect_token(colon),
            expect(newline, "newline"),
            expect(indent, "indent"),
            parse.many(stmt, ElseBlock),
            expect(dedent, "statement or dedent")
        ;
            {ElseBlock = []}
        ),
        {Stmt = if_then_else(Cond, ThenBlock, ElseBlock)}
    ; token(while) ->
        expect(expr, "expression", Cond),
        expect_token(colon),
        expect(newline, "newline"),
        expect(indent, "indent"),
        parse.many(stmt, Body),
        expect(dedent, "statement or dedent"),
        {Stmt = while(Cond, Body)}
    ;
        {fail}
    ).

% Converts a char to an integer, then wraps it as an intlit expression.
:- func char_to_intlit(char) = expr.
char_to_intlit(C) = expr(intlit(integer(char.to_int(C))), no).

% expr(Expr, !Parser).
% Consumes an expression.
:- pred expr(expr::out, tokenparser::in, tokenparser::out) is semidet.
expr(Expr) -->
    ( token(intlit(I)) ->
        {Expr0 = expr(intlit(I), no)}
    ; token(charlit(C)) ->
        {Expr0 = char_to_intlit(C)}
    ; token(stringlit(S)) ->
        {Ints = map(char_to_intlit, string.to_char_list(S))},
        {Expr0 = expr(arraylit(Ints), no)}
    ; token(opensquare) ->
        sepbycomma(expr, "expression", Elems),
        expect_token(closesquare),
        {Expr0 = expr(arraylit(Elems), no)}
    ; token(lident(S)) ->
        {Expr0 = expr(varref(svname(S)), no)}
    ; token(uident(S)) ->
        {Expr0 = expr(ctorref(S), no)}
    ; token(openparen) ->
        expect(expr, "expression", Expr0),
        expect_token(closeparen)
    ;
        {fail}
    ),
    expr_prime(Expr0, Expr).

:- pred expr_prime(expr::in, expr::out, tokenparser::in, tokenparser::out)
            is det.
expr_prime(!Expr) -->
    ( token(openparen) ->
        expr_list_dotdotdot(Args, Dots),
        ( {Dots = yes} ->
            % Partial application
            expect_token(closeparen),
            {!:Expr = expr(parapp(!.Expr, Args), no)}
        ;
            expect_token(closeparen),
            {!:Expr = expr(app(!.Expr, Args), no)}
        ),
        expr_prime(!Expr)
    ; token(dot) ->
        expect(lident, "lowercase identifier", Field),
        {!:Expr = expr(fieldref(!.Expr, Field), no)},
        expr_prime(!Expr)
    ;
        {true}
    ).

% Parses a comma-separated expression list, followed by an optional "..."
% token. Returns the list of expressions, and a bool, which is "yes" if the
% "..." was present, and "no" if it wasn't.
:- pred expr_list_dotdotdot(list(expr)::out, bool::out,
    tokenparser::in, tokenparser::out) is det.
expr_list_dotdotdot(Exprs, Dots) -->
    ( expr_list_dotdotdot1(Exprs_, Dots_) ->
        {Exprs = Exprs_, Dots = Dots_}
    ;
        {Exprs = [], Dots = no}
    ).

:- pred expr_list_dotdotdot1(list(expr)::out, bool::out,
    tokenparser::in, tokenparser::out) is semidet.
expr_list_dotdotdot1(Exprs, Dots) -->
    ( token(dotdotdot) ->
        {Exprs = [], Dots = yes}
    ; expr(E) ->
        ( token(comma) ->
            ( expr_list_dotdotdot1(Exprs0, Dots_) ->
                {Exprs = [E|Exprs0], Dots = Dots_}
            ;
                throw_expected("expression or \"...\"", showToken)
            )
        ;
            {Exprs = [E], Dots = no}
        )
    ;
        {fail}
    ).

:- pred case(case_stmt::out,
            tokenparser::in, tokenparser::out) is semidet.
case(case_stmt(Pattern, Stmts, Context)) -->
    parse.get_context(Context),
    token(case),        % Can fail
    expect(pattern, "pattern", Pattern),
    expect_token(colon),
    expect(newline, "newline"),
    expect(indent, "indent"),
    parse.many(stmt, Stmts),
    expect(dedent, "statement or dedent").

:- pred pattern(pattern::out, tokenparser::in, tokenparser::out) is semidet.
pattern(pattern(Pattern, no)) -->
    ( token(underscore) ->
        {Pattern = pat_any}
    ; token(lident(Name)) ->
        {Pattern = pat_var(svname(Name))}
    ; token(uident(Name)) ->
        ( token(openparen) ->
            sepbycomma(pattern, "pattern", YesArgs),
            expect_token(closeparen),
            {Args = yes(YesArgs)}
        ;
            {Args = no}
        ),
        {Pattern = pat_ctor(Name, Args)}
    ; token(intlit(I)) ->
        {Pattern = pat_intlit(I)}
    ; token(charlit(C)) ->
        {char.to_int(C, I)},
        {Pattern = pat_intlit(integer(I))}
    ;
        {fail}
    ).

% -------------------------------------------------------------------------- %

% Top-level exported predicates

parse_prog(Toks, Prog) :-
    parse_prog("", Toks, Prog).

parse_prog(Filename, Toks, Prog) :-
    parse.new_parser(Toks, Filename, Parser),
    program(Prog, Parser, _).
    % Note: Resulting parser guaranteed to be empty, no need to check.

parse_stmt(Toks, Stmt) :-
    parse_stmt("", Toks, Stmt).

parse_stmt(Filename, Toks, Stmt) :-
    some [!Parser] (
        parse.new_parser(Toks, Filename, !:Parser),
        expect(stmt, "statement", Stmt, !Parser),
        % Allow newlines at the end
        parse.many(token(newline), !Parser),
        % Make sure stream is empty
        ( parse.stream_empty(!Parser) ->
            true
        ;
            throw_expected("End of statement", !Parser)
        )
    ).

parse_expr(Toks, Expr) :-
    parse_expr("", Toks, Expr).

parse_expr(Filename, Toks, Expr) :-
    some [!Parser] (
        parse.new_parser(Toks, Filename, !:Parser),
        expect(expr, "expression", Expr, !Parser),
        % Allow newlines at the end
        parse.many(token(newline), !Parser),
        % Make sure stream is empty
        ( parse.stream_empty(!Parser) ->
            true
        ;
            throw_expected("End of expression", !Parser)
        )
    ).

parse_type(Toks, Varset, Type) :-
    parse_type("", Toks, Varset, Type).

parse_type(Filename, Toks, Varset, Type) :-
    some [!Parser] (
        parse.new_parser(Toks, Filename, !:Parser),
        ( typeexpr(Type_, varset.init, Varset_, map.init, _, !Parser) ->
            Type = Type_, Varset = Varset_
        ;
            throw_expected("type", !Parser)
        ),
        % Allow newlines at the end
        parse.many(token(newline), !Parser),
        % Make sure stream is empty
        ( parse.stream_empty(!Parser) ->
            true
        ;
            throw_expected("End of expression", !Parser)
        )
    ).
