% -------------------------------------------------------------------------- %
% Mars/Venus pretty-printer
% Copyright (C) 2008 Matt Giuca
% 28/5/2008
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- module pretty.

:- interface.

:- import_module string.
:- import_module pair.
:- import_module io.
:- import_module varset.

:- import_module ir.
:- import_module cfg.
:- import_module types.

% string_typeval(Type) displays a typeval as a string.
:- func string_typeval(varset, typeval) = string.

% string_kind(Kind) displays a kind as a string.
:- func string_kind(kind) = string.

% print_ctor(IndentLevel, Varset, Ctor, !IO) prints the ctor to output.
% Block level (expects to be at start of line, ends with a newline).
:- pred print_ctor(int::in, varset::in, ctor::in, io::di, io::uo) is det.

% print_typedef_head(TypeDef, !IO) prints the header of the typedef.
% This is used by code generators when generating comments.
% Does not print indent or newline.
:- pred print_typedef_head(typedef::in, io::di, io::uo) is det.

% print_typedef(IndentLevel, TypeDef, !IO) prints the typedef to output.
% Block level (expects to be at start of line, ends with a newline).
:- pred print_typedef(int::in, typedef::in, io::di, io::uo) is det.

% string_expr(Expr) displays an expression as a string.
:- func string_expr(expr) = string.

% string_pattern(Pattern) displays a pattern as a string.
:- func string_pattern(pattern) = string.

% print_decl(IndentLevel, Varset, Decl, !IO) prints the decl to output.
% Block level (expects to be at start of line, ends with a newline).
:- pred print_decl(int::in, varset::in, pair(varname, typeval)::in,
        io::di, io::uo) is det.

% print_stmt(IndentLevel, Stmt, !IO) prints the statement to output.
% Block level (expects to be at start of line, ends with a newline).
:- pred print_stmt(int::in, stmt::in, io::di, io::uo) is det.

% print_function_head(Func, !IO) prints the header of the function.
% This is used by code generators when generating comments.
% Does not print indent or newline.
:- pred print_function_head(function::in, io::di, io::uo) is det.

% print_function(IndentLevel, Func, !IO) prints the function to output.
% Block level (expects to be at start of line, ends with a newline).
:- pred print_function(int::in, function::in, io::di, io::uo) is det.

% print_cfg(IndentLevel, CFG, !IO) prints the CFG to output.
% Block level (expects to be at start of line, ends with a newline).
:- pred print_cfg(int::in, cfg(S)::in, io::di, io::uo) is det.

% print_program(IndentLevel, Prog, !IO) prints the program to output.
% Block level (expects to be at start of line, ends with a newline).
:- pred print_program(int::in, program::in, io::di, io::uo) is det.
% print_program(Prog, !IO) prints the program to output.
:- pred print_program(program::in, io::di, io::uo) is det.

% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- implementation.

:- import_module list.
:- import_module map.
:- import_module int.
:- import_module integer.
:- import_module bool.
:- import_module maybe.
:- import_module require.
:- import_module term.

% Number of spaces per indent
:- func indent_width = int.
indent_width = 4.

:- pred indent(int::in, io::di, io::uo) is det.
indent(Indent, !IO) :-
    io.write_string(string.duplicate_char(' ', Indent * indent_width), !IO).

:- func string_typevar(varset, var) = string.
string_typevar(Varset, V) = Str :-
    % If the variable is bound, get its bound value (recurse)
    ( types.search_var(Varset, V, T) ->
        Str = string_typeval(Varset, T)
    ;
        % Not bound, print out its name.
        Str = types.varset_lookup_name(Varset, V)
    ).

string_typeval(Varset, Type0) = Str :-
    types.deref(Varset, Type0, Type),
    ( Type = variable(V) ->
        % Free variable (deref guarantees this).
        Str = types.varset_lookup_name(Varset, V)
    ; Type = const(Name) ->
        Str = Name
    ; Type = functype ->
        Str = "->"
    ; Type = app(functype, ArgsPlusReturn0@[_|_]) ->
        % Special case for function type
        % Dereference all argument types and return type
        map(types.deref(Varset), ArgsPlusReturn0, ArgsPlusReturn),
        % Pull the return type out (it is the last element)
        list.split_last_det(ArgsPlusReturn, Args, Return),
        StrArgs0 = string.join_list(", ", map(string_typeval(Varset), Args)),
        ( list.length(Args) = 1 ->
            % Drop the parens iff there is exactly 1 non-function argument
            ( Args = [app(functype, _)] ->
                StrArgs = "(" ++ StrArgs0 ++ ")"
            ;
                StrArgs = StrArgs0
            )
        ;
            StrArgs = "(" ++ StrArgs0 ++ ")"
        ),
        Str = StrArgs ++ " -> " ++ string_typeval(Varset, Return)
    ; Type = app(Func, Args) ->
        ArgStrings = map(string_typeval(Varset), Args),
        Str = string_typeval(Varset, Func)
            ++ "(" ++ string.join_list(", ", ArgStrings) ++ ")"
    ; % Should not get here
        error("string_typeval: Invalid type value")
    ).

string_kind(kind_star) = "*".
string_kind(kind_arrow(Arity, IsVariadic)) = Str :-
    Stars0 = list.duplicate(Arity, "*"),
    ( IsVariadic = yes ->
        Stars = Stars0 ++ ["**"]
    ;
        Stars = Stars0
    ),
    ( list.length(Stars) = 1 ->
        Str = string.join_list(", ", Stars) ++ " -> *"
    ;
        Str = "(" ++ string.join_list(", ", Stars) ++ ") -> *"
    ).

:- func string_maybeparam(varset, pair(maybe(string), typeval)) = string.
string_maybeparam(Varset, yes(Name) - Type)
    = Name ++ " :: " ++ string_typeval(Varset, Type).
string_maybeparam(Varset, no - Type)
    = string_typeval(Varset, Type).

:- func string_param(varset, {string, typeval, argmode}) = string.
string_param(Varset, {Name, Type, ArgMode})
    = Name ++ " :: " ++ string_typeval(Varset, Type)
        ++ string_argmode(ArgMode).

:- func string_argmode(argmode) = string.
string_argmode(AM) = S :-
    ( AM = argmode(no, no) ->
        S = ""
    ;
        S = "`" ++ K ++ I ++ "`",
        ( AM ^ arg_kill = yes ->
            K = "k"
        ;
            K = ""
        ),
        ( AM ^ arg_interest = yes ->
            I = "i"
        ;
            I = ""
        )
    ).

:- func string_param_varname(varset, pair(varname, typeval)) = string.
string_param_varname(Varset, Name - Type)
    % Note: varname_to_string will escape '_' as "__", so this will modify
    % existing variable names.
    = ir.varname_to_string_noescape(Name)
        ++ " :: " ++ string_typeval(Varset, Type).

print_ctor(Indent, Varset, ctor(Name, Args, _), !IO) :-
    indent(Indent, !IO),
    io.write_string(Name, !IO),
    (
        Args = no
    ;
        Args = yes(YesArgs),
        io.write_string("(", !IO),
        io.write_string(string.join_list(", ",
                            list.map(string_maybeparam(Varset), YesArgs)),
                        !IO),
        io.write_string(")", !IO)
    ),
    io.nl(!IO).

print_typedef_head(typedef(Name, no, _Ctors, _Varset, _), !IO) :-
    io.write_string(Name, !IO).
print_typedef_head(typedef(Name, yes(Params), _Ctors, Varset, _), !IO) :-
    io.write_string(Name, !IO),
    io.write_string("(", !IO),
    io.write_string(string.join_list(", ",
                        list.map(string_typevar(Varset), Params)), !IO),
    io.write_string(")", !IO).

print_typedef(Indent, T@typedef(_Name, _Params, Ctors, Varset, _), !IO) :-
    indent(Indent, !IO),
    io.write_string("type ", !IO),
    print_typedef_head(T, !IO),
    io.write_string(":\n", !IO),
    foldl(print_ctor(Indent+1, Varset), Ctors, !IO),
    io.nl(!IO).

string_expr(Expr) = string_expr_(Expr ^ expr_expr).

:- func string_expr_(expr_) = string.
string_expr_(intlit(I)) = integer.to_string(I).
string_expr_(arraylit(A))
    = "[" ++ string.join_list(", ", list.map(string_expr, A)) ++ "]".
string_expr_(varref(Name)) = ir.varname_to_string_noescape(Name).
string_expr_(ctorref(Name)) = Name.
string_expr_(fieldref(Obj, Field)) = string_expr(Obj) ++ "." ++ Field.
string_expr_(app(Func, Args)) = string_expr(Func) ++ "(" ++ ArgsStr ++ ")" :-
    ArgsStr = string.join_list(", ", list.map(string_expr, Args)).
string_expr_(parapp(Func, Args)) = string_expr(Func) ++ "(" ++ ArgsStr
    ++ ")" :-
    ArgsStr = string.join_list(", ", list.map(string_expr, Args) ++ ["..."]).

print_stmt(Indent, basic_stmt(Stmt, _), !IO) :-
    print_basic_stmt(Indent, Stmt, !IO).
print_stmt(Indent, compound_stmt(Stmt, _), !IO) :-
    print_compound_stmt(Indent, Stmt, !IO).

:- pred print_basic_stmt(int::in, basic_stmt::in, io::di, io::uo) is det.
print_basic_stmt(Indent, pass, !IO) :-
    indent(Indent, !IO),
    io.write_string("pass\n", !IO).
print_basic_stmt(Indent, assign(Target, Expr), !IO) :-
    indent(Indent, !IO),
    io.write_string(ir.varname_to_string_noescape(Target), !IO),
    io.write_string(" = ", !IO),
    io.write_string(string_expr(Expr), !IO),
    io.nl(!IO).
print_basic_stmt(Indent, fieldset(Target, FieldName, Expr), !IO) :-
    indent(Indent, !IO),
    io.write_string(string_expr(Target), !IO),
    io.write_string(".", !IO),
    io.write_string(FieldName, !IO),
    io.write_string(" = ", !IO),
    io.write_string(string_expr(Expr), !IO),
    io.nl(!IO).
print_basic_stmt(Indent, fieldreplace(Target, FieldName, Expr), !IO) :-
    indent(Indent, !IO),
    io.write_string(string_expr(Target), !IO),
    io.write_string(".", !IO),
    io.write_string(FieldName, !IO),
    io.write_string(" := ", !IO),
    io.write_string(string_expr(Expr), !IO),
    io.nl(!IO).
print_basic_stmt(Indent, eval(Expr), !IO) :-
    indent(Indent, !IO),
    io.write_string(string_expr(Expr), !IO),
    io.nl(!IO).

:- pred print_compound_stmt(int::in, compound_stmt::in, io::di,io::uo) is det.
print_compound_stmt(Indent, return(Expr), !IO) :-
    indent(Indent, !IO),
    io.write_string("return ", !IO),
    io.write_string(string_expr(Expr), !IO),
    io.nl(!IO).
print_compound_stmt(Indent, switch(Expr, Cases), !IO) :-
    indent(Indent, !IO),
    io.write_string("switch ", !IO),
    io.write_string(string_expr(Expr), !IO),
    io.write_string(":\n", !IO),
    foldl(print_case(Indent+1), Cases, !IO).
print_compound_stmt(Indent, if_then_else(Cond, ThenPart, ElsePart), !IO) :-
    indent(Indent, !IO),
    io.write_string("if ", !IO),
    io.write_string(string_expr(Cond), !IO),
    io.write_string(":\n", !IO),
    foldl(print_stmt(Indent+1), ThenPart, !IO),
    ( list.is_empty(ElsePart) ->
        true
    ;
        indent(Indent, !IO),
        io.write_string("else:\n", !IO),
        foldl(print_stmt(Indent+1), ElsePart, !IO)
    ).
print_compound_stmt(Indent, while(Cond, Stmts), !IO) :-
    indent(Indent, !IO),
    io.write_string("while ", !IO),
    io.write_string(string_expr(Cond), !IO),
    io.write_string(":\n", !IO),
    foldl(print_stmt(Indent+1), Stmts, !IO).

:- pred print_instr(int::in, instr::in, io::di, io::uo) is det.
print_instr(Indent, Instr, !IO) :-
    print_instr_(Indent, Instr ^ instr_instr, !IO),
    io.nl(!IO).

:- pred print_instr_(int::in, instr_::in, io::di, io::uo) is det.
print_instr_(Indent, nop, !IO) :-
    indent(Indent, !IO),
    io.write_string("nop", !IO).
print_instr_(Indent, mov(Dst, Src), !IO) :-
    indent(Indent, !IO),
    io.write_string(ir.varname_to_string_noescape(Dst), !IO),
    io.write_string(" = ", !IO),
    io.write_string(ir.varname_to_string_noescape(Src), !IO).
print_instr_(Indent, i_fieldset(Target, FieldName, Expr), !IO) :-
    indent(Indent, !IO),
    io.write_string(ir.varname_to_string_noescape(Target), !IO),
    io.write_string(".", !IO),
    io.write_string(FieldName, !IO),
    io.write_string(" = ", !IO),
    io.write_string(ir.varname_to_string_noescape(Expr), !IO).
print_instr_(Indent, i_fieldreplace(Dst, Target, FieldName, Expr), !IO) :-
    indent(Indent, !IO),
    io.write_string(ir.varname_to_string_noescape(Dst), !IO),
    io.write_string(" = ", !IO),
    io.write_string(ir.varname_to_string_noescape(Target), !IO),
    io.write_string(".", !IO),
    io.write_string(FieldName, !IO),
    io.write_string(" := ", !IO),
    io.write_string(ir.varname_to_string_noescape(Expr), !IO).
print_instr_(Indent, ld_intlit(Dst, I), !IO) :-
    indent(Indent, !IO),
    io.write_string(ir.varname_to_string_noescape(Dst), !IO),
    io.write_string(" = ", !IO),
    io.write_string(integer.to_string(I), !IO).
print_instr_(Indent, ld_arraylit(Dst, Elems), !IO) :-
    indent(Indent, !IO),
    io.write_string(ir.varname_to_string_noescape(Dst), !IO),
    io.write_string(" = ", !IO),
    io.write_string("[" ++ string.join_list(", ",
        list.map(ir.varname_to_string_noescape, Elems)) ++ "]", !IO).
print_instr_(Indent, ld_ctor(Dst, Name), !IO) :-
    indent(Indent, !IO),
    io.write_string(ir.varname_to_string_noescape(Dst), !IO),
    io.write_string(" = ", !IO),
    io.write_string(Name, !IO).
print_instr_(Indent, ld_field(Dst, Obj, Field), !IO) :-
    indent(Indent, !IO),
    io.write_string(ir.varname_to_string_noescape(Dst), !IO),
    io.write_string(" = ", !IO),
    io.write_string(ir.varname_to_string_noescape(Obj) ++ "." ++ Field, !IO).
print_instr_(Indent, call(Dst, Func, Args), !IO) :-
    indent(Indent, !IO),
    io.write_string(ir.varname_to_string_noescape(Dst), !IO),
    io.write_string(" = ", !IO),
    ArgsStr = string.join_list(", ",
        list.map(ir.varname_to_string_noescape, Args)),
    io.write_string(
        ir.varname_to_string_noescape(Func) ++ "(" ++ ArgsStr ++ ")", !IO).
print_instr_(Indent, parcall(Dst, Func, Args), !IO) :-
    indent(Indent, !IO),
    io.write_string(ir.varname_to_string_noescape(Dst), !IO),
    io.write_string(" = ", !IO),
    ArgsStr = string.join_list(", ",
        list.map(ir.varname_to_string_noescape, Args) ++ ["..."]),
    io.write_string(
        ir.varname_to_string_noescape(Func) ++ "(" ++ ArgsStr ++ ")", !IO).

:- pred print_terminator_instr(int::in, cfg(S)::in, terminator_instr(S)::in,
    io::di,io::uo) is det.
print_terminator_instr(Indent, CFG, t_switch(Control, Cases), !IO) :-
    indent(Indent, !IO),
    io.write_string("switch ", !IO),
    io.write_string(ir.varname_to_string_noescape(Control), !IO),
    io.write_string(":\n", !IO),
    foldl(print_cfg_case(Indent+1, CFG), Cases, !IO).
print_terminator_instr(Indent, CFG, cond_branch(Cond, ThenPart, ElsePart),
    !IO) :-
    indent(Indent, !IO),
    io.write_string("if ", !IO),
    io.write_string(ir.varname_to_string_noescape(Cond), !IO),
    io.write_string(" goto ", !IO),
    io.write(cfg.ref_id(ThenPart, CFG), !IO),
    io.write_string(" else goto ", !IO),
    io.write(cfg.ref_id(ElsePart, CFG), !IO),
    io.nl(!IO).
print_terminator_instr(Indent, CFG, branch(Target), !IO) :-
    indent(Indent, !IO),
    io.write_string("goto ", !IO),
    io.write(cfg.ref_id(Target, CFG), !IO),
    io.nl(!IO).
print_terminator_instr(Indent, _CFG, null_terminator, !IO) :-
    indent(Indent, !IO),
    io.write_string("# Null block terminator.\n", !IO).

:- pred print_case(int::in, case_stmt::in, io::di, io::uo) is det.
print_case(Indent, case_stmt(Pattern, Stmts, _), !IO) :-
    indent(Indent, !IO),
    io.write_string("case ", !IO),
    io.write_string(string_pattern(Pattern), !IO),
    io.write_string(":\n", !IO),
    foldl(print_stmt(Indent+1), Stmts, !IO).

:- pred print_cfg_case(int::in, cfg(S)::in, cfg_case_stmt(S)::in,
    io::di, io::uo) is det.
print_cfg_case(Indent, CFG, cfg_case_stmt(Pattern, Target, _), !IO) :-
    indent(Indent, !IO),
    io.write_string("case ", !IO),
    io.write_string(string_pattern(Pattern), !IO),
    io.write_string(" goto ", !IO),
    io.write(cfg.ref_id(Target, CFG), !IO),
    io.nl(!IO).

string_pattern(pattern(P,_)) = string_pattern_(P).

:- func string_pattern_(pattern_) = string.
string_pattern_(pat_any) = "_".
string_pattern_(pat_var(Name)) = varname_to_string_noescape(Name).
string_pattern_(pat_ctor(Name, Args)) = Name ++ ArgString :-
    (
        Args = no,
        ArgString = ""
    ;
        Args = yes(YesArgs),
        ArgString = "("
                    ++ string.join_list(", ",
                        list.map(string_pattern, YesArgs))
                    ++ ")"
    ).
string_pattern_(pat_intlit(I)) = integer.to_string(I).

print_function_head(Func, !IO) :-
    io.write_string(Func^func_name, !IO),
    (
        Func^func_params = no
    ;
        Func^func_params = yes(Params),
        io.write_string("(", !IO),
        io.write_string(string.join_list(", ",
                            list.map(string_param(Func^func_varset), Params)),
                        !IO),
        io.write_string(")", !IO)
    ),
    io.write_string(" :: ", !IO),
    io.write_string(string_typeval(Func^func_varset,Func^func_ret_type), !IO).

print_function(Indent, Func, !IO) :-
    indent(Indent, !IO),
    io.write_string("def ", !IO),
    print_function_head(Func, !IO),
    io.write_string(":\n", !IO),
    (
        Func^func_body = func_builtin
        % TODO: Builtin functions
    ;
        Func^func_body = func_body_ast(Decls, Stmts),
        foldl(print_decl(Indent+1, Func^func_varset), Decls, !IO),
        foldl(print_stmt(Indent+1), Stmts, !IO)
    ;
        Func^func_body = func_body_cfg(Decls, CFG),
        foldl(print_decl(Indent+1, Func^func_varset), Decls, !IO),
        indent(Indent+1, !IO),
        io.write_string("# XXX CFG representation (not valid Mars code)\n",
                        !IO),
        print_cfg(Indent+1, CFG, !IO)
    ),
    io.nl(!IO).

print_cfg(Indent, CFG, !IO) :-
    RetV = cfg.get_retv(CFG),
    indent(Indent, !IO),
    io.write_string("# Return variable: " ++
        ir.varname_to_string_noescape(RetV) ++ "\n", !IO),
    cfg.foldl_cfg(print_block(Indent, CFG), CFG, !IO).

:- pred print_block(int::in, cfg(S)::in, bbref(S)::in, basic_block(S)::in,
    io::di, io::uo) is det.
print_block(Indent, CFG, BBRef, Block, !IO) :-
    indent(Indent, !IO),
    io.write_string("Block ", !IO),
    io.write(cfg.ref_id(BBRef, CFG), !IO),
    io.write_string(":\n", !IO),
    indent(Indent+1, !IO),
    io.write_string("# Preds: ", !IO),
    list.map((pred(A0::in, A::out) is det :- A = ref_id(A0, CFG)),
        Block ^ bb_predecessors, PredNums),
    io.write(PredNums, !IO),
    io.nl(!IO),
    list.foldl(print_phi(Indent+1, CFG), Block ^ bb_phis, !IO),
    cfg.foldl_body(print_instr(Indent+1), Block, !IO),
    print_terminator_instr(Indent+1, CFG, Block ^ bb_terminator, !IO).

:- pred print_phi(int::in, cfg(S)::in, phi(S)::in, io::di, io::uo) is det.
print_phi(Indent, CFG, Phi, !IO) :-
    indent(Indent, !IO),
    io.write_string("PHI ", !IO),
    io.write_string(varname_to_string_noescape(Phi ^ phi_varname), !IO),
    io.write_string(" ", !IO),
    map.foldr((pred(A0::in, B0::in, L0::in, L::out) is det :-
            A = ref_id(A0, CFG), B = ir.varname_to_string_noescape(B0),
            L = [(A-B):pair(int,string)|L0]),
        Phi ^ phi_preds, [], PredNums),
    io.write(PredNums, !IO),
    io.nl(!IO).

print_decl(Indent, Varset, Decl, !IO) :-
    indent(Indent, !IO),
    io.write_string("var ", !IO),
    io.write_string(string_param_varname(Varset, Decl), !IO),
    io.nl(!IO).

:- pred print_import(int::in, import::in, io::di, io::uo) is det.
print_import(Indent, Module - _, !IO) :-
    indent(Indent, !IO),
    io.write_string("import ", !IO),
    io.write_string(Module, !IO),
    io.write_string("\n\n", !IO).

:- pred print_program_node(int::in, program_node::in, io::di, io::uo) is det.
print_program_node(Indent, ptypedef(TypeDef), !IO) :-
    print_typedef(Indent, TypeDef, !IO).
print_program_node(Indent, pfunction(Function), !IO) :-
    print_function(Indent, Function, !IO).
print_program_node(Indent, pimport(Import), !IO) :-
    print_import(Indent, Import, !IO).

print_program(Program, !IO) :-
    print_program(0, Program, !IO).
print_program(Indent, program(Nodes), !IO) :-
    foldl(print_program_node(Indent), Nodes, !IO).
