% -------------------------------------------------------------------------- %
% Mars/Venus source file reader
% Copyright (C) 2008 Matt Giuca
% 9/6/2008
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- module reader.

:- interface.

:- import_module io.
:- import_module string.
:- import_module maybe.

:- import_module ir.

% read_source_file(Filename, Program, !IO).
% Opens the file of the given name, parses it, and returns the parsed program.
% Note that this will recursively load all imported files, so the resulting
% program will not have any import statements, but it will contain the nodes
% of all imported modules.
% Throws context.errors if there is a parse error or a file cannot be opened.
:- pred read_source_file(string::in, program::out, io::di, io::uo)
        is det.

% locate_module(ModuleName, Filename, !IO).
% Given the name of a module or filename, find a source file.
% Will search all paths on the environment variable MARSPATH.
% Returns Filename = no if no file can be found matching the module name.
:- pred locate_module(string::in, maybe(string)::out, io::di, io::uo) is det.

:- implementation.

:- import_module char.
:- import_module list.
:- import_module set.
:- import_module pair.
:- import_module dir.

:- import_module token.
:- import_module parsem.
:- import_module context.

% Default search locations, excluding "."
:- func default_search_paths = list(string).
default_search_paths = ["/usr/lib/mars", "/usr/local/lib/mars"].

% read_source_file(Context, PathList, Filename, Program, !IO).
% (Also takes a Context, of the file which is importing, and a PathList).
:- pred read_source_file(context.context::in, list(string)::in,
        string::in, program::out, set(string)::in, set(string)::out,
        io::di, io::uo) is det.

read_source_file(Filename, Program, !IO) :-
    % Get the MARSPATH from the environment
    get_marspath(PathList, !IO),
    set.init(SeenSet),
    read_source_file(context.blank_context, PathList, Filename, Program,
                     SeenSet, _, !IO).
read_source_file(Context, PathList, Filename, Program, !SeenSet, !IO) :-
    % Open the filename given (a Mars source file)
    % SeenSet is a set of all modules already imported or being imported
    io.open_input(Filename, OpenRes, !IO),
    (
        OpenRes = ok(InFile),
        io.read_file(InFile, Res, !IO),
        (
            Res = ok(Text)
        ;
            Res = error(Text, Err),
            io.write_string(io.error_message(Err), !IO)
        ),
        % Convert the text into a stream of tokens
        token.tokenize(Filename, Text, Tokens),         % Can throw
        % Convert the stream of tokens into a program
        parsem.parse_prog(Filename, Tokens, Program0),  % Can throw
        io.close_input(InFile, !IO),
        % Read all the imported files and bring them into the program
        do_imports(PathList, Program0, Program, !SeenSet, !IO)
    ;
        OpenRes = error(_),
        context.throw_error("Can't open file: " ++ Filename, Context)
    ).

% do_imports(!Program, !IO).
% Transitive closure of program imports. Recursively imports all modules
% found in "import" statements in the program.
% The resulting program is complete, with all imported nodes, and contains no
% import statements.
:- pred do_imports(list(string)::in, program::in, program::out,
        set(string)::in, set(string)::out, io::di, io::uo) is det.
do_imports(PathList, program(Nodes0), program(Nodes), !SeenSet, !IO) :-
    do_imports_nodes(PathList, Nodes0, Nodes, !SeenSet, !IO).

:- pred do_imports_nodes(list(string)::in, list(program_node)::in,
        list(program_node)::out, set(string)::in, set(string)::out,
        io::di, io::uo) is det.
do_imports_nodes(_PathList, [], [], !SeenSet, !IO).
do_imports_nodes(PathList, [N0|Ns0], N ++ Ns, !SeenSet, !IO) :-
    do_imports_node(PathList, N0, N, !SeenSet, !IO),
    do_imports_nodes(PathList, Ns0, Ns, !SeenSet, !IO).

:- pred do_imports_node(list(string)::in, program_node::in,
        list(program_node)::out, set(string)::in, set(string)::out,
        io::di, io::uo) is det.
do_imports_node(PathList, Node, Nodes, !SeenSet, !IO) :-
    ( Node = pimport(ModuleName - Context) ->
        ( set.member(ModuleName, !.SeenSet) ->
            % We've already imported or are currently importing this module.
            % Silently drop this import statement
            % (This allows for multiple files importing the same module, as
            % well as cyclic imports).
            Nodes = []
        ;
            % Cache the name of this module to avoid trying to import it again
            set.insert(!.SeenSet, ModuleName, !:SeenSet),
            % Find the file for this module
            locate_module(PathList, ModuleName, MaybeFilename, !IO),
            (
                MaybeFilename = yes(Filename),
                % Import node - load the module and replace this node with the
                % full program imported.
                read_source_file(Context, PathList, Filename, program(Nodes),
                                 !SeenSet, !IO)
            ;
                MaybeFilename = no,
                % Could not find a file corresponding to the module
                context.throw_error("Can't find module named \""
                    ++ ModuleName ++ "\"", Context)
            )
        )
    ;
        % Some other node, just copy it into the output
        Nodes = [Node]
    ).

locate_module(ModuleName, Filename, !IO) :-
    get_marspath(PathList, !IO),
    locate_module(PathList, ModuleName, Filename, !IO).

% locate_module(SearchPaths, ModuleName, Filename, !IO).
% Given the name of a module or filename, find a source file.
% SearchPaths is a list of absolute or relative pathnames on which to search.
% Must include ".", or the current directory won't be searched.
% Returns Filename = no if no file can be found matching the module name.
:- pred locate_module(list(string)::in, string::in, maybe(string)::out,
                io::di, io::uo) is det.
locate_module([], _ModuleName, no, !IO).
locate_module([Path|Paths], ModuleName, Filename, !IO) :-
    TryFilename = dir.make_path_name(Path, ModuleName ++ ".mar"),
    io.check_file_accessibility(TryFilename, [io.read], Res, !IO),
    (
        Res = ok,
        Filename = yes(TryFilename)
    ;
        Res = error(_),
        locate_module(Paths, ModuleName, Filename, !IO)
    ).

% get_marspath(PathList, !IO).
% Retrieves a list of module search paths, by looking for the environment
% variable MARSPATH (colon-separated list of paths).
% Always appends "." to the path if it isn't there already, along with other
% default search locations. ("." is appended to the front, other defaults
% appended to the back).
% If not found, just returns the default search locations.
:- pred get_marspath(list(string)::out, io::di, io::uo) is det.
get_marspath(PathList, !IO) :-
    io.get_environment_var("MARSPATH", MaybeVal, !IO),
    (
        MaybeVal = yes(Val),
        PathList0 = string.words_separator((pred(C::in) is semidet :-
                                    (char.to_int(C) = 0x3a)),   % ':'
                                 Val),
        % Append "." to the front, if not there already
        ( PathList0 = [dir.this_directory|_] ->
            PathList = PathList0 ++ default_search_paths
        ;
            PathList = [dir.this_directory|PathList0] ++ default_search_paths
        )
    ;
        MaybeVal = no,
        PathList = [dir.this_directory] ++ default_search_paths
    ).
