% -------------------------------------------------------------------------- %
% Mars/Venus symbol table data structures
% Copyright (C) 2008 Matt Giuca
% 1/6/2008
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

% This module extends IR by having symbol table data structures, for
% efficient searching.

:- module tables.

:- interface.

:- import_module ir.
:- import_module types.
:- import_module map.
:- import_module pair.

% ProgTable: A program (collection of types and functions), but stored
% without any ordering in an efficiently searchable structure.
:- type progtable
    --->    progtable(
                prog_types :: map(string, pair(typedef, fieldtable)),
                prog_ctors :: map(string, typedef),
                prog_funcs :: map(string, function)
            ).

% FieldTable: A symbol table for a typedef, mapping field names to types.
% Allows efficient searching for field names within a typedef.
:- type fieldtable == map(string, typeval).

% LocalTable: A local symbol table (mapping variable names to types).
% Note that this does not store a varset (which is needed to make use of
% the types).
:- type localtable == map(varname, typeval).

:- func empty_progtable = progtable.
:- func empty_localtable = localtable.

% build_progtable(Prog, !ProgTable) builds a program table from a program.
% Throws a context.error if duplicate names are found.
:- pred build_progtable(program::in, progtable::in, progtable::out) is det.
:- pred build_progtable(program::in, progtable::out) is det.
:- pred update_progtable_typedef(typedef::in, progtable::in, progtable::out)
        is det.
:- pred update_progtable_function(function::in, progtable::in, progtable::out)
        is det.

% build_localtable(Func, !LocalTable) builds a local table from a function.
% Throws a context.error if duplicate names are found.
:- pred build_localtable(function::in, localtable::in, localtable::out)
        is det.
:- pred build_localtable(function::in, localtable::out) is det.
% update_localtable_string(VarName_String, VarType, !LocalTable).
% Inserts VarName-VarType as a new pair in the table.
% If VarName already exists, overwrites.
:- pred update_localtable_string(string::in, typeval::in,
                          localtable::in, localtable::out) is det.
% update_localtable(VarName, VarType, !LocalTable).
% Inserts VarName-VarType as a new pair in the table.
% If VarName already exists, overwrites.
:- pred update_localtable(varname::in, typeval::in,
                          localtable::in, localtable::out) is det.

% lookup_type(ProgTable, Name, TypeDef, FieldTable).
% Gets the typedef and fieldtable associated with a particular type name.
% Note: This even works for built-in types "Array" and "Int".
% However, there is no table entry for the built-in type "->" (function). This
% must be handled specially.
:- pred lookup_type(progtable::in, string::in, typedef::out, fieldtable::out)
            is semidet.
% lookup_type(ProgTable, Name, TypeDef).
% Gets the typedef associated with a particular type name.
% Note: This even works for built-in types "Array" and "Int".
% However, there is no table entry for the built-in type "->" (function). This
% must be handled specially.
:- pred lookup_type(progtable::in, string::in, typedef::out)
            is semidet.
% lookup_ctor(ProgTable, Name, TypeDef).
% Gets the typedef for the constructor associated with a particular ctor name,
% and the ctor itself.
:- pred lookup_ctor(progtable::in, string::in, typedef::out, ctor::out)
    is semidet.
% lookup_function(ProgTable, Name, Function).
% Gets the function associated with a particular global variable name.
:- pred lookup_function(progtable::in, string::in, function::out) is semidet.
% lookup_field(FieldTable, Name, Type).
% Gets the type of a field in a field table.
:- pred lookup_field(fieldtable::in, string::in, typeval::out) is semidet.
% lookup_field(ProgTable, TypeName, FieldName, Type).
% Gets the type of a field associated with a particular typedef.
:- pred lookup_field(progtable::in, string::in, string::in, typeval::out)
            is semidet.
% lookup_local_string(LocalTable, Name_String, Type).
% Gets the type of a local variable associated with a particular name.
:- pred lookup_local_string(localtable::in, string::in, typeval::out)
            is semidet.
% lookup_local(LocalTable, Name, Type).
% Gets the type of a local variable associated with a particular name.
:- pred lookup_local(localtable::in, varname::in, typeval::out) is semidet.

:- implementation.

:- import_module list.
:- import_module maybe.
:- import_module string.
:- import_module exception.

:- import_module context.

empty_progtable = progtable(map.init, map.init, map.init).
empty_localtable = map.init.

build_progtable(program(Nodes), !Table) :-
    list.foldl(add_to_progtable, Nodes, !Table).
build_progtable(program(Nodes), Table) :-
    build_progtable(program(Nodes), empty_progtable, Table).

:- pred add_to_progtable(program_node::in, progtable::in, progtable::out)
        is det.
add_to_progtable(ptypedef(TypeDef), progtable(Types0, Ctors0, Funcs),
                                    progtable(Types, Ctors, Funcs)) :-
    % Build the fieldtable for this typedef
    build_fieldtable(TypeDef, FieldTable),
    ( map.insert(Types0, TypeDef^typedef_name, TypeDef-FieldTable, Types_) ->
        Types = Types_
    ;
        context.throw_error("Duplicate type name: " ++ TypeDef ^ typedef_name,
                            TypeDef ^ typedef_context)
    ),
    TypeDef = typedef(_,_,TypedefCtors,_,_),
    list.foldl(add_ctor_to_progtable(TypeDef), TypedefCtors, Ctors0,Ctors).
add_to_progtable(pfunction(Func), progtable(Types, Ctors, Funcs0),
                                    progtable(Types, Ctors, Funcs)) :-
    ( map.insert(Funcs0, Func ^ func_name, Func, Funcs_) ->
        Funcs = Funcs_
    ;
        % FuncContextTemp does NOT have the function name in the context stack
        % (simply because the function name is already printed in the error
        % message).
        context.context_pop_block(Func ^ func_context, FuncContextTemp),
        context.throw_error("Duplicate function name: " ++ Func ^ func_name,
                            FuncContextTemp)
    ).
add_to_progtable(pimport(_), !ProgTable).       % Can't process
    % (XXX: should this be a software error?)

update_progtable_typedef(TypeDef, progtable(Types0, Ctors0, Funcs),
                                    progtable(Types, Ctors, Funcs)) :-
    % Build the fieldtable for this typedef
    build_fieldtable(TypeDef, FieldTable),
    map.set(Types0, TypeDef^typedef_name, TypeDef-FieldTable, Types),
    TypeDef = typedef(_,_,TypedefCtors,_,_),
    list.foldl(add_ctor_to_progtable(TypeDef), TypedefCtors, Ctors0,Ctors).
update_progtable_function(Func, progtable(Types, Ctors, Funcs0),
                                    progtable(Types, Ctors, Funcs)) :-
    map.set(Funcs0, Func ^ func_name, Func, Funcs).

:- pred add_ctor_to_progtable(typedef::in, ctor::in, map(string, typedef)::in,
                                        map(string, typedef)::out) is det.
add_ctor_to_progtable(TypeDef, Ctor, Ctors0, Ctors) :-
    ( map.insert(Ctors0, Ctor ^ ctor_name, TypeDef, Ctors_) ->
        Ctors = Ctors_
    ;
        context.throw_error("Duplicate constructor name: " ++ Ctor ^ ctor_name,
                            Ctor ^ ctor_context)
    ).

build_localtable(Func, !Table) :-
    (
        Func^func_params = no
    ;
        Func^func_params = yes(Params),
        list.foldl(add_to_localtable_string(Func, "argument"), Params,
            !Table)
    ),
    ( Func^func_body = func_body_ast(Decls,_) ->
        list.foldl(add_to_localtable(Func, "variable"), Decls, !Table)
    ; Func^func_body = func_body_cfg(Decls,_) ->
        list.foldl(add_to_localtable(Func, "variable"), Decls, !Table)
    ;
        true
    ).
build_localtable(Func, Table) :-
    build_localtable(Func, empty_localtable, Table).

% build_fieldtable(Typedef, FieldTable) builds a field table from a typedef.
% Throws a context.error if fields of the same name in different ctors do not
% have matching types. (That is: exactly matching, not just unifiable).
:- pred build_fieldtable(typedef::in, fieldtable::out) is det.
build_fieldtable(TypeDef, FieldTable) :-
    list.foldl(add_ctor_to_fieldtable, TypeDef ^ typedef_ctors,
               map.init, FieldTable).

:- pred add_ctor_to_fieldtable(ctor::in, fieldtable::in,
                                         fieldtable::out) is det.
add_ctor_to_fieldtable(Ctor, Table0, Table) :-
    Args = Ctor ^ ctor_args,
    (
        Args = no,
        Table = Table0
    ;
        Args = yes(YesArgs),
        list.foldl(add_ctor_field_to_fieldtable(Ctor ^ ctor_context),
                   YesArgs, Table0, Table)
    ).

:- pred add_ctor_field_to_fieldtable(context::in,
                pair(maybe(string), typeval)::in,
                fieldtable::in, fieldtable::out) is det.
add_ctor_field_to_fieldtable(Context, yes(FieldName) - Type, Table0, Table) :-
    ( map.insert(Table0, FieldName, Type, Table_) ->
        Table = Table_
    ; map.lookup(Table0, FieldName) = Type ->
        % Duplicate name - OK as long as the types match exactly
        Table = Table0
    ;
        context.throw_error("Duplicate field name with differing types: "
                            ++ FieldName, Context)
    ).
% No field name, no field.
add_ctor_field_to_fieldtable(_Context, no - _, !FieldTable).

% Note: First argument is the "what" we are inserting,
% eg. "argument" or "variable".
% In this version, the 3rd argument is a string - type pair, as opposed to a
% varname - type pair.
:- pred add_to_localtable_string(function::in, string::in,
                        {string, typeval, argmode}::in,
                        localtable::in, localtable::out) is det.
add_to_localtable_string(Func, What, {Name, Type, _Argmode}, !LT) :-
    add_to_localtable(Func, What, svname(Name) - Type, !LT).

% Note: First argument is the "what" we are inserting,
% eg. "argument" or "variable".
% In this version, the 3rd argument is a varname - type pair.
:- pred add_to_localtable(function::in, string::in,
                        pair(varname, typeval)::in,
                        localtable::in, localtable::out) is det.
add_to_localtable(Func, What, Name - Type, LT0, LT) :-
    ( map.insert(LT0, Name, Type, LT_) ->
        LT = LT_
    ;
        context.throw_error("Duplicate " ++ What ++ " name: " ++
                            ir.varname_to_string(Name), Func ^ func_context)
    ).

update_localtable_string(Name, Type, !LT) :-
    update_localtable(svname(Name), Type, !LT).

update_localtable(Name, Type, LT0, LT) :-
    map.set(LT0, Name, Type, LT).

lookup_type(PT, Name, Typedef, FieldTable) :-
    map.search(PT ^ prog_types, Name, Typedef - FieldTable).

lookup_type(PT, Name, Typedef) :-
    lookup_type(PT, Name, Typedef, _).

lookup_ctor(PT, Name, Typedef, Ctor) :-
    map.search(PT ^ prog_ctors, Name, Typedef),
    search_ctor(Typedef ^ typedef_ctors, Name, Ctor).

:- pred search_ctor(list(ctor)::in, string::in, ctor::out) is det.
search_ctor([Ctor|Ctors], Name, C) :-
    ( Ctor ^ ctor_name = Name ->
        C = Ctor
    ;
        search_ctor(Ctors, Name, C)
    ).
search_ctor([], _, _) :-
    throw("Error: Constructor not found in list").

lookup_function(PT, Name, Func) :-
    map.search(PT ^ prog_funcs, Name, Func).

lookup_field(FT, Name, Type) :-
    map.search(FT, Name, Type).

lookup_field(PT, TypeName, FieldName, Type) :-
    map.search(PT ^ prog_types, TypeName, _Typedef - FT),
    map.search(FT, FieldName, Type).

lookup_local_string(LT, Name, Type) :-
    lookup_local(LT, svname(Name), Type).

lookup_local(LT, Name, Type) :-
    map.search(LT, Name, Type).
