% -------------------------------------------------------------------------- %
% ArrayRef Library - Test Suite
% Copyright (C) 2008 Matt Giuca
% 11/11/2008
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %
% This test suite forms a standalone executable program.
% It tests both the arrayref and vectorref modules.
% -------------------------------------------------------------------------- %

:- module test_arrayref.

:- interface.

:- import_module io.

:- pred main(io::di, io::uo) is det.

:- implementation.

:- import_module arrayref.
:- import_module vectorref.

:- import_module array.
:- import_module list.
:- import_module int.
:- import_module string.
:- import_module exception.
:- import_module store.

% -------------------------------------------------------------------------- %
% ARRAYREF TESTS
% -------------------------------------------------------------------------- %

% assert_eq_list(ArrayRef, List, !Store)
% If the elements of ArrayRef are equal to those of List, does nothing.
% Otherwise, raises an exception.
:- pred assert_eq_list(arrayref(T, S)::in, list(T)::in, S::di, S::uo)
    is det <= store(S).
assert_eq_list(ArrayRef, List, !S) :-
    arrayref.extract(ArrayRef, Array, !S),
    array.to_list(Array, List0),
    ( List0 = List ->
        true
    ;
        throw("assert_eq_list: arrayref(" ++ string(List0) ++ ") != "
            ++ string(List))
    ).

:- pred test_init(S::di, S::uo) is det <= store(S).
test_init(!S) :-
    arrayref.init(3, 49, A, !S),
    L = [49, 49, 49],
    assert_eq_list(A, L, !S).

:- pred test_bounds(S::di, S::uo) is det <= store(S).
test_bounds(!S) :-
    arrayref.init(3, 49, A, !S),
    arrayref.min(A, Min, !S),
    ( Min = 0 ->
        true
    ;
        throw("test_bounds: Min != 0")
    ),
    arrayref.max(A, Max, !S),
    ( Max = 2 ->
        true
    ;
        throw("test_bounds: Max != 2")
    ),
    arrayref.size(A, Size, !S),
    ( Size = 3 ->
        true
    ;
        throw("test_bounds: Size != 3")
    ),
    arrayref.bounds(A, BMin, BMax, !S),
    ( BMin = 0 ->
        true
    ;
        throw("test_bounds: BMin != 0")
    ),
    ( BMax = 2 ->
        true
    ;
        throw("test_bounds: BMax != 2")
    ).

:- pred test_lookup(S::di, S::uo) is det <= store(S).
test_lookup(!S) :-
    arrayref.init(3, 49, A, !S),
    arrayref.lookup(A, 2, V, !S),
    ( V = 49 ->
        true
    ;
        throw("test_lookup: V != 49")
    ).

:- pred test_set(S::di, S::uo) is det <= store(S).
test_set(!S) :-
    arrayref.init(3, 49, A, !S),
    % Test set on its own
    arrayref.set(A, 1, 42, !S),
    L = [49, 42, 49],
    assert_eq_list(A, L, !S),
    % Test aliased set. Set Alias, should affect A.
    Alias = A,
    arrayref.set(Alias, 0, 57, !S),
    L1 = [57, 42, 49],
    assert_eq_list(A, L1, !S).

:- pred test_copy(S::di, S::uo) is det <= store(S).
test_copy(!S) :-
    arrayref.init(3, 49, A, !S),
    arrayref.copy(A, Copy, !S),
    L = [49, 49, 49],
    assert_eq_list(Copy, L, !S),
    % Test set of a copy. Set Copy, should not affect A.
    arrayref.set(Copy, 0, 57, !S),
    L1 = [57, 49, 49],
    assert_eq_list(Copy, L1, !S),
    assert_eq_list(A, L, !S).

:- pred test_resize(S::di, S::uo) is det <= store(S).
test_resize(!S) :-
    arrayref.init(3, 49, A, !S),
    L = [49, 49, 49],
    assert_eq_list(A, L, !S),   % Sanity check

    % Resize bigger, then test setting the copied array; should not affect A.
    arrayref.resize(A, 5, 201, A1, !S),
    L1 = [49, 49, 49, 201, 201],
    assert_eq_list(A1, L1, !S),
    arrayref.set(A1, 3, 57, !S),
    L1_1 = [49, 49, 49, 57, 201],
    assert_eq_list(A1, L1_1, !S),
    assert_eq_list(A, L, !S),   % Original unmodified

    % Resize smaller, then test setting the copied array; should not affect A.
    arrayref.resize(A, 1, 201, A2, !S),
    L2 = [49],
    assert_eq_list(A2, L2, !S),
    arrayref.set(A2, 0, 57, !S),
    L2_1 = [57],
    assert_eq_list(A2, L2_1, !S),
    assert_eq_list(A, L, !S),   % Original unmodified

    % Resize same, then test setting the copied array; should not affect A.
    arrayref.resize(A, 3, 201, A3, !S),
    L3 = [49, 49, 49],
    assert_eq_list(A3, L3, !S),
    arrayref.set(A3, 1, 57, !S),
    L3_1 = [49, 57, 49],
    assert_eq_list(A3, L3_1, !S),
    assert_eq_list(A, L, !S).   % Original unmodified

:- pred test_shrink(S::di, S::uo) is det <= store(S).
test_shrink(!S) :-
    arrayref.init(3, 49, A, !S),
    L = [49, 49, 49],
    assert_eq_list(A, L, !S),   % Sanity check

    % Resize smaller, then test setting the copied array; should not affect A.
    arrayref.shrink(A, 1, A2, !S),
    L2 = [49],
    assert_eq_list(A2, L2, !S),
    arrayref.set(A2, 0, 57, !S),
    L2_1 = [57],
    assert_eq_list(A2, L2_1, !S),
    assert_eq_list(A, L, !S),   % Original unmodified

    % Resize same, then test setting the copied array; should not affect A.
    arrayref.shrink(A, 3, A3, !S),
    L3 = [49, 49, 49],
    assert_eq_list(A3, L3, !S),
    arrayref.set(A3, 1, 57, !S),
    L3_1 = [49, 57, 49],
    assert_eq_list(A3, L3_1, !S),
    assert_eq_list(A, L, !S).   % Original unmodified

:- pred test_unsafe_init(S::di, S::uo) is det <= store(S).
test_unsafe_init(!S) :-
    arrayref.unsafe_init(3, A, !S),
    % Test the size
    arrayref.size(A, Size, !S),
    ( Size = 3 ->
        true
    ;
        throw("test_unsafe_init: Size != 3")
    ),
    % Set one element, then look it up
    arrayref.set(A, 1, 18, !S),
    arrayref.lookup(A, 1, V, !S),
    ( V = 18 ->
        true
    ;
        throw("test_unsafe_init: V != 18")
    ),
    % Set all elements of the array, then test
    arrayref.set(A, 0, 17, !S),
    arrayref.set(A, 2, 19, !S),
    L = [17, 18, 19],
    assert_eq_list(A, L, !S).

:- pred test_unsafe_resize(S::di, S::uo) is det <= store(S).
test_unsafe_resize(!S) :-
    arrayref.init(3, 49, OldA, !S),
    % Expand, unsafely
    arrayref.unsafe_resize(OldA, 6, A, !S),
    % Test the size
    arrayref.size(A, Size, !S),
    ( Size = 6 ->
        true
    ;
        throw("test_unsafe_resize: Size != 6")
    ),
    % Set one element, then look it up
    arrayref.set(A, 4, 18, !S),
    arrayref.lookup(A, 4, V, !S),
    ( V = 18 ->
        true
    ;
        throw("test_unsafe_resize: V != 18")
    ),
    % Set all elements of the array, then test
    arrayref.set(A, 3, 17, !S),
    arrayref.set(A, 5, 19, !S),
    L = [49, 49, 49, 17, 18, 19],
    assert_eq_list(A, L, !S).

:- pred test_arrayref(io::di, io::uo) is det.
test_arrayref(!IO) :-
    test_init(!IO),
    test_bounds(!IO),
    test_lookup(!IO),
    test_set(!IO),
    test_copy(!IO),
    test_resize(!IO),
    test_shrink(!IO),
    test_unsafe_init(!IO),
    test_unsafe_resize(!IO),
    io.write_string("arrayref: All tests successful.\n", !IO).

% -------------------------------------------------------------------------- %
% VECTORREF TESTS
% -------------------------------------------------------------------------- %

% assert_v_eq_list(VectorRef, List, !Store)
% If the elements of VectorRef are equal to those of List, does nothing.
% Otherwise, raises an exception.
:- pred assert_v_eq_list(vectorref(T, S)::in, list(T)::in, S::di, S::uo)
    is det <= store(S).
assert_v_eq_list(VectorRef, List, !S) :-
    vectorref.extract(VectorRef, Array, !S),
    array.to_list(Array, List0),
    ( List0 = List ->
        true
    ;
        throw("assert_v_eq_list: vectorref(" ++ string(List0) ++ ") != "
            ++ string(List))
    ).

:- pred test_v_init(S::di, S::uo) is det <= store(S).
test_v_init(!S) :-
    vectorref.init(A, !S),
    L = [] : list(int),
    assert_v_eq_list(A, L, !S).

:- pred test_v_lots(S::di, S::uo) is det <= store(S).
% Tests add, remove, size, capacity, bounds
test_v_lots(!S) :-
    vectorref.init(3, A, !S), % Will need more capacity
    vectorref.add(A, 23, !S),
    vectorref.add(A, 14, !S),
    L = [23, 14],
    assert_v_eq_list(A, L, !S),
    % Test size, capacity, bounds
    vectorref.size(A, Size, !S),
    ( Size = 2 ->
        true
    ;
        throw("test_bounds: Size != 2")
    ),
    vectorref.capacity(A, Capacity, !S),
    ( Capacity = 3 ->
        true
    ;
        throw("test_bounds: Capacity != 3")
    ),
    vectorref.bounds(A, BMin, BMax, !S),
    ( BMin = 0 ->
        true
    ;
        throw("test_bounds: BMin != 0")
    ),
    ( BMax = 1 ->
        true
    ;
        throw("test_bounds: BMax != 1")
    ),

    % Test effects of aliasing. Add to Alias, check A has been modified.
    Alias = A,
    vectorref.add(Alias, 37, !S),
    vectorref.add(Alias, 42, !S),
    vectorref.add(Alias, 16, !S),
    L1 = [23, 14, 37, 42, 16],
    assert_v_eq_list(A, L1, !S),
    % Test that capacity has doubled for both A and Alias.
    vectorref.capacity(A, Capacity1, !S),
    ( Capacity1 = 6 ->
        true
    ;
        throw("test_bounds: Capacity1 != 6")
    ),
    vectorref.capacity(Alias, CapacityAlias, !S),
    ( CapacityAlias = 6 ->
        true
    ;
        throw("test_bounds: CapacityAlias != 6")
    ),
    vectorref.remove(A, 1, !S),
    vectorref.remove(A, 2, !S),
    L2 = [23, 37, 16],
    assert_v_eq_list(A, L2, !S).

:- pred test_v_lookup(S::di, S::uo) is det <= store(S).
test_v_lookup(!S) :-
    vectorref.init(3, A, !S),
    vectorref.add(A, 49, !S),
    vectorref.add(A, 49, !S),
    vectorref.add(A, 49, !S),
    vectorref.lookup(A, 2, V, !S),
    ( V = 49 ->
        true
    ;
        throw("test_v_lookup: V != 49")
    ).

:- pred test_v_set(S::di, S::uo) is det <= store(S).
test_v_set(!S) :-
    vectorref.init(3, A, !S),
    vectorref.add(A, 49, !S),
    vectorref.add(A, 49, !S),
    vectorref.add(A, 49, !S),
    % Test set on its own
    vectorref.set(A, 1, 42, !S),
    L = [49, 42, 49],
    assert_v_eq_list(A, L, !S),
    % Test aliased set. Set Alias, should affect A.
    Alias = A,
    vectorref.set(Alias, 0, 57, !S),
    L1 = [57, 42, 49],
    assert_v_eq_list(A, L1, !S).

:- pred test_v_copy(S::di, S::uo) is det <= store(S).
test_v_copy(!S) :-
    vectorref.init(5, A, !S),
    vectorref.add(A, 23, !S),
    vectorref.add(A, 14, !S),
    vectorref.add(A, 37, !S),
    vectorref.copy(A, Copy, !S),
    L = [23, 14, 37],
    assert_v_eq_list(A, L, !S),
    assert_v_eq_list(Copy, L, !S),
    % Test add to a copy. Update Copy, should not affect A.
    vectorref.add(Copy, 57, !S),
    L1 = [23, 14, 37, 57],
    assert_v_eq_list(Copy, L1, !S),
    assert_v_eq_list(A, L, !S).

:- pred test_v_ensure_capacity(S::di, S::uo) is det <= store(S).
test_v_ensure_capacity(!S) :-
    vectorref.init(3, A, !S), % Will need more capacity
    vectorref.add(A, 23, !S),
    vectorref.add(A, 14, !S),
    vectorref.capacity(A, Cap0, !S),
    ( Cap0 = 3 ->
        true
    ;
        throw("test_bounds: Cap0 != 3")
    ),
    % Make the capacity larger; should extend the array
    vectorref.ensure_capacity(A, 7, !S),
    vectorref.capacity(A, Cap1, !S),
    ( Cap1 = 7 ->
        true
    ;
        throw("test_bounds: Cap1 != 7")
    ),
    % Make the capacity smaller; should ignore and stay at 7
    vectorref.ensure_capacity(A, 2, !S),
    vectorref.capacity(A, Cap2, !S),
    ( Cap2 = 7 ->
        true
    ;
        throw("test_bounds: Cap2 != 7")
    ).

:- pred test_vectorref(io::di, io::uo) is det.
test_vectorref(!IO) :-
    test_v_init(!IO),
    test_v_lots(!IO),
    test_v_lookup(!IO),
    test_v_set(!IO),
    test_v_copy(!IO),
    test_v_ensure_capacity(!IO),
    io.write_string("vectorref: All tests successful.\n", !IO).

main(!IO) :-
    test_arrayref(!IO),
    test_vectorref(!IO).
