% -------------------------------------------------------------------------- %
% Mars/Venus tokenizer
% Copyright (C) 2008 Matt Giuca
% 27/5/2008
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- module token.

:- interface.

:- import_module integer.
:- import_module char.
:- import_module string.
:- import_module list.

:- import_module parse.

% -------------------------------------------------------------------------- %

:- type token
    --->    newline
    ;       indent
    ;       dedent
    ;       uident(string)  % Starts with uppercase
    ;       lident(string)  % Starts with lowercase
    ;       intlit(integer)
    ;       charlit(char)
    ;       stringlit(string)
    ;       pass
    ;       return
    ;       if_t
    ;       else_t
    ;       switch
    ;       case
    ;       while
    ;       import
    ;       def
    ;       type_t
    ;       var
    ;       builtin
    ;       openparen
    ;       closeparen
    ;       opensquare
    ;       closesquare
    ;       dot
    ;       dotdotdot
    ;       comma
    ;       backtick
    ;       underscore
    ;       arrow
    ;       colon
    ;       coloncolon
    ;       eq
    ;       eqeq
    ;       coloneq.

% "nice" show representation of tokens.
:- func showToken(token) = string.

% Converts a character stream into a token stream.
% Throws an exception on error (unexpected character).
% First arg is Filename.
:- pred tokenize(string::in, list(char)::in, list(token)::out) is det.
:- pred tokenize(list(char)::in, list(token)::out) is det.

% A token parser. (That is a parser which accepts tokens as input).
:- type tokenparser(S) == parser(token, S).
:- type tokenparser == parser(token).

% Consume a token and return it.
:- pred token(token::out, tokenparser(S)::in, tokenparser(S)::out) is semidet.
:- pred uident(string::out, tokenparser(S)::in, tokenparser(S)::out)
            is semidet.
:- pred lident(string::out, tokenparser(S)::in, tokenparser(S)::out)
            is semidet.
:- pred intlit(integer::out, tokenparser(S)::in, tokenparser(S)::out)
            is semidet.

% Predicates by special request
% is_lident(S) succeeds if S is a lident.
:- pred is_lident(string::in) is semidet.

% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- implementation.

:- import_module int.
:- import_module maybe.
:- import_module stack.
:- import_module require.
:- import_module exception.

:- import_module context.

% IParser is a character parser which stores an internal state which
% is a stack of Ints, representing all of the indentation levels up to the
% current element.
% Invariant: Stack is non-empty. (Errors are thrown if it is).
% Invariant: Ints are pushed in strictly-ascending order on the stack.
% (As described here: http://docs.python.org/ref/indentation.html)
% This is used by scanners to transform strings into lists of tokens.
:- type iparser == cparser(stack(int)).

showToken(newline) = "newline".
showToken(indent) = "indent".
showToken(dedent) = "dedent".
showToken(uident(S)) = "identifier: " ++ S.
showToken(lident(S)) = "identifier: " ++ S.
showToken(intlit(I)) = "integer literal: " ++ integer.to_string(I).
showToken(charlit(C)) = "character literal: " ++ string(C).
showToken(stringlit(_)) = "string literal".
showToken(pass) = "\"pass\"".
showToken(return) = "\"return\"".
showToken(if_t) = "\"if\"".
showToken(else_t) = "\"else\"".
showToken(switch) = "\"switch\"".
showToken(case) = "\"case\"".
showToken(while) = "\"while\"".
showToken(import) = "\"import\"".
showToken(def) = "\"def\"".
showToken(type_t) = "\"type\"".
showToken(var) = "\"var\"".
showToken(builtin) = "\"builtin\"".
showToken(openparen) = "\"(\"".
showToken(closeparen) = "\")\"".
showToken(opensquare) = "\"[\"".
showToken(closesquare) = "\"]\"".
showToken(dot) = "\".\"".
showToken(dotdotdot) = "\"...\"".
showToken(comma) = "\",\"".
showToken(backtick) = "\"`\"".
showToken(underscore) = "\"_\"".
showToken(arrow) = "\"->\"".
showToken(colon) = "\":\"".
showToken(coloncolon) = "\"::\"".
showToken(eq) = "\"=\"".
showToken(eqeq) = "\"==\"".
showToken(coloneq) = "\":=\"".

% -------------------------------------------------------------------------- %

%% State handlers %%

% getcolumn(Column, !Parser).
% Gets the column of the parser on the current line.
:- pred get_column(int::out, iparser::in, iparser::out) is det.
get_column(Ctx ^ ctx_column) -->
    parse.get_context(Ctx).

% Gets the indent level on top of the current stack.
:- pred get_current_indent_level(int::out, iparser::in, iparser::out) is det.
get_current_indent_level(stack.top_det(State)) -->
    parse.get_state(State).

% Push a new indent level on top of the current stack.
% Must be greater than the top indent level.
:- pred push_indent_level(int::in, iparser::in, iparser::out) is det.
push_indent_level(Indent) -->
    parse.get_state(State),
    parse.set_state(stack.push(State, Indent)).

% pop_indents_until(Indent, NumDedents, !Parser).
% Pops the indent stack of all elements greater than the int specified,
% leaving the specified int on top of the stack.
% Fails if the int specified was not found (expects the stack to be
% descending from the top).
% Returns the number of elements popped in NumDedents.
:- pred pop_indents_until(int::in, int::out, iparser::in, iparser::out)
        is semidet.
pop_indents_until(Indent, NumDedents, !P) :-
    parse.get_state(State, !P),
    stack.pop_det(State, Top, Rest),
    ( Top = Indent ->
        % Match - finished popping (keep old state)
        NumDedents = 0
    ; Top < Indent ->
        % Failed to match exactly - fail
        fail
    ;
        % Not there yet. Pop and continue.
        parse.set_state(Rest, !P),
        pop_indents_until(Indent, NumDedents0, !P),
        NumDedents = NumDedents0 + 1
    ).

% -------------------------------------------------------------------------- %

%% Actual token parsers %%

% Consumes 1 whitespace token.
:- pred whitespace(iparser::in, iparser::out) is semidet.
whitespace -->
    ( parse.char(' ') -> {true}
    ; parse.char('\t') -> {true}
    ; {fail}).

% Consumes a '#' character then all characters up until a newline or EOF.
% (Consumes the newline character).
:- pred linecomment(iparser::in, iparser::out) is semidet.
linecomment -->
    parse.char('#'),
    linecomment_.       % Always succeeds
:- pred linecomment_(iparser::in, iparser::out) is det.
linecomment_ -->
    ( parse.char(C) ->
        ( {C = '\n'} ->
            {true}          % Finish, having consumed the newline
        ;
            linecomment_    % Recurse; not done yet
        )
    ;
        {true}              % Finish, having consumed all input
    ).

% Consumes an identifier and returns it as a token.
% May also return other tokens - checks for alphanumeric tokens
% (such as underscore and keywords), and returns those instead.
:- pred ident(token::out, iparser::in, iparser::out) is semidet.
ident(Token, !P) :-
    parse.alpha_or_underscore(C, !P),
    parse.many(parse.alnum_or_underscore, Cs, !P),
    string.from_char_list([C|Cs], Ident),
    ( Ident = "pass" -> Token = pass
    ; Ident = "return" -> Token = return
    ; Ident = "if" -> Token = if_t
    ; Ident = "else" -> Token = else_t
    ; Ident = "switch" -> Token = switch
    ; Ident = "case" -> Token = case
    ; Ident = "while" -> Token = while
    ; Ident = "import" -> Token = import
    ; Ident = "def" -> Token = def
    ; Ident = "type" -> Token = type_t
    ; Ident = "var" -> Token = var
    ; Ident = "builtin" -> Token = builtin
    ; Ident = "_" -> Token = underscore
    % Not special - either uppercase or lowercase identifier
    ; char.is_upper(C) ->
        Token = uident(Ident)
    ;
        Token = lident(Ident)
    ).

% Consumes an integer literal and returns it as a token.
:- pred tok_intlit(token::out, iparser::in, iparser::out) is semidet.
tok_intlit(intlit(Val)) -->
    ( parse.char('-') ->
        {Neg = ['-']}
    ;
        {Neg = []}
    ),
    parse.many1(parse.digit, Chrs),     % Can fail
    {Val = integer.det_from_string(string.from_char_list(Neg ++ Chrs))}.

% stringlit_char(What, Char, !Parser).
% Consumes a single "character" in a string or character literal.
% Note that a "character" in this context is any number of source characters
% which make up a single character in the literal.
% eg. "c" is a character, but so is "\\n" and so is "\\x48".
% Escapes are C specific control character escapes, and \x__.
% Only characters in the range [0x20, 0x7f] may be used unescaped.
% What is either "string" or "character", for error messages.
:- pred stringlit_char(string::in, char::out, iparser::in, iparser::out)
    is det.
stringlit_char(What, C) -->
    ( stringlit_char_(What, C_) ->
        {C = C_}
    ;
        parse.throw_error(
                "Unexpected end-of-file in " ++ What ++ " literal")
    ).

:- pred stringlit_escape_special(char::in, char::out) is semidet.
stringlit_escape_special('0', char.det_from_int(0)).
stringlit_escape_special('a', '\a').
stringlit_escape_special('b', '\b').
stringlit_escape_special('t', '\t').
stringlit_escape_special('n', '\n').
stringlit_escape_special('v', '\v').
stringlit_escape_special('f', '\f').
stringlit_escape_special('r', '\r').
stringlit_escape_special('e', char.det_from_int(27)).
stringlit_escape_special('\\', '\\').
stringlit_escape_special('\'', '\'').
stringlit_escape_special('\"', '\"').

% Semidet - fails if EOF encountered.
:- pred stringlit_char_(string::in, char::out, iparser::in, iparser::out)
    is semidet.
stringlit_char_(What, C) -->
    parse.char(C1),         % Can fail
    {Code = char.to_int(C1)},
    ( {C1 = char.det_from_int(92)} ->   % '\\' (bug in Mercury)
        % Escape character - have to read more
        parse.char(C2),     % Can fail
        ( {stringlit_escape_special(C2, C_)} ->
            {C = C_}
        ; {C2 = 'x'} ->
            % Hex escape
            ( parse.hex_digit(C3), parse.hex_digit(C4) ->
                {char.digit_to_int(C3, I3)},    % Shouldn't fail
                {char.digit_to_int(C4, I4)},    % Shouldn't fail
                {C = char.det_from_int(I3 * 16 + I4)}
            ;
                parse.throw_error(
                        "Expected 2 hex digits following \\x in "
                        ++ What ++ " literal")
            )
        ;
            parse.throw_error(
                    "Invalid escape character \\" ++ string.from_char(C2)
                    ++ " in " ++ What ++ " literal")
        )
    ; {Code >= 32}, {Code < 128} ->
        % Valid regular character
        {C = C1}
    ;
        % Invalid string character (including tab and newline)
        ( {C1 = '\n'} ->
            parse.throw_error(
                    "Newline in " ++ What ++ " literal (illegal)")
        ; {C1 = '\t'} ->
            parse.throw_error(
                    "Tab in " ++ What ++ " literal (illegal)")
        ;
            parse.throw_error(
                    "Invalid character in " ++ What ++ " literal")
        )
    ).

% Consumes a character literal and returns it as a token.
:- pred tok_charlit(token::out, iparser::in, iparser::out) is semidet.
tok_charlit(charlit(Chr)) -->
    parse.char('\''),       % Can fail
    stringlit_char("character", Chr),
    ( parse.char('\'') ->
        {true}
    ;
        parse.throw_error(
            "Expected end of character literal")
    ).

% Consumes a string literal and returns it as a token.
:- pred tok_stringlit(token::out, iparser::in, iparser::out) is semidet.
tok_stringlit(stringlit(Str)) -->
    parse.char('\"'),       % Can fail
    tok_stringlit_(Chrs),
    {string.from_char_list(Chrs, Str)}.

:- pred tok_stringlit_(list(char)::out, iparser::in, iparser::out) is det.
tok_stringlit_(Chrs) -->
    ( parse.char('\"') ->
        % Found unescaped '"' - end the string
        {Chrs = []}
    ;
        % Read the next character
        stringlit_char("string", C),
        tok_stringlit_(Cs),
        {Chrs = [C|Cs]}
    ).

% start_of_line(IndentTokens, !Parser).
% Consume whitespace, but assume at the start of a line, and generate
% indent/dedent tokens.
% May consume 0 or more characters of input (0 if there is no whitespace).
% May produce 0 or more tokens (0 if same indent level as previous).
%
% Note: If it encounters blank lines (including those with just whitespace and
% comments), consumes the entire line and continues. Consumes as many lines as
% necessary until it finds a line with code on it, or EOF.
%
% If this line is more indented than the previous, pushes a new indent level
% on the stack and returns [indent].
% If this line is less indented than the previous, and it matches one of the
% indent levels on the stack, pops back the stack to that level and returns
% a list of dedent tokens, for as many as there were levels popped.
% If this line is less indented than the previous, but does not match one
% on the stack, throws a parse error, "Bad indentation".
% If it is the same as the previous, does nothing and returns [].
:- pred start_of_line(list(token)::out, iparser::in, iparser::out) is det.
start_of_line(Tokens) -->
    % Consume all whitespace (may be 0).
    parse.many(whitespace),
    % Check for either a comment or a newline (indicating nothing on line)
    ( linecomment ->
        start_of_line(Tokens0),      % Continue with next line
        {Tokens = [newline|Tokens0]} % Record this newline in the token stream
    ; parse.char('\n') ->
        start_of_line(Tokens0),      % Continue with next line
        {Tokens = [newline|Tokens0]} % Record this newline in the token stream
    ; parse.stream_empty ->
        % Done - empty tokens
        {Tokens = []}
    ;
        % We must have some real characters on this line
        % We'll let the caller handle those - we just want to get the
        % indents/dedents for this line first.
        get_current_indent_level(OldLevel),
        get_column(NewLevel),
        ( {NewLevel = OldLevel} ->
            % Same as previous line, so no tokens.
            {Tokens = []}
        ; {NewLevel > OldLevel} ->
            % More indented than previous line. Push and return an indent.
            push_indent_level(NewLevel),
            {Tokens = [indent]}
        ;
            % Less indented than previous line.
            ( pop_indents_until(NewLevel, NumDedents) ->
                % Successfully matched dedents. Return dedents.
                {Tokens = list.duplicate(NumDedents, dedent)}
            ;
                % Failed to match. Syntax error.
                parse.throw_error("Bad indentation")
            )
        )
    ).

% Retrieves a token from a character stream.
% Returns "no" if it consumed something meaningless (like whitespace),
% but still consumes something. Never succeeds without consuming.
% Fails on error (unexpected character).
:- pred onetoken(maybe(token)::out, iparser::in, iparser::out) is semidet.
onetoken(Tok) -->
    ( parse.many1(whitespace) -> {Tok = no}
    ; linecomment -> {Tok = yes(newline)}
    ; ident(T) -> {Tok = yes(T)}        % Note: Matches any alphanum tokens
    ; tok_intlit(T) -> {Tok = yes(T)}
    ; tok_charlit(T) -> {Tok = yes(T)}
    ; tok_stringlit(T) -> {Tok = yes(T)}
    ; parse.char('\n') -> {Tok = yes(newline)}
    ; parse.char('(') -> {Tok = yes(openparen)}
    ; parse.char(')') -> {Tok = yes(closeparen)}
    ; parse.char('[') -> {Tok = yes(opensquare)}
    ; parse.char(']') -> {Tok = yes(closesquare)}
    ; parse.string("...") -> {Tok = yes(dotdotdot)}
    ; parse.char('.') -> {Tok = yes(dot)}
    ; parse.char(',') -> {Tok = yes(comma)}
    ; parse.char('\x60\') -> {Tok = yes(backtick)}
    ; parse.string("->") -> {Tok = yes(arrow)}
    ; parse.string("::") -> {Tok = yes(coloncolon)}
    ; parse.string(":=") -> {Tok = yes(coloneq)}
    ; parse.char(':') -> {Tok = yes(colon)}
    ; parse.string("==") -> {Tok = yes(eqeq)}
    ; parse.char('=') -> {Tok = yes(eq)}
    ; {fail}).

% Retrieves all tokens from a character stream.
% Throws an exception on error (unexpected character).
% The resulting parser will be empty.
:- pred tokenize_(list(token)::out, iparser::in, iparser::out) is det.
tokenize_(Toks) -->
    % If we're at the start of line, consume all input up to the first
    % non-whitespace character, and compute indent/dedent tokens.
    get_column(C),
    ( {C = 1} ->
        % At start of line
        start_of_line(Indents)
    ;
        {Indents = []}
    ),
    ( parse.stream_empty ->
        % Base case - EOF
        % Push the last tokens - any Indents we found above, a newline, and
        % all dedent tokens required to fully pop the stack.
        ( pop_indents_until(1, NumDedents) ->
            {Dedents = list.duplicate(NumDedents, dedent)}
        ;
            {error("tokenize_: No '1' on stack, during EOF condition")}
        ),
        {Toks = Indents ++ [newline|Dedents]}
    ;
        ( onetoken(T) ->
            % Successful parse
            (
                % But it was just whitespace. Recurse.
                {T = no},
                tokenize_(Ts),
                {Toks = Indents ++ Ts}
            ;
                % Prepend, and recurse
                {T = yes(T_)},
                tokenize_(Ts),
                {Toks = Indents ++ [T_|Ts]}
            )
        ;
            % onetoken failed - error in character sequence
            ( parse.char(BadChar) ->
                parse.throw_error(
                        "Invalid character: '" ++ string.from_char(BadChar)
                        ++ "'")
            ;
                % Should not fail
                {throw("tokenize: parse.char failed (unexpected)")}
            )
        )
    ).

% High-level tokenize (exported)
tokenize(Chars, Toks) :-
    tokenize("", Chars, Toks).

% High-level tokenize (exported)
tokenize(Filename, Chars, Toks) :-
    % Create a new parser.
    % Start indent stack off with [1] - which will always stay on the bottom
    % of the stack to ensure the invariant.
    parse.new_parser(Chars, Filename, stack.push(stack.init, 1), Parser),
    tokenize_(Toks, Parser, _).
    % Note: Resulting parser guaranteed to be empty, no need to check.

:- pred token_nextpos(token::in, context::in, context::out) is det.
token_nextpos(Tok, Ctx0, Ctx) :-
    ( Tok = newline ->
        Ctx0 = context(F, L, C, B),
        Ctx = context(F, L+1, C, B)
    ;
        Ctx = Ctx0
    ).

token(Tok) -->
    token_prim(token_nextpos, Tok).
uident(Name) -->
    token(uident(Name)).
lident(Name) -->
    token(lident(Name)).
intlit(Name) -->
    token(intlit(Name)).

% Special request predicates
is_lident(S) :-
    string.to_char_list(S, Chars),
    parse.new_parser(Chars, "", stack.push(stack.init, 1), Parser0),
    ident(lident(_), Parser0, Parser1),     % Can fail (if not lident)
    parse.stream_empty(Parser1, _).         % Can fail (if not stream end)
