% -------------------------------------------------------------------------- %
% Mars/Venus type checking system
% Copyright (C) 2008 Matt Giuca
% 29/5/2008
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- module typecheck.

:- interface.

:- import_module varset.

:- import_module ir.
:- import_module tables.
:- import_module context.
:- import_module types.

% expr_type(Context, ProgTable, LocalTable, !Varset, !Expr, Type).
% Determines the type of an expression.
% If Expr already has a stored type, just returns that.
% Updates Varset with any bindings and new variables as a result of this.
% Updates Expr so it stores the computed type.
% May refer to elements of the program table or local table. These should
% reflect the context which the expression is in.
% Throws a context.error if:
%   - A reference is made to a variable which does not exist in the table.
%   - A type error occurs.
% Otherwise, succeeds. Note that this may bind variables, or create new
% variables, which will be reflected in the output varset.
:- pred expr_type(context::in, progtable::in, localtable::in,
                    varset::in, varset::out,
                    expr::in, expr::out, typeval::out)
                    is det.

% type_kind(Context, ProgTable, Varset, Type, Kind).
% Determines the kind (effectively, arity) of a type.
% May refer to elements of the program table.
% Throws a context.error if:
%   - A reference is made to a typedef which does not exist in the table.
%   - Too many arguments are supplied to a type.
% Otherwise, succeeds.
:- pred type_kind(context::in, progtable::in,
                    varset::in, typeval::in, kind::out) is det.

% check_stmt(ProgTable, LocalTable, FuncRetType, !Varset, !Stmt).
% Checks to see if a statement is type-correct, and annotates the type of
% expressions within the statement.
% Updates Varset with any bindings and new variables as a result of this.
% Updates Stmt so its expressions hold their type values.
% Throws a context.error on reference or type errors.
:- pred check_stmt(progtable::in, localtable::in, typeval::in,
                    varset::in, varset::out,
                    stmt::in, stmt::out) is det.

% check_basic_stmt(Context, ProgTable, LocalTable, !Varsete, !BasicStmt).
% Checks to see if a basic statement is type-correct.
% Does not require a function return type.
:- pred check_basic_stmt(context::in, progtable::in, localtable::in,
                    varset::in, varset::out,
                    basic_stmt::in, basic_stmt::out) is det.

% check_function(ProgTable, LocalTable, !Function).
% Checks to see if a function is type-correct, and annotates the type of
% expressions within the function.
% Updates the function's Varset with any bindings and new variables as a
% result of this.
% Throws a context.error on reference or type errors.
:- pred check_function(progtable::in, localtable::in,
                    function::in, function::out) is det.

% check_typedef(ProgTable, Typedef).
% Checks to see if a typedef is type-correct.
% Throws a context.error on reference or type errors.
:- pred check_typedef(progtable::in, typedef::in) is det.

% check_program(ProgTable, !Program).
% Checks to see if a program is type-correct, and annotates the type
% of expressions within the program.
% Updates the varsets and expressions of the whole program with annotations
% and extra variables.
:- pred check_program(progtable::in, program::in, program::out) is det.

% det_unify(Context, What, Term, Type1, Type2, !Varset).
% Unify two types with respect to a single varset.
% Throws a context.error if the types do not unify.
% Succeeds if they do, and possibly updates varset with new bindings.
% "What" : see type_error
:- pred det_unify(context::in, string::in, expr::in, typeval::in, typeval::in,
                  varset::in, varset::out) is det.

% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- implementation.

:- import_module int.
:- import_module bool.
:- import_module string.
:- import_module maybe.
:- import_module pair.
:- import_module list.
:- import_module require.

:- import_module pretty.

% list_elem_type(Context, ProgTable, LocalTable, What, Type1, Elems, !Varset).
% Unify the type of all elements of an expression list with a type.
% Throws a context.error if all the types do not unify, or if there are type
% errors in any of the Elems.
% Succeeds if they do, and possibly updates varset with new bindings.
% Also updates the list of expressions with type bindings.
% "What" : see type_error
:- pred list_elem_type(context::in, progtable::in, localtable::in, string::in,
                        typeval::in, list(expr)::in, list(expr)::out,
                        varset::in, varset::out) is det.
list_elem_type(_Ctx, _PT, _LT, _What, _T1, [], [], !Varset).
list_elem_type(Ctx, PT, LT, What, T1, [E0|Es0], [E|Es], !Varset) :-
    expr_type(Ctx, PT, LT, !Varset, E0, E, EType),
    det_unify(Ctx, What, E, EType, T1, !Varset),
    list_elem_type(Ctx, PT, LT, What, T1, Es0, Es, !Varset).

% expr_type for expressions with a computed type - just return it.
expr_type(_Ctx, _PT, _LT, !Varset, E@expr(_Expr, yes(Type)), E, Type).
% expr_type for expressions without a computed type.
% Compute the type and store it in the output expr.
expr_type(Ctx, PT, LT, !Varset, expr(Expr0,no), expr(Expr,yes(Type)), Type) :-
    expr_type_(Ctx, PT, LT, !Varset, Expr0, Expr, Type).

% expr_type_(ProgTable, LocalTable, !Varset, !Expr_, Type).
% Determines the type of an expr_ (the data structure inside expr).
:- pred expr_type_(context::in, progtable::in, localtable::in,
                    varset::in, varset::out,
                    expr_::in, expr_::out, typeval::out)
                    is det.
% Type of intlit is Int.
expr_type_(_Ctx, _PT, _LT, !Varset, E@intlit(_), E, types.const("Int")).
expr_type_(Ctx, PT, LT, !Varset, E@arraylit(Elems0), arraylit(Elems),
            types.app(types.const("Array"), [ElemType])) :-
    % Find the type of each element. (Type error unless all unify).
    types.new_var(ElemType, !Varset),
    What = "expression '" ++ pretty.string_expr(expr(E,no)) ++ "'",
    list_elem_type(Ctx, PT, LT, What, ElemType, Elems0, Elems, !Varset).
expr_type_(Ctx, PT, LT, !Varset, E@varref(VarName), E, Type) :-
    ( tables.lookup_local(LT, VarName, Type_) ->
        % Found in local table
        Type = Type_
    ;
        VarName = svname(UnqualVarName),
        tables.lookup_function(PT, UnqualVarName, Func)
    ->
        % Found in global function table - get its type and varset
        function_type(Func, FuncVarset, Type0),
        % Augment our own varset with this function's varset and
        % translate the function's type to refer to the combined varset
        types.varset_mergeone(FuncVarset, !Varset, Type0, Type)
    ;
        context.throw_error("Undefined variable: "
            ++ ir.varname_to_string_noescape(VarName), Ctx)
    ).
expr_type_(Ctx, PT, _LT, !Varset, E@ctorref(VarName), E, Type) :-
    ( tables.lookup_ctor(PT, VarName, TypeDef, _) ->
        % Found in global ctor table - get its type and varset
        ( ctor_type(TypeDef, VarName, CtorVarset, Type0) ->
            % Augment our own varset with this function's varset and
            % translate the function's type to refer to the combined varset
            types.varset_mergeone(CtorVarset, !Varset, Type0, Type)
        ;
            % Should not fail since we just looked it up
            error("expr_type_: ctor_type failed on known ctor")
        )
    ;
        context.throw_error("Undefined variable: " ++ VarName, Ctx)
    ).
expr_type_(Ctx, PT, LT, !Varset,
            fieldref(Expr0, FieldName), fieldref(Expr, FieldName), Type) :-
    expr_type(Ctx, PT, LT, !Varset, Expr0, Expr, ObjType0),
    types.deref(!.Varset, ObjType0, ObjType),
    % ObjType is the type which we are accessing
    ( ObjType = variable(_) ->
        % We can't handle a free type variable being dereferenced.
        context.throw_error("Type error: Expression '"
            ++ pretty.string_expr(Expr) ++ "' has unknown type; can't "
            ++ "access field", Ctx)
    ;
        % TODO: Handle this properly
        context.throw_error("Sorry: Type checker cannot yet handle field "
            ++ "references", Ctx)
    ),
    Type = ObjType.
expr_type_(Ctx, PT, LT, !Varset, app(Func0, Args0), Expr@app(Func, Args),
        RetType) :-
    expr_type(Ctx, PT, LT, !Varset, Func0, Func, FuncType0),
    expr_types(Ctx, PT, LT, !Varset, Args0, Args, ArgTypes),
    types.deref(!.Varset, FuncType0, FuncType),
    What = "expression '" ++ pretty.string_expr(expr(Expr,no)) ++ "'",
    % Unify: FuncType = ArgType -> Type
    % Switch here simply to customize better error messages
    ( FuncType = types.app(types.functype, FuncTypeParams@[_|_]) ->
        % Check the arity
        list.split_last_det(FuncTypeParams, InputTypes, RetType),
        Arity = list.length(InputTypes),
        NumArgs = list.length(Args),
        ( Arity = NumArgs ->
            det_unify_lists(Ctx, What, Args, ArgTypes, InputTypes, !Varset)
        ;
            type_error_arity(Ctx, !.Varset, What, Func, FuncType, Arity,
                NumArgs)
        )
    ; FuncType = types.variable(_) ->
        % Function is a variable type - simply bind it
        types.new_var(RetType, !Varset),
        det_unify(Ctx, What, Func, FuncType,
                  types.app(types.functype, ArgTypes ++ [RetType]),
                  !Varset)
    ;
        % All other alternatives - not going to unify with ArgType -> Type.
        % Hence, "this is not a function" error.
        error_not_a_function(Ctx, !.Varset, What, Func, FuncType)
    ).
expr_type_(Ctx, PT, LT, !Varset, parapp(Func0, Args0),
        Expr@parapp(Func, Args), ResultType) :-
    expr_type(Ctx, PT, LT, !Varset, Func0, Func, FuncType0),
    expr_types(Ctx, PT, LT, !Varset, Args0, Args, ArgTypes),
    types.deref(!.Varset, FuncType0, FuncType),
    What = "expression '" ++ pretty.string_expr(expr(Expr,no)) ++ "'",
    % Unify: FuncType = ArgType -> Type
    % Switch here simply to customize better error messages
    ( FuncType = types.app(types.functype, FuncTypeParams@[_|_]) ->
        % Check the arity
        list.split_last_det(FuncTypeParams, InputTypes, FuncRetType),
        Arity = list.length(InputTypes),
        NumArgs = list.length(Args),
        ( Arity >= NumArgs ->
            list.det_split_list(NumArgs, InputTypes, GotInputTypes,
                RemInputTypes),
            % Unify all supplied args (remaining input types are not checked)
            det_unify_lists(Ctx, What, Args, ArgTypes, GotInputTypes, !Varset)
        ;
            type_error_arity(Ctx, !.Varset, What, Func, FuncType, Arity,
                NumArgs)
        ),
        ResultType = types.app(types.functype, RemInputTypes ++ [FuncRetType])
    ; FuncType = types.variable(_) ->
        % XXX Function is a variable type - this is quite hard to deal with
        % since we don't know what type skeleton to bind it to.
        % (It could be any arity >= the number of supplied arguments).
        error("Not Implemented: Partial application on an unknown type")
    ;
        % All other alternatives - not going to unify with ArgType -> Type.
        % Hence, "this is not a function" error.
        error_not_a_function(Ctx, !.Varset, What, Func, FuncType)
    ).

% expr_types(Context, ProgTable, LocalTable, !Varset, !Exprs, Types).
% Determines the type of each of a list of expressions.
% Applies expr_types to each element of the list, updating the list and
% returning a list of types.
:- pred expr_types(context::in, progtable::in, localtable::in,
                   varset::in, varset::out,
                   list(expr)::in, list(expr)::out, list(typeval)::out)
                   is det.
expr_types(_Ctx, _PT, _LT, !Varset, [], [], []).
expr_types(Ctx, PT, LT, !Varset, [E0|Es0], [E|Es], [T|Ts]) :-
    expr_type(Ctx, PT, LT, !Varset, E0, E, T),
    expr_types(Ctx, PT, LT, !Varset, Es0, Es, Ts).

type_kind(Ctx, PT, Varset, Type0, Kind) :-
    types.deref(Varset, Type0, Type),
    ( Type = types.functype ->
        % Special case for function type. Has kind ((*, **) -> *).
        Kind = kind_arrow(1, yes)
    ; Type = types.const(Name) ->
        ( tables.lookup_type(PT, Name, Typedef) ->
            Kind = typedef_kind(Typedef)
        ;
            context.throw_error("Undefined type: " ++ Name, Ctx)
        )
    ; Type = types.variable(_Var) ->
        % In this simplified kind system, assume the kind of any variable
        % is *.
        % (In a real system you would need to infer the kind of a variable).
        Kind = kind_star
    ; Type = types.app(Func, Args) ->
        type_kind(Ctx, PT, Varset, Func, FuncKind),
        % Unify FuncKind = ArgKind -> Kind
        (
            FuncKind = kind_arrow(_,_),
            kind_arity_check(Ctx, Varset, Type, Func, FuncKind, length(Args)),
            % In our simplified kind system, all argument types must be of
            % kind *, as will be the result kind
            % (this is implicit in kind_arrow).
            check_types_star(Ctx, PT, Varset, Args),
            Kind = kind_star
        ;
            FuncKind = kind_star,
            % Cannot pass an argument to this type
            error_not_a_type_function(Ctx, Varset, Type, Func, kind_star)
        )
    ;
        error("type_kind: Invalid type value")
    ).

% check_types_star(Ctx, ProgTable, Varset, Types)
:- pred check_types_star(context::in, progtable::in, varset::in,
            list(typeval)::in) is det.
check_types_star(_Ctx, _PT, _Varset, []).
check_types_star(Ctx, PT, Varset, [T|Ts]) :-
    check_type_star(Ctx, PT, Varset, T),
    check_types_star(Ctx, PT, Varset, Ts).

% check_type_star(Ctx, ProgTable, Varset, Type)
% Like type_kind, but doesn't return the kind - simply forces the kind to
% be kind_star, and throws an error if it isn't.
:- pred check_type_star(context::in, progtable::in, varset::in, typeval::in)
            is det.
check_type_star(Ctx, PT, Varset, Type) :-
    type_kind(Ctx, PT, Varset, Type, Kind),
    ( Kind = kind_star ->
        true
    ;
        kind_error(Ctx, Varset, Type, Type, Kind, kind_star)
    ).

% check_pattern(Context, ProgTable, !LocalTable, !Varset,
%               What, SwitchType, Pattern).
% Checks to see if a pattern is type-correct, with respect to its parent
% switch type.
% Updates the varset and localtable with the extra type and term variables
% created by the pattern.
% Updates pattern by embedding the switch type.
% What: The entire top-level pattern, for error purposes.
:- pred check_pattern(context::in, progtable::in,
                    localtable::in, localtable::out,
                    varset::in, varset::out,
                    pattern::in, typeval::in,
                    pattern::in, pattern::out)
                    is det.
check_pattern(Ctx, PT, !LT, !Varset, What, SwitchType,
    pattern(Pattern0, _), pattern(Pattern, yes(SwitchType))) :-
    check_pattern_(Ctx, PT, !LT, !Varset, What, SwitchType,
        Pattern0, Pattern).

:- pred check_pattern_(context::in, progtable::in,
                    localtable::in, localtable::out,
                    varset::in, varset::out,
                    pattern::in, typeval::in,
                    pattern_::in, pattern_::out)
                    is det.
check_pattern_(_Ctx, _PT, !LT, !Varset, _What, _SwitchType, pat_any, pat_any).
check_pattern_(_Ctx, _PT, !LT, !Varset, _What, SwitchType,
        P@pat_var(Name), P) :-
    % Create a new variable in the localtable
    tables.update_localtable(Name, SwitchType, !LT).
check_pattern_(Ctx, PT, !LT, !Varset, What, SwitchType,
        P@pat_ctor(VarName, Args0), pat_ctor(VarName, Args)) :-
    ( tables.lookup_ctor(PT, VarName, TypeDef, _) ->
        % Found in global ctor table
        typedef_type(TypeDef, OutputType0),
        ( ctor_type(TypeDef, VarName, CtorVarset_, CtorType_, Ctor) ->
            CtorVarset = CtorVarset_,
            CtorType = CtorType_,
            CtorArgTypes0 = ctor_arg_types(Ctor)
        ;
            % Should not fail since we just looked it up
            error("check_pattern: ctor_type failed on known ctor")
        ),
        (
            CtorArgTypes0 = yes(YesCtorArgTypes0),
            % Augment our own varset with this ctor's varset and
            % translate the function's type to refer to the combined varset
            ( types.varset_merge(CtorVarset, !Varset,
                    [OutputType0|YesCtorArgTypes0],
                    [OutputType_|CtorArgTypes_]) ->
                OutputType = OutputType_, CtorArgTypes = CtorArgTypes_
            ;
                error("check_pattern: varset_merge returned bad output")
            ),
            % Make sure the pattern also has arguments
            ( list.length(CtorArgTypes) = 1 ->
                Arguments = " argument"
            ;
                Arguments = " arguments"
            ),
            (
                Args0 = yes(YesArgs0)
            ;
                Args0 = no,
                context.throw_error("Type error in pattern '"
                    ++ pretty.string_pattern(What) ++ "'"
                    ++ "\n        Term:     " ++ VarName
                    ++ "\n        Type:     "
                        ++ pretty.string_typeval(CtorVarset, CtorType)
                    ++ "\n        Expected: "
                        ++ string.from_int(list.length(CtorArgTypes))
                        ++ Arguments,
                    Ctx)
            ),
            % Make sure there are the same number of arguments in the pattern
            ( list.length(YesArgs0) : int = list.length(CtorArgTypes) ->
                true
            ;
                context.throw_error("Type error in pattern '"
                    ++ pretty.string_pattern(What) ++ "'"
                    ++ "\n        Term:     " ++ VarName
                    ++ "\n        Type:     "
                        ++ pretty.string_typeval(CtorVarset, CtorType)
                    ++ "\n        Expected: "
                        ++ string.from_int(list.length(CtorArgTypes))
                        ++ Arguments
                        ++ ", got " ++ string.from_int(list.length(YesArgs0)),
                    Ctx)
            ),
            % Make sure each argument matches the ctor argument type
            check_patterns(Ctx, PT, !LT, !Varset, What, CtorArgTypes,
                YesArgs0, YesArgs),
            Args = yes(YesArgs)
        ;
            CtorArgTypes0 = no,
            % Augment our own varset with this ctor's varset and
            % translate the function's type to refer to the combined varset
            ( types.varset_merge(CtorVarset, !Varset, [OutputType0],
                                                    [OutputType_]) ->
                OutputType = OutputType_
            ;
                error("check_pattern: varset_merge returned bad output")
            ),
            % Make sure the pattern is also parameter-less
            (
                Args0 = yes(_),
                context.throw_error("Type error in pattern '"
                    ++ pretty.string_pattern(What) ++ "'"
                    ++ "\n        Term:     " ++ VarName
                    ++ "\n        Type:     "
                        ++ pretty.string_typeval(CtorVarset, CtorType)
                    ++ "\n        Arguments not allowed",
                    Ctx)
            ;
                Args0 = no
            ),
            Args = no
        ),
        % Finally, check that the constructor matches the given switch type.
        ( types.unify(OutputType, SwitchType, !Varset) ->
            true
        ;
            type_error_pattern(Ctx, !.Varset, What, pattern(P,no), OutputType,
                               SwitchType)
        )
    ;
        context.throw_error("Undefined constructor: " ++ VarName, Ctx)
    ).
check_pattern_(Ctx, _PT, !LT, !Varset, What, SwitchType0,
        P@pat_intlit(_), P) :-
    deref(!.Varset, SwitchType0, SwitchType),
    ( SwitchType = types.const("Int") ->
        true
    ;
        type_error_pattern(Ctx, !.Varset, What, pattern(P,no),
                           types.const("Int"), SwitchType)
    ).

% Check a list of patterns against a list of switch types.
% Note it stops when either list reaches empty. Should check lengths
% separately.
:- pred check_patterns(context::in, progtable::in,
                    localtable::in, localtable::out,
                    varset::in, varset::out,
                    pattern::in, list(typeval)::in,
                    list(pattern)::in, list(pattern)::out)
                    is det.
check_patterns(_Ctx, _PT, !LT, !Varset, _What, [], Ps, Ps).
check_patterns(_Ctx, _PT, !LT, !Varset, _What, [_|_], [], []).
check_patterns(Ctx, PT, !LT, !Varset, What, [S|Ss], [P0|Ps0], [P|Ps]) :-
    check_pattern(Ctx, PT, !LT, !Varset, What, S, P0, P),
    check_patterns(Ctx, PT, !LT, !Varset, What, Ss, Ps0, Ps).

:- func ctor_arg_types(ctor) = maybe(list(typeval)).
ctor_arg_types(ctor(_,no,_)) = no.
ctor_arg_types(ctor(_,yes(Args),_)) = yes(map(pair.snd, Args)).

check_stmt(PT, LT, _RetType, !Varset,
    basic_stmt(Stmt0, Ctx), basic_stmt(Stmt, Ctx)) :-
    check_basic_stmt(Ctx, PT, LT, !Varset, Stmt0, Stmt).
check_stmt(PT, LT, RetType, !Varset,
    compound_stmt(Stmt0, Ctx), compound_stmt(Stmt, Ctx)) :-
    check_compound_stmt(Ctx, PT, LT, RetType, !Varset, Stmt0, Stmt).

check_basic_stmt(_Ctx, _PT, _LT, !Varset, pass, pass).
check_basic_stmt(Ctx, PT, LT, !Varset,
        assign(VarName, Expr0), assign(VarName, Expr)) :-
    % Get the type of both sides
    get_local_var(Ctx, PT, LT, VarName, TargetType),  % Throws any errors
    expr_type(Ctx, PT, LT, !Varset, Expr0, Expr, ExprType),
    What = "statement '" ++ ir.varname_to_string_noescape(VarName)
        ++ " = " ++ pretty.string_expr(Expr) ++ "'",
    % Unify: TargetType = ExprType
    det_unify(Ctx, What, Expr, ExprType, TargetType, !Varset).
check_basic_stmt(Ctx, _PT, _LT, !Varset,
            fieldset(Target0, FieldName, Expr0),
            fieldset(Target0, FieldName, Expr0)) :-
    context.throw_error("Sorry: Type checker can't handle fieldset", Ctx).
check_basic_stmt(Ctx, _PT, _LT, !Varset,
            fieldreplace(Target0, FieldName, Expr0),
            fieldreplace(Target0, FieldName, Expr0)) :-
    context.throw_error("Sorry: Type checker can't handle fieldreplace", Ctx).
check_basic_stmt(Ctx, PT, LT, !Varset, eval(Expr0), eval(Expr)) :-
    expr_type(Ctx, PT, LT, !Varset, Expr0, Expr, _).

:- pred check_compound_stmt(context::in, progtable::in, localtable::in,
            typeval::in, varset::in, varset::out,
            compound_stmt::in, compound_stmt::out) is det.
check_compound_stmt(Ctx, PT, LT, Ret, !Varset, return(Expr0), return(Expr)) :-
    expr_type(Ctx, PT, LT, !Varset, Expr0, Expr, ExprType),
    What = "statement 'return " ++ pretty.string_expr(Expr) ++ "'",
    % Unify with function return type
    det_unify(Ctx, What, Expr, ExprType, Ret, !Varset).
check_compound_stmt(Ctx, PT, LT, Ret, !Varset, switch(Expr0, Cases0),
            switch(Expr, Cases)) :-
    expr_type(Ctx, PT, LT, !Varset, Expr0, Expr, SwitchType),
    check_cases(PT, LT, Ret, SwitchType, !Varset, Cases0, Cases).
check_compound_stmt(Ctx, PT, LT, Ret, !Varset,
            if_then_else(Cond0, ThenPart0, ElsePart0),
            if_then_else(Cond, ThenPart, ElsePart)) :-
    expr_type(Ctx, PT, LT, !Varset, Cond0, Cond, CondType),
    What = "statement 'if " ++ pretty.string_expr(Cond) ++ ":'",
    % Unify with Int type (all conditions must be Int).
    det_unify(Ctx, What, Cond, CondType, types.const("Int"), !Varset),
    check_stmts(PT, LT, Ret, !Varset, ThenPart0, ThenPart),
    check_stmts(PT, LT, Ret, !Varset, ElsePart0, ElsePart).
check_compound_stmt(Ctx, PT, LT, Ret, !Varset, while(Cond0, Stmts0),
            while(Cond, Stmts)) :-
    expr_type(Ctx, PT, LT, !Varset, Cond0, Cond, CondType),
    What = "statement 'while " ++ pretty.string_expr(Cond) ++ ":'",
    % Unify with Int type (all conditions must be Int).
    det_unify(Ctx, What, Cond, CondType, types.const("Int"), !Varset),
    check_stmts(PT, LT, Ret, !Varset, Stmts0, Stmts).

% check_stmts(PT, LT, RetType, !Varset, !Stmts).
:- pred check_stmts(progtable::in, localtable::in, typeval::in,
    varset::in, varset::out, list(stmt)::in, list(stmt)::out) is det.
check_stmts(_,_,_,!Varset,[],[]).
check_stmts(PT, LT, RetType, !Varset, [Stmt0|Stmts0],[Stmt|Stmts]) :-
    check_stmt(PT, LT, RetType, !Varset, Stmt0, Stmt),
    check_stmts(PT, LT, RetType, !Varset, Stmts0, Stmts).

% check_cases(PT, LT, RetType, SwitchType, !Varset, !Stmts).
:- pred check_cases(progtable::in, localtable::in,
    typeval::in, typeval::in, varset::in, varset::out,
    list(case_stmt)::in,list(case_stmt)::out) is det.
check_cases(_,_,_,_,!Varset,[],[]).
check_cases(PT, LT, RetType, SwitchType, !Varset,
        [case_stmt(Pattern0, Stmts0, Ctx)|Cases0],
        [case_stmt(Pattern, Stmts, Ctx)|Cases]) :-
    check_pattern(Ctx, PT, LT, LT_Inside, !Varset,
                  Pattern0, SwitchType, Pattern0, Pattern),
    % Note: Checking the statements uses the new LT which contains the
    % newly-bound variables by this pattern.
    check_stmts(PT, LT_Inside, RetType, !Varset, Stmts0, Stmts),
    check_cases(PT, LT, RetType, SwitchType, !Varset, Cases0, Cases).

% Gets the name of a local variable, giving an appropriate error if not found.
% (Note: Takes a progtable so it gives a different error if it's found
% globally than not at all).
:- pred get_local_var(context::in, progtable::in, localtable::in,
    varname::in, typeval::out) is det.
get_local_var(Ctx, PT, LT, VarName, Type) :-
    ( tables.lookup_local(LT, VarName, Type_) ->
        % Found in local table
        Type = Type_
    ;
        VarName = svname(UnqualVarName),
        tables.lookup_function(PT, UnqualVarName, _)
    ->
        % Found in global function table - cannot assign
        context.throw_error("Cannot assign global variable: "
            ++ UnqualVarName, Ctx)
    ;
        context.throw_error("Undefined variable: "
            ++ ir.varname_to_string_noescape(VarName), Ctx)
    ).

check_function(_PT, _LT, !Func) :-
    !.Func^func_body = func_builtin.
check_function(PT, LT, !Func) :-
    !.Func^func_body = func_body_ast(Decls, Stmts0),
    Ctx = !.Func^func_context,
    % Type/kind check the params, ret type and decls - must all be of kind *
    (
        !.Func^func_params = no
    ;
        !.Func^func_params = yes(Params),
        check_types_star(Ctx, PT, !.Func^func_varset,
                         list.map(func({_,B,_})=B, Params))
    ),
    check_type_star(Ctx, PT, !.Func^func_varset, !.Func^func_ret_type),
    check_types_star(Ctx, PT, !.Func^func_varset, list.map(pair.snd, Decls)),
    % Universally-quantify all the variables in Params, by binding them to
    % constants of the same name. (Note: It is important to ensure no name
    % conflict with actual type names, but this is already ensured by having
    % type vars in a different namespace).
    some [!Varset] (
        !:Varset = !.Func^func_varset,
        (
            !.Func^func_params = no
        ;
            !.Func^func_params = yes(Params2),
            foldl(bind_type, list.map(func({_,B,_})=B, Params2), !Varset)
        ),
        bind_type(!.Func^func_ret_type, !Varset),
        check_stmts(PT, LT, !.Func^func_ret_type, !Varset, Stmts0, Stmts),
        !Func ^ func_varset := !.Varset
    ),
    !Func ^ func_body := func_body_ast(Decls, Stmts).
% TODO: Type-check CFG-bodied functions.
% This is difficult due to existential type. For now, does not do any checking
% or annotation!
check_function(_PT, _LT, !Func) :-
    !.Func^func_body = func_body_cfg(_,_),
    error("Not Implemented: Type checking on a function in CFG form").

% bind_type(Type, !Varset).
% Bind all variables in the Type to a const type of the same name.
:- pred bind_type(typeval::in, varset::in, varset::out) is det.
bind_type(Type, !Varset) :-
    ( Type = variable(V) ->
        Term = types.const(types.varset_lookup_name(!.Varset, V)),
        types.varset_bind_var(V, Term, !Varset)
    ; Type = app(T1, T2) ->
        bind_type(T1, !Varset),
        foldl(bind_type, T2, !Varset)
    ;
        true
    ).

check_typedef(PT, Typedef) :-
    check_ctors(PT, Typedef ^ typedef_varset, Typedef ^ typedef_ctors).

:- pred check_ctors(progtable::in, varset::in, list(ctor)::in) is det.
check_ctors(_PT, _Varset, []).
check_ctors(PT, Varset, [C|Cs]) :-
    % Typecheck all types in the ctor, and make sure they have kind *
    Args = C ^ ctor_args,
    (
        Args = no
    ;
        Args = yes(YesArgs),
        check_types_star(C ^ ctor_context, PT, Varset,
            list.map(pair.snd, YesArgs))
    ),
    check_ctors(PT, Varset, Cs).

check_program(PT, program(Nodes0), program(Nodes)) :-
    check_program_(PT, Nodes0, Nodes).

:- pred check_program_(progtable::in, list(program_node)::in,
            list(program_node)::out) is det.
check_program_(_PT, [], []).
check_program_(PT, [Node0|Nodes0], [Node|Nodes]) :-
    (
        Node0 = pfunction(Func0),
        tables.build_localtable(Func0, LT),
        check_function(PT, LT, Func0, Func),
        Node = pfunction(Func)
    ;
        Node0 = ptypedef(Typedef),
        check_typedef(PT, Typedef),
        Node = Node0
    ;
        Node0 = pimport(_),     % Can't check
        % (XXX: should this be a software error?)
        Node = Node0
    ),
    check_program_(PT, Nodes0, Nodes).

det_unify(Ctx, What, Term, T1, T2, !Varset) :-
    ( types.unify(T1, T2, !Varset) ->
        true
    ;
        type_error(Ctx, !.Varset, What, Term, T1, T2)
    ).

% det_unify_lists(Context, What, Term, Type1, Type2, !Varset).
% Unify two lists types with respect to a single varset.
% Zips with det_unify across 3 lists (with Term corresponding to Type1).
% Precond: All 3 lists are the same size
:- pred det_unify_lists(context::in, string::in, list(expr)::in,
                  list(typeval)::in, list(typeval)::in,
                  varset::in, varset::out) is det.
det_unify_lists(_Ctx, _What, [], _, _, !Varset).
det_unify_lists(_Ctx, _What, [_|_], [], _, !Varset).
det_unify_lists(_Ctx, _What, [_|_], [_|_], [], !Varset).
det_unify_lists(Ctx, What, [Term|Terms], [T1|T1s], [T2|T2s], !Varset) :-
    det_unify(Ctx, What, Term, T1, T2, !Varset),
    det_unify_lists(Ctx, What, Terms, T1s, T2s, !Varset).

% type_error(Context, Varset, What, Term, Term_Type, Other_Type).
% Throws an error resulting from failing to unify Type1 and Type2 with
% respect to a particular Varset.
% "What" is a string describing the whole thing we have a problem with.
%   eg. "expression 'x'"
%   or  "statement 'x = 4'"
:- pred type_error(context::in, varset::in, string::in, expr::in,
    typeval::in,typeval::in) is erroneous.
type_error(Ctx, Varset, What, Term, T1, T2) :-
    context.throw_error("Type error in " ++ What
        ++ "\n        Term:     " ++ pretty.string_expr(Term)
        ++ "\n        Type:     " ++ pretty.string_typeval(Varset, T1)
        ++ "\n        Expected: " ++ pretty.string_typeval(Varset, T2),
        Ctx).

% type_error_pattern(Context, Varset, What, Term, Term_Type, Other_Type).
% Throws an error resulting from failing to unify Type1 and Type2 with
% respect to a particular Varset.
% "What" is the whole pattern, Term is the particular piece of the pattern
% which failed.
:- pred type_error_pattern(context::in, varset::in, pattern::in, pattern::in,
    typeval::in,typeval::in) is erroneous.
type_error_pattern(Ctx, Varset, What, Term, T1, T2) :-
    context.throw_error("Type error in pattern '"
        ++ pretty.string_pattern(What) ++ "'"
        ++ "\n        Term:     " ++ pretty.string_pattern(Term)
        ++ "\n        Type:     " ++ pretty.string_typeval(Varset, T1)
        ++ "\n        Expected: " ++ pretty.string_typeval(Varset, T2),
        Ctx).

% type_error_arity(Context, Varset, What, Term, Term_Type, Arity, NumGot).
% Throws an error resulting from applying the wrong number of arguments to a
% function.
% "What" is a string describing the whole thing we have a problem with.
%   eg. "expression 'x'"
%   or  "statement 'x = 4'"
:- pred type_error_arity(context::in, varset::in, string::in, expr::in,
    typeval::in, int::in, int::in) is erroneous.
type_error_arity(Ctx, Varset, What, Term, Term_Type, Arity, NumArgs) :-
    ( Arity = 1 ->
        Arguments = " argument"
    ;
        Arguments = " arguments"
    ),
    context.throw_error("Type error in " ++ What
        ++ "\n        Term:     " ++ pretty.string_expr(Term)
        ++ "\n        Type:     " ++ pretty.string_typeval(Varset, Term_Type)
        ++ "\n        Expected: " ++ string.from_int(Arity) ++ Arguments
            ++ ", got " ++ string.from_int(NumArgs),
        Ctx).

% error_not_a_function(Context, Varset, What, Term, Term_Type).
% Throws an error resulting from applying the wrong number of arguments to a
% function.
% "What" is a string describing the whole thing we have a problem with.
%   eg. "expression 'x'"
%   or  "statement 'x = 4'"
:- pred error_not_a_function(context::in, varset::in, string::in, expr::in,
    typeval::in) is erroneous.
error_not_a_function(Ctx, Varset, What, Term, Term_Type) :-
    context.throw_error("Type error in " ++ What
        ++ "\n        Term:     " ++ pretty.string_expr(Term)
        ++ "\n        Type:     " ++ pretty.string_typeval(Varset, Term_Type)
        ++ "\n        Not a function",
        Ctx).

% error_not_a_type_function(Ctx, Varset, What, Type, Kind)
% Gives "too many arguments" error, after stripping away all arguments
:- pred error_not_a_type_function(context::in, varset::in, typeval::in,
    typeval::in, kind::in) is erroneous.
error_not_a_type_function(Ctx, Varset, What, Type0, Kind) :-
    deref(Varset, Type0, Type),
    context.throw_error("Kind error in type '"
        ++ pretty.string_typeval(Varset, What) ++ "'"
        ++ "\n        Type:     " ++ pretty.string_typeval(Varset, Type)
        ++ "\n        Kind:     " ++ pretty.string_kind(Kind)
        ++ "\n        Type does not accept parameters",
        Ctx).

% kind_arrow_expects(Arity, IsVariadic)
% Returns a description of how many arguments this kind expects
% (eg. "2 arguments", "at least 2 arguments")
:- func kind_arrow_expects(int, bool) = string.
kind_arrow_expects(Arity, IsVariadic) = S :-
    ( IsVariadic = yes ->
        VarText = "at least "
    ;
        VarText = ""
    ),
    ( Arity = 1 ->
        Arguments = " argument"
    ;
        Arguments = " arguments"
    ),
    S = VarText ++ string.from_int(Arity) ++ Arguments.

% kind_arity_check(Context, Varset, What, Type, KindExpected, NumArgsGiven).
% Checks that NumArgsGiven matches the arity of KindExpected, and that
% KindExpected is a type constructor (arrow) kind.
% If not, throws an error.
% "What" is the type of the whole thing we have a problem with.
%   eg. List(List)
:- pred kind_arity_check(context::in, varset::in, typeval::in, typeval::in,
    kind::in, int::in) is det.
kind_arity_check(Ctx, Varset, What, Type, kind_star, _NumArgs) :-
    error_not_a_type_function(Ctx, Varset, What, Type, kind_star).
kind_arity_check(Ctx, Varset, What, Type, Kind@kind_arrow(Arity, IsVariadic),
    NumArgs) :-
    ( IsVariadic = yes ->
        ( NumArgs >= Arity ->
            Error = no
        ;
            Error = yes
        )
    ;
        ( NumArgs = Arity ->
            Error = no
        ;
            Error = yes
        )
    ),
    ( Error = yes ->
        context.throw_error("Kind error in type '"
            ++ pretty.string_typeval(Varset, What) ++ "'"
            ++ "\n        Type:     " ++ pretty.string_typeval(Varset, Type)
            ++ "\n        Kind:     " ++ pretty.string_kind(Kind)
            ++ "\n        Expected " ++ kind_arrow_expects(Arity, IsVariadic)
                ++ ", got " ++ string.from_int(NumArgs),
            Ctx)
    ;
        true
    ).

% kind_error(Context, Varset, What, Type, Type_Kind, Other_Kind).
% Throws an error resulting from failing to unify Kind1 and Kind1 with
% respect to a particular Varset.
% "What" is the type of the whole thing we have a problem with.
%   eg. List(List)
:- pred kind_error(context::in, varset::in, typeval::in, typeval::in,
    kind::in, kind::in) is erroneous.
kind_error(Ctx, Varset, What, Type, K1, K2) :-
    % If K2 is *, be more helpful
    ( K1 = kind_arrow(Arity, IsVariadic) ->
        MoreHelp = "\n        (Type requires "
            ++ kind_arrow_expects(Arity, IsVariadic) ++ ")"
    ;
        MoreHelp = ""
    ),
    context.throw_error("Kind error in type '"
        ++ pretty.string_typeval(Varset, What) ++ "'"
        ++ "\n        Type:     " ++ pretty.string_typeval(Varset, Type)
        ++ "\n        Kind:     " ++ pretty.string_kind(K1)
        ++ "\n        Expected: " ++ pretty.string_kind(K2) ++ MoreHelp,
        Ctx).
