% -------------------------------------------------------------------------- %
% Mars/Venus type data structures
% Copyright (C) 2008 Matt Giuca
% 29/5/2008
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- module types.

:- interface.

:- import_module term.  % For var.
:- import_module varset.
:- import_module list.
:- import_module string.

% typeval - A type value. This is any value in the type system.
% It can be a simple reference to a typedef, a partially or
% wholly-instantiated type, or a type variable.
% Also function types are special.
% Behaves like the following algebraic type:
%:- type typeval
%    --->    const(string)
%    ;       functype
%    ;       variable(var)
%    ;       app(typeval, typeval).
:- type typeval.

% Note: There is a subtle "leak" in the opaqueness of this type
% (for it is internally represented as a Mercury term).
% The vars stored inside typevals are regular Mercury vars, bound to
% term objects. This means that if you were to look inside a bound var,
% you would see a Mercury term. You should only bind vars to valid typevals.

% The functions that make up this "fake" ADT.

% const(Name) is a reference to a named type.
:- func const(string) = typeval.
:- mode const(in) = out is det.
:- mode const(out) = in is semidet.

% functype is the type (->).
% This isn't really anything special (other than syntactically). We make a
% special case of it just because of some refactoring going on which may
% require it.
:- func functype = typeval.
:- mode functype = out is det.
:- mode functype = in is semidet.

% variable(Var) is a type variable.
:- func variable(var) = typeval.
:- mode variable(in) = out is det.
:- mode variable(out) = in is semidet.

% app(T, Ts) yields the result of applying the types Ts to T.
% That is T(T1, T2, ..., Tn), where Ts is [T1, T2, ..., Tn].
:- func app(typeval, list(typeval)) = typeval.
:- mode app(in, in) = out is det.
:- mode app(out, out) = in is semidet.

% new_var(Type, !Varset) creates a new type, which is a new free variable.
% Updates varset to include this new variable.
:- pred new_var(typeval::out, varset::in, varset::out) is det.

% search_var(Varset, Var, Type) succeeds if Var is bound to Type in Varset.
:- pred search_var(varset::in, var::in, typeval::out) is semidet.

% deref(Varset, Type0, Type) succeeds if Type is Type0 dereferenced in Varset.
% If Type0 is not bound, makes no change.
% If Type0 is a type application, recursively dereferences where possible.
:- pred deref(varset::in, typeval::in, typeval::out) is det.

% Unify two types with respect to a single varset.
% Fails if the types can't unify.
% Succeeds if they do, and possibly updates varset with new bindings.
:- pred unify(typeval::in, typeval::in, varset::in, varset::out) is semidet.

% Get the name of a variable from a varset.
% Prefix some number with "t_" if not found.
:- func varset_lookup_name(varset, var) = string.

% Bind a variable to a typeval in a given varset.
:- pred varset_bind_var(var::in, typeval::in, varset::in, varset::out)
    is det.

% varset_merge(NewVarset, !Varset, !Types).
% Updates !Varset, merging all the variables from NewVarset into Varset.
% Also takes a list of Types in NewVarset and modifies them to refer to
% variables in the resulting !Varset.
:- pred varset_merge(varset::in, varset::in, varset::out,
        list(typeval)::in, list(typeval)::out) is det.

% varset_mergeone(NewVarset, !Varset, !Type).
% Like varset_merge, but works on a single type rather than a list.
:- pred varset_mergeone(varset::in, varset::in, varset::out,
        typeval::in, typeval::out) is det.

% function_arity(Type, Arity) succeeds if Type is a function type and Arity is
% the function's arity. Fails if Type is not a function type.
:- pred function_arity(typeval::in, int::out) is semidet.

% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %

:- implementation.

:- import_module int.
:- import_module require.

% Since we don't care about context, we just set all contexts to this
:- func blank_context = context.
blank_context = context("",0).

:- type typeval == term.
% Notes on typeval.
% We would like our own specific representation (as described above),
% but it is impractical to reimplement all the features like our own type
% variables, varsets and unification algorithm.
% It's much easier to "cheat" and use Mercury's term
% representation. This means we need special rules on how to represent types,
% and provide an interface almost like an ADT.

% Below, you can see the definitions of const, variable, and app, which show
% how to map a typeval fake ADT onto a Mercury term.
% Any term which does not match one of these patterns is not a valid typeval.

const(S) = functor(atom(S), [], blank_context).
functype = functor(atom("->"), [Void], blank_context) :-
    % Void is needed to distinguish a function symbol from a const
    Void = functor(integer(0), [], blank_context).
variable(V) = term.variable(V, blank_context).
app(T, Ts) = functor(atom("a"), [T|Ts], blank_context).

new_var(types.variable(V), !Varset) :-
    varset.new_var(!.Varset, V, !:Varset).

search_var(Varset, Var, Type) :-
    varset.search_var(Varset, Var, Type).

deref(Varset, Type0, Type) :-
    ( Type0 = const(_) ->
        Type = Type0
    ; Type0 = functype ->
        Type = Type0
    ; Type0 = types.variable(V) ->
        ( types.search_var(Varset, V, Type1) ->
            deref(Varset, Type1, Type)
        ;
            % Free variable - just leave as-is
            Type = Type0
        )
    ; Type0 = app(T, Ts) ->
        deref(Varset, T, T_),
        map(deref(Varset), Ts, Ts_),
        Type = app(T_, Ts_)
    ;
        error("deref: Invalid type value")
    ).

unify(T1, T2, !Varset) :-
    some [!Sub] (
        varset.get_bindings(!.Varset, !:Sub),
        term.unify_term(T1, T2, !Sub),   % Can fail
        varset.set_bindings(!.Varset, !.Sub, !:Varset)
    ).

varset_lookup_name(Varset, Var) = varset.lookup_name(Varset, Var, "t_").

varset_bind_var(Var, Type, !Varset) :-
    varset.bind_var(!.Varset, Var, Type, !:Varset).

varset_merge(NewVarset, !Varset, !Types) :-
    varset.merge(!.Varset, NewVarset, !.Types, !:Varset, !:Types).

varset_mergeone(NewVarset, !Varset, !Type) :-
    ( varset.merge(!.Varset, NewVarset, [!.Type], !:Varset, [!:Type]) ->
        true
    ;
        % Can only fail for 1 reason - the last arg is not exactly 1 element.
        % (Which it should never be).
        require.error("varset_mergeone: varset.merge returned the wrong "
                        ++ "number of elements")
    ).

function_arity(Type, Arity) :-
    Type = app(functype, Args@[_|_]),   % Can fail
    Arity = list.length(Args) - 1.      % -1 for return type
