% -------------------------------------------------------------------------- %
% Mars/Venus Implementation
% Copyright (C) 2008 Matt Giuca
% 10/11/2008
%
% Utilities
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %
% -------------------------------------------------------------------------- %
% Various general utility functions.
% -------------------------------------------------------------------------- %

:- module util.

:- interface.

:- import_module list.
:- import_module exception.
:- import_module io.
:- import_module string.

% Given two lists of equal length, combine each pair with a predicate and
% construct the output list.
% Fails if one list is shorter than the other.
:- pred zipwith(pred(L1, L2, A), list(L1), list(L2), list(A)).
:- mode zipwith(pred(in, in, out) is det, in, in, out) is semidet.
:- mode zipwith(pred(in, in, out) is semidet, in, in, out) is semidet.

% Given two lists of equal length, combine each pair with a predicate and
% construct the output list.
% If one list is shorter than the other, the output list is as long as the
% shortest list.
:- pred zipwith_det(pred(L1, L2, A), list(L1), list(L2), list(A)).
:- mode zipwith_det(pred(in, in, out) is det, in, in, out) is det.
:- mode zipwith_det(pred(in, in, out) is semidet, in, in, out) is semidet.

% Given two lists of equal length, test each pair with a predicate.
% Fails if one list is shorter than the other.
:- pred zipwith0(pred(L1, L2), list(L1), list(L2)).
:- mode zipwith0(pred(in, in) is semidet, in, in) is semidet.

:- pred list_foldr2(pred(L, A, A, Z, Z), list(L), A, A, Z, Z).
:- mode list_foldr2(pred(in, in, out, di, uo) is det, in, in, out, di, uo)
    is det.
% Add more modes as required.

:- pred list_foldr3(pred(L, A, A, B, B, C, C), list(L), A, A, B, B, C, C).
:- mode list_foldr3(pred(in, in, out, in, out, in, out) is det, in, in, out,
    in, out, in, out) is det.
:- mode list_foldr3(pred(in, in, out, in, out, di, uo) is det, in, in, out,
    in, out, di, uo) is det.
% Add more modes as required.

:- pred list_map_foldr(pred(L, L, A, A), list(L), list(L), A, A).
:- mode list_map_foldr(pred(in, out, in, out) is det, in, out, in, out)
    is det.
% Add more modes as required.

:- pred reverse_append(list(T)::in, list(T)::in, list(T)::out) is det.

% Repeatedly applies a predicate to an accumulator, a given number of times.
:- pred repeat(pred(A, A), int, A, A).
:- mode repeat(pred(in, out) is det, in, in, out) is det.
:- mode repeat(pred(in, out) is semidet, in, in, out) is semidet.
:- mode repeat(pred(in, out) is multi, in, in, out) is multi.
:- mode repeat(pred(in, out) is nondet, in, in, out) is nondet.
:- mode repeat(pred(di, uo) is det, in, di, uo) is det.
:- mode repeat(pred(di, uo) is semidet, in, di, uo) is semidet.

% A semidet predicate which always succeeds.
% (Useful for tricking the determinism system into thinking something is
% semidet when it's det, or nondet when it's multi, to avoid warnings).
:- pred always_succeeds_semidet is semidet.

% split_ext(Filename, Base, Ext).
% Filename = Base ++ "." ++ Ext. Fails if there is no "." in Filename after
% the last directory separator.
:- pred split_ext(string::in, string::out, string::out) is semidet.

% try_io_thread1(Pred, ExcResult, !SV, !IO).
% Like try_io, but threads an extra state variable through, as well as IO.
% If an exception is thrown, the state variable is not updated.
% Note: We can't thread unique objects, because it might fail and we'd have to
% backtrack.
:- pred try_io_thread1(pred(T, S, S, io, io), exception_result(T),
    S, S, io, io).
:- mode try_io_thread1(pred(out, in, out, di, uo) is det,
    out(cannot_fail), in, out, di, uo) is cc_multi.
:- mode try_io_thread1(pred(out, in, out, di, uo) is cc_multi,
    out(cannot_fail), in, out, di, uo) is cc_multi.

% try_io_thread2(Pred, ExcResult, !SV1, !SV2, !IO).
% Like try_io, but threads two extra state variables through, as well as IO.
% If an exception is thrown, the state variables are not updated.
% Note: We can't thread unique objects, because it might fail and we'd have to
% backtrack.
:- pred try_io_thread2(pred(T, S, S, U, U, io, io), exception_result(T),
    S, S, U, U, io, io).
:- mode try_io_thread2(pred(out, in, out, in, out, di, uo) is det,
    out(cannot_fail), in, out, in, out, di, uo) is cc_multi.
:- mode try_io_thread2(pred(out, in, out, in, out, di, uo) is cc_multi,
    out(cannot_fail), in, out, in, out, di, uo) is cc_multi.

% Call an IO predicate without requiring an IO state.
% This is marked impure, because it may have side-effects which are not
% IO-threaded.
:- impure pred invoke_io(pred(io, io)).
:- mode invoke_io(pred(di, uo) is det) is det.

:- impure pred invoke_io1(pred(T, io, io), T).
:- mode invoke_io1(pred(out, di, uo) is det, out) is det.
:- mode invoke_io1(pred(uo, di, uo) is det, uo) is det.

% Like io.write_string, but does not thread IO state.
% For debug use.
% Note: Technically impure, but not declared so to avoid having to propagate
% impurity upwards, just for debugging.
:- pred debug_write_string(string::in) is det.

% Like io.write, but does not thread IO state.
% For debug use.
:- pred debug_write(T::in) is det.

% debug_writef(Format, Params).
% Wrapper for debug_write_string(string.format(Format, Params)).
:- pred debug_writef(string::in, list(string.poly_type)::in) is det.

% Like io.nl, but does not thread IO state.
% For debug use.
:- pred debug_nl is det.

% Get the approximate address of the current stack frame, as an int.
% For debug use.
:- pred debug_get_stackframe_addr(int::out) is det.

:- implementation.

:- import_module pair.
:- import_module int.
:- import_module char.
:- import_module dir.

zipwith(_Pred, [], [], []).
zipwith(Pred, [X|Xs], [Y|Ys], [A|As]) :-
    Pred(X, Y, A),
    zipwith(Pred, Xs, Ys, As).

zipwith_det(_Pred, [], [], []).
zipwith_det(_Pred, [], [_|_], []).
zipwith_det(_Pred, [_|_], [], []).
zipwith_det(Pred, [X|Xs], [Y|Ys], [A|As]) :-
    Pred(X, Y, A),
    zipwith_det(Pred, Xs, Ys, As).

zipwith0(_Pred, [], []).
zipwith0(Pred, [X|Xs], [Y|Ys]) :-
    Pred(X, Y),
    zipwith0(Pred, Xs, Ys).

list_foldr2(_, [], !A, !Z).
list_foldr2(P, [H | T], !A, !Z) :-
    list_foldr2(P, T, !A, !Z),
    P(H, !A, !Z).

list_foldr3(_, [], !A, !B, !C).
list_foldr3(P, [H | T], !A, !B, !C) :-
    list_foldr3(P, T, !A, !B, !C),
    P(H, !A, !B, !C).

list_map_foldr(_, [], [], !A).
list_map_foldr(P, [H0 | T0], [H | T], !A) :-
    list_map_foldr(P, T0, T, !A),
    P(H0, H, !A).

reverse_append([], L, L).
reverse_append([X|Xs], L0, L) :-
    reverse_append(Xs, [X|L0], L).

repeat(Pred, Count, !A) :-
    ( Count =< 0 ->
        true
    ;
        Pred(!A),
        repeat(Pred, Count-1, !A)
    ).

always_succeeds_semidet :-
    % Pass 0 to the id function, then check that the result equals 0.
    % The determinism checker doesn't know that it will always be true, so it
    % infers this as semidet. However, logically it will always succeed.
    (func(X) = X)(0) = 0.

split_ext(Filename, Base, Ext) :-
    % Drop the directory part, then get the num of chars after the last dot.
    dir.basename(Filename, DirBase),        % Fails if no filename part
    Dot = char.det_from_int(46),    % ?? '.' is a syntax error?
    ExtLen = string.suffix_length((pred(C::in) is semidet :- C \= Dot),
                                  DirBase),
    ExtLen < string.length(DirBase),        % Fails if no "."
    Base = string.left(Filename, string.length(Filename) - ExtLen - 1),
    Ext = string.right(Filename, ExtLen).

try_io_thread1(Pred, ExcResult, !SV, !IO) :-
    % Make a Pred fit for try_io.
    TryPred = (pred(Result-SVOut::out, I::di, O::uo) is cc_multi :-
        Pred(Result, !.SV, SVOut, I, O)),
    try_io(TryPred, TryResult, !IO),
    (
        TryResult = succeeded(InnerResult - !:SV),
        ExcResult = succeeded(InnerResult)
    ;
        TryResult = exception(E),
        ExcResult = exception(E)
    ).

try_io_thread2(Pred, ExcResult, !SV1, !SV2, !IO) :-
    % Make a Pred fit for try_io.
    TryPred = (pred({Result,SVOut1,SVOut2}::out, I::di, O::uo) is cc_multi :-
        Pred(Result, !.SV1, SVOut1, !.SV2, SVOut2, I, O)),
    try_io(TryPred, TryResult, !IO),
    (
        TryResult = succeeded({InnerResult, !:SV1, !:SV2}),
        ExcResult = succeeded(InnerResult)
    ;
        TryResult = exception(E),
        ExcResult = exception(E)
    ).

invoke_io(Pred) :-
    % Create a new IO state out of nothing
    impure fabricate_io_state(IO),
    % Call the predicate with this state as input; forget the result state
    Pred(IO, _).

invoke_io1(Pred, V) :-
    % Create a new IO state out of nothing
    impure fabricate_io_state(IO),
    % Call the predicate with this state as input; forget the result state
    Pred(V, IO, _).

% Create a unique IO world without having one as input.
% This is marked impure, because it can be used to generate side-effects which
% are not IO-threaded.
:- impure pred fabricate_io_state(io::uo) is det.
:- pragma foreign_proc("C",
    fabricate_io_state(IO::uo),
    [will_not_call_mercury, thread_safe, terminates, no_sharing],
"
    /* Just create a new IO value out of nothing.
     * IO state values are just 0-valued ints. */
    IO = 0
").

:- pragma promise_pure(debug_write_string/1).
debug_write_string(Str) :-
    impure invoke_io(io.write_string(Str)).

debug_write(X) :-
    XStr = string.string(X),
    debug_write_string(XStr).

debug_writef(Format, Params) :-
    debug_write_string(string.format(Format, Params)).

debug_nl :-
    debug_write_string("\n").

:- pragma foreign_proc("C",
    debug_get_stackframe_addr(Addr::out),
    [will_not_call_mercury, thread_safe, terminates, no_sharing,
     promise_pure],
"
    /* We said we'd get an approximate address.
     * Just create a local variable and return its address.
     */

    int x;
    Addr = (MR_Integer) (&x);
").
