% -------------------------------------------------------------------------- %
% VectorRef Library
% Copyright (C) 2008 Matt Giuca
% 11/11/2008
%
% Provides functionality similar to the 'arrayref' module, but for 'vectors',
% which are arrays whose capacity may be mutably changed.
% -------------------------------------------------------------------------- %
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------- %

:- module vectorref.

:- interface.

:- import_module store.
:- import_module array.
:- import_module int.

% An vectorref(T, S) (extendable array) is much like an arrayref(T, S), but
% its capacity may be changed through mutation, and all references to the
% array will see the change.
% This allows the implementation of methods such as C++'s "vector::push_back"
% or Java's "Vector.add", which mutably append elements to the end of a
% vector.
%
% Unlike arrayrefs, vectors have both a size (number of elements stored in the
% vector) and a capacity (number of elements the vector supports without
% resizing). Capacity need not be directly spoken of, but is part of the
% public interface to allow for greater customization.
:- type vectorref(T, S).

% vectorref.init(VectorRef, !S).
% Creates a new vector in a store S, and returns a reference to it.
% The vector is initially empty, and has a reasonable amount of capacity.
:- pred vectorref.init(vectorref(T, S)::out, S::di, S::uo)
    is det <= store(S).
% vectorref.init(InitialCapacity, VectorRef, !S).
% Optional InitialCapacity sets the initial number of allocated elements (it
% does not affect the semantics of the vector).
:- pred vectorref.init(int::in, vectorref(T, S)::out, S::di, S::uo)
    is det <= store(S).

% vectorref.extract(VectorRef, Array, !S).
% Given a vector reference, extracts a copy of the array from the store.
% The resulting array is unique, and does not interfere with the store.
% Note that the array's length is the vector's SIZE, not its CAPACITY.
:- pred vectorref.extract(vectorref(T, S)::in, array(T)::array_uo, S::di, S::uo)
    is det <= store(S).

% Get the size of the vector (number of elements stored).
:- pred vectorref.size(vectorref(_T, S)::in, int::out, S::di, S::uo)
    is det <= store(S).

% Get the capacity of the vector (number of elements allocated for).
:- pred vectorref.capacity(vectorref(_T, S)::in, int::out, S::di, S::uo)
    is det <= store(S).

% vectorref.bounds(VectorRef, Min, Max, !Store).
% Get the minimum and maximum accessible bounds of the vector.
% Min is 0, Max is size - 1.
:- pred vectorref.bounds(vectorref(_T, S)::in, int::out, int::out,
    S::di, S::uo) is det <= store(S).

% vectorref.lookup returns the Nth element of a vector.
% Throws an exception if the index is out of bounds.
:- pred vectorref.lookup(vectorref(T, S)::in, int::in, T::out, S::di, S::uo)
    is det <= store(S).

% vectorref.set sets the Nth element of a vector, destructively modifying the
% vector through the store.
% Throws an exception if the index is out of bounds.
:- pred vectorref.set(vectorref(T, S)::in, int::in, T::in, S::di, S::uo)
    is det <= store(S).

% vectorref.add appends an element to the end of the vector, destructively
% modifying the vector through the store. This increases the vector's size by
% 1, and may also increase its capacity.
:- pred vectorref.add(vectorref(T, S)::in, T::in, S::di, S::uo)
    is det <= store(S).

% vectorref.remove removes an element from the vector, ad a given index,
% destructively modifying the vector through the store. This decreases the
% vector's size by 1.
:- pred vectorref.remove(vectorref(T, S)::in, int::in, S::di, S::uo)
    is det <= store(S).

% vectorref.copy(VectorRef0, VectorRef, !Store).
% Makes a new unique copy of a vector, and returns a reference to it.
% This makes a shallow copy.
:- pred vectorref.copy(vectorref(T, S)::in, vectorref(T, S)::out,
    S::di, S::uo) is det <= store(S).

% vectorref.ensure_capacity(VectorRef, MinCapacity, !Store).
% If VectorRef's capacity is less than `MinCapacity', extends the vector so
% its capacity is equal to `MinCapacity'.
% This has no semantic effect on the vector.
:- pred vectorref.ensure_capacity(vectorref(T, S)::in, int::in,
    S::di, S::uo) is det <= store(S).

%-----------------------------------------------------------------------------%
%-----------------------------------------------------------------------------%

:- implementation.

:- import_module arrayref.

:- import_module store.
:- import_module exception.
:- import_module string.
:- import_module list.

:- type vectorref(T, S) == store.generic_mutvar(vector(T, S), S).

% NOTE: For efficiency, the unused elements of vr_array are stored UNSAFELY as
% undefined values (using arrayref.unsafe_init/unsafe_resize).
% The elements >= vr_size should not be accessed under any circumstances.
:- type vector(T, S)
    --->    vector(
                % vr_array's "size" is the vector's "capacity"
                vr_array :: arrayref.arrayref(T, S),
                % actual number of defined elements
                vr_size  :: int
            ).

% examine_vector(VectorRef, ArrayRef, Size, !Store).
% Pull the vr_array and vr_size out of a vectorref.
:- pred examine_vector(vectorref(T, S)::in, arrayref.arrayref(T, S)::out,
    int::out, S::di, S::uo) is det <= store(S).
examine_vector(VR, AR, Size, !S) :-
    store.get_mutvar(VR, vector(AR, Size), !S).

vectorref.init(V, !S) :-
    % By default, start with an initial capacity of 8 elements
    vectorref.init(8, V, !S).
vectorref.init(Cap, VR, !S) :-
    arrayref.unsafe_init(Cap, AR, !S),
    V = vector(AR, 0),
    store.new_mutvar(V, VR, !S).

vectorref.extract(VR, Array, !S) :-
    examine_vector(VR, AR, Size, !S),
    % Do not copy the whole array (that contains undefined elements above
    % size). Instead use shrink_array to create a copy which is exactly as
    % long as the vector's size.
    arrayref.shrink_array(AR, Size, Array, !S).

vectorref.size(VR, Size, !S) :-
    examine_vector(VR, _AR, Size, !S).

vectorref.capacity(VR, X, !S) :-
    examine_vector(VR, AR, _Size, !S),
    arrayref.size(AR, X, !S).

vectorref.bounds(VR, 0, Size-1, !S) :-
    examine_vector(VR, _AR, Size, !S).

vectorref.lookup(VR, N, V, !S) :-
    examine_vector(VR, AR, Size, !S),
    check_upper_bound(Size, N, "vectorref.lookup"),
    arrayref.lookup(AR, N, V, !S).

vectorref.set(VR, N, V, !S) :-
    examine_vector(VR, AR, Size, !S),
    check_upper_bound(Size, N, "vectorref.set"),
    arrayref.set(AR, N, V, !S).

vectorref.add(VR, V, !S) :-
    examine_vector(VR, AR0, Size, !S),
    arrayref.size(AR0, Capacity, !S),
    ( Size >= Capacity ->
        arrayref.unsafe_resize(AR0, Capacity*2, AR, !S)
    ;
        AR = AR0
    ),
    arrayref.set(AR, Size, V, !S),
    % Mutate the vectorref with the (potentially)-new AR and new Size.
    store.set_mutvar(VR, vector(AR, Size+1), !S).

vectorref.remove(VR, N, !S) :-
    examine_vector(VR, AR, Size, !S),
    check_upper_bound(Size, N, "vectorref.remove"),
    vectorref.remove_from(AR, N, Size-1, !S),
    % Mutate the vectorref with the new Size.
    store.set_mutvar(VR, vector(AR, Size-1), !S).

:- pred vectorref.remove_from(arrayref(T, S)::in, int::in, int::in,
    S::di, S::uo) is det <= store(S).
vectorref.remove_from(AR, N, Max, !S) :-
    ( N < Max ->
        arrayref.lookup(AR, N+1, V, !S),
        arrayref.set(AR, N, V, !S),
        vectorref.remove_from(AR, N+1, Max, !S)
    ;
        true
    ).

vectorref.copy(VR0, VR, !S) :-
    % Need to copy both the vector's mutvar, and the array itself
    examine_vector(VR0, AR0, Size, !S),
    arrayref.copy(AR0, AR, !S),
    store.new_mutvar(vector(AR, Size), VR, !S).

vectorref.ensure_capacity(VR, MinCapacity, !S) :-
    examine_vector(VR, AR0, Size, !S),
    arrayref.size(AR0, Capacity, !S),
    ( Capacity < MinCapacity ->
        arrayref.unsafe_resize(AR0, MinCapacity, AR, !S),
        % Mutate the vectorref with the new AR.
        store.set_mutvar(VR, vector(AR, Size), !S)
    ;
        true
    ).

%-----------------------------------------------------------------------------%

:- pred check_upper_bound(int::in, int::in, string::in) is det.
check_upper_bound(Size, Index, PredName) :-
    ( bounds_checks, Index >= Size ->
        out_of_bounds_error(Index, 0, Size-1, PredName)
    ;
        true
    ).

:- pred bounds_checks is semidet.
:- pragma inline(bounds_checks/0).

:- pragma foreign_proc("C",
    bounds_checks,
    [will_not_call_mercury, promise_pure, thread_safe, will_not_modify_trail],
"
#ifdef ML_OMIT_ARRAY_BOUNDS_CHECKS
    SUCCESS_INDICATOR = MR_FALSE;
#else
    SUCCESS_INDICATOR = MR_TRUE;
#endif
").

:- pragma foreign_proc("C#",
    bounds_checks,
    [will_not_call_mercury, promise_pure, thread_safe],
"
#if ML_OMIT_ARRAY_BOUNDS_CHECKS
    SUCCESS_INDICATOR = false;
#else
    SUCCESS_INDICATOR = true;
#endif
").

:- pragma foreign_proc("Java",
    bounds_checks,
    [will_not_call_mercury, promise_pure, thread_safe],
"
    // never do bounds checking for Java (throw exceptions instead)
    succeeded = false;
").

%-----------------------------------------------------------------------------%

    % throw an exception indicating an array bounds error
:- pred out_of_bounds_error(int, int, int, string).
:- mode out_of_bounds_error(in, in, in, in) is erroneous.

out_of_bounds_error(Index, Min, Max, PredName) :-
    throw(array.index_out_of_bounds(
        string.format("%s: index %d not in range [%d, %d]",
            [s(PredName), i(Index), i(Min), i(Max)]))).

%-----------------------------------------------------------------------------%
