#!/usr/bin/env python
# -------------------------------------------------------------------------- %
# Mars/Venus Compiler Test Framework
# Copyright (C) 2008 Matt Giuca
# 27/8/2008
# -------------------------------------------------------------------------- %
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -------------------------------------------------------------------------- %
# -------------------------------------------------------------------------- %

# Based on main function structure, by Guido van Rossum.
# http://www.artima.com/weblogs/viewpost.jsp?thread=4829

# Requires PyYAML
# http://pyyaml.org

# Note: `mars` must be in $PATH.
#       Any modules imported by the modules being tested must be in $MARSPATH.

import warnings
import sys
import getopt
import os, os.path
import subprocess
import logging
import random

import yaml

logging.basicConfig(level=logging.INFO,
                    format="%(levelname)s: %(message)s")

# Constants
# Use ANSI colour when printing
PRINT_COLOR = True

# Report all failures, even if expected
REPORT_FAILURES = False

# Globals - Overall pass/fail rate
num_as_expected = 0     # Also includes all passes
num_pass = 0    # Unexpected passes
num_fail = 0    # Unexpected fails
num_skip = 0    # Unexpected skips
num_total = 0
# outcomes is a list of (char, char, str) triples, logging:
# - the 'passed' character value
# - the 'expected' character value
# - reason for each case.
outcomes = []

class Usage(Exception):
    pass

class TestingError(Exception):
    """
    Some error with the test cases themselves (not failure, but badly defined
    test cases).
    """
    pass

# report_outcome passed values
PASS = 'P'
COMPILE_ERROR = 'E'
FAIL = 'F'
CANT_RUN = 'x'

ANSI_COLOR = {
    # ANSI colour code values for each outcome
    PASS:           '92',
    COMPILE_ERROR:  '91',
    FAIL:           '91',
    CANT_RUN:       '90',
}

rand_letters = ('abcdefghijklmnopqrstuvwxyz'
                'ABCDEFGHIJKLMNOPQRSTUVWXYZ')
def random_filename(prefix='', suffix=''):
    return (prefix
            + ''.join(random.sample(rand_letters, 1)[0] for _ in range(8))
            + suffix)

def report_outcome(passed=PASS, expected=PASS, reason=None):
    """
    Report pass or fail.
    Logs the pass or failure, and the reason, and prints a character to
    stdout.
    (Pass prints a '.', fail prints a 'F', compiler error prints a 'E').
    The reason is logged, and printed at the end.
    """
    global num_as_expected, num_pass, num_skip, num_fail, num_total, failures
    if PRINT_COLOR:
        sys.stdout.write('\033[%sm' % ANSI_COLOR[passed])
    if REPORT_FAILURES and passed == expected == FAIL:
        # If failed, and expected to fail, treat as if we expected a pass
        # (since we are reporting ALL failures)
        expected = PASS
    if passed == expected:
        # Print a '.' to indicate that this was expected
        sys.stdout.write('.')
    else:
        sys.stdout.write(passed)
    if PRINT_COLOR:
        sys.stdout.write('\033[0m')
    if passed == expected:
        num_as_expected += 1
    elif passed == PASS:
        num_as_expected += 1    # Also includes pass
        num_pass += 1
    elif passed == CANT_RUN:
        num_skip += 1
    else:
        num_fail += 1
    num_total += 1
    sys.stdout.flush()
    outcomes.append((passed, expected, reason))

def compile_mars(source, output, backend=None, main=None, environ=None):
    """
    Run the Mars compiler, and potentially a command or function as well.
    Waits until the compiler and execution are complete.
    source: Mars source filename.
    output: Output filename (temporary).
    backend: Backend string (--backend switch).
    main: Name of main function (-m switch).
    environ: Dictionary for Mars process's environment.
    Returns a triple (retcode, stdout, stderr), as returned by the
    compiler.
    Deletes the "output" file (which is a temporary).
    """
    args = ["marsc", "-o", output]
    if backend is not None:
        args.extend(["--backend", backend])
    if main is not None:
        args.extend(["-m", main])
    args.append(source)
    logging.debug(' '.join(args))
    p = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE, env=environ)
    stdout, stderr = p.communicate("")
    try:
        os.unlink(output)
    except OSError:
        # May not have been created at all (eg. error)
        pass
    return p.returncode, stdout, stderr

def run_mars(source, backend=None, main=None, command=None,
             environ=None, stdin=None):
    """
    Run the Mars interpreter, running a command or function as well.
    Waits until the compiler and execution are complete.
    source: Mars source filename.
    backend: Backend string (--backend switch).
    main: Name of main function (-m switch).
    command: Command to call (-c switch).
    environ: Dictionary for Mars process's environment.
    stdin: Bytes send to stdin.
    Returns a triple (retcode, stdout, stderr), as returned by the
    compiler/execution.
    """
    args = ["mars"]
    if backend is not None:
        args.extend(["--backend", backend])
    if main is not None:
        args.extend(["-m", main])
    if command is not None:
        args.extend(["-c", command])
    args.append(source)
    logging.debug(' '.join(args))
    p = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE, env=environ)
    stdout, stderr = p.communicate(stdin)
    return p.returncode, stdout, stderr

def do_test_real(marfilename, mtc):
    """
    marfilename: Name of .mar file to test.
    mtc: MTC dictionary (parsed from MTC source file).
    """
    outfile = random_filename(prefix="__", suffix=".py")
    retcode, stdout, stderr = compile_mars(marfilename, outfile)

    compile_success = False

    try:
        compile = mtc['compile']
        try:
            expect_outcome = compile['outcome']
        except KeyError:
            expect_outcome = 'succeed'
        try:
            expect_errors = compile['errors']
        except KeyError:
            expect_errors = ''
        try:
            expect = compile['expect']
        except KeyError:
            expect = 'pass'
    except KeyError:
        expect_outcome = 'succeed'
        expect_errors = ''
        expect = 'pass'

    if expect == 'pass':
        expect = PASS
    elif expect == 'fail':
        expect = FAIL
    elif expect == 'compile_error':
        expect = COMPILE_ERROR
    else:
        raise TestingError("Invalid 'expect' value: %s" % expect)

    if retcode == 0:
        if expect_outcome == 'succeed':
            # Succeed compile = Pass case
            report_outcome(PASS, expect)
            compile_success = True
        else:
            # Succeed compile = Fail case
            report_outcome(FAIL, expect,
                           "Expected: %s\nInstead, compiled successfully.\n"
                           % expect_errors)
    else:
        # Fail compile = Fail case
        if expect_outcome == 'succeed':
            # Fail compile = Fail case
            report_outcome(COMPILE_ERROR, expect, stderr)
        else:
            # Expected (and got) fail compile 
            # Check stderr matches
            if stderr == expect_errors:
                report_outcome(PASS, expect)
            else:
                report_outcome(FAIL, expect, "Expected: %s\nGot:      %s\n"
                                             % (expect_errors, stderr))

    # Runtime tests
    try:
        run = mtc['run']
    except KeyError:
        pass
    else:
        if compile_success:
            for runtimetest in run:
                do_runtime_test(marfilename, outfile, runtimetest)
        else:
            # Can't run any runtime cases. Report all as "can't run"
            for runtimetest in run:
                report_outcome(CANT_RUN, PASS)

    # Unit tests
    # First run with ":b" to get the names of all the functions
    _, stdout, _ = run_mars(marfilename, command=":ta")
    for func in stdout.split('\n'):
        try:
            fname, ftype = func.split('::')
        except ValueError:
            pass
        else:
            fname = fname.strip()
            ftype = ftype.strip()
            # All functions starting with "test_" are testable, and must
            # have type () -> Int or Int -> Int.
            if not fname.startswith("test_"):
                continue
            if ftype == "() -> Int":
                do_unit_test(marfilename, outfile, fname, "()")
            elif ftype == "Int -> Int":
                do_unit_test(marfilename, outfile, fname, "(0)")
            else:
                report_outcome(FAIL, PASS, "Test function %s does not have "
                    "type () -> Int or Int -> Int\n" % fname)

    print
    # Remove sundry ".pyc" files that have been created
    try:
        os.unlink(outfile + 'c')
    except OSError:
        # May not have been created at all (eg. error)
        pass

def do_runtime_test(marfilename, outfile, runtimetest):
    """
    Performs a single runtime test. Prints success code to the screen, and
    logs the outcome.
    marfilename: Filename of Mars file (should have successfully compiled).
    """
    call = runtimetest['call']
    try:
        new_environ = runtimetest['environ']
    except KeyError:
        new_environ = {}
    try:
        stdin = runtimetest['stdin']
    except KeyError:
        stdin = ""
    try:
        expect_stdout = runtimetest['stdout']
    except KeyError:
        expect_stdout = ""
    try:
        expect_stderr = runtimetest['stderr']
    except KeyError:
        expect_stderr = ""

    try:
        expect = runtimetest['expect']
    except KeyError:
        expect = 'pass'

    if expect == 'pass':
        expect = PASS
    elif expect == 'fail':
        expect = FAIL
    else:
        raise TestingError("Invalid 'expect' value: %s" % expect)

    # Get the union of the current os.environ and the specified environment
    environ = dict(os.environ)
    for k,v in new_environ.items():
        environ[k] = str(v)
    # Run Mars with -c the given call, passing environ and stdin
    _, stdout, stderr = run_mars(marfilename, command=call,
                                 environ=environ, stdin=stdin)
    if stdout != expect_stdout:
        if stderr != expect_stderr:
            report_outcome(FAIL, expect,
                           "Expected:\n%s(stderr):\n%sGot:\n%s(stderr):\n%s"
                           % (expect_stdout, expect_stderr, stdout, stderr))
        else:
            report_outcome(FAIL, expect, "Expected:\n%sGot:\n%s"
                                         % (expect_stdout, stdout))
    elif stderr != expect_stderr:
        report_outcome(FAIL, expect, "Expected (stderr):\n%sGot:\n%s"
                                     % (expect_stderr, stderr))
    else:
        report_outcome(PASS, expect)

def do_unit_test(marfilename, outfile, function_name, args="()"):
    """
    Calls a unit test function. Prints success code to the screen, and
    logs the outcome.
    marfilename: Filename of Mars file (should have successfully compiled).
    function_name: Name of test function to call (expected to produce lots of
        stdout).
    args: String of arguments to pass to function, including parens
        Defaults to "()".
    """
    # Run Mars with -c the given call, passing environ and stdin
    _, stdout, stderr = run_mars(marfilename, command=function_name + args)
    if len(stderr) > 0:
        report_outcome(FAIL, PASS, "Got stderr:\n%s\n" % stderr)
        return

    current_case = None     # Name of current test case
    expect = None
    case_already_failed = None

    for line in stdout.split('\n'):
        if not line.startswith(":test:"): continue
        if line.startswith(":test:begin_case:"):
            cont = line[17:]
            if current_case is not None:
                report_outcome(FAIL, PASS,
                    "Begin case, when one is already active\n")
                return
            current_case = cont
            expect = PASS
            case_already_failed = False
        elif line.startswith(":test:begin_case_pending:"):
            cont = line[25:]
            if current_case is not None:
                report_outcome(FAIL, PASS,
                    "Begin case, when one is already active\n")
                return
            current_case = cont
            expect = FAIL
            case_already_failed = False
        elif line.startswith(":test:end_case:"):
            cont = line[15:]
            if current_case is None:
                report_outcome(FAIL, PASS, "End case, when none active\n")
                return
            elif cont != current_case:
                report_outcome(FAIL, PASS, "End case for %r, while in %r\n"
                                % (cont, current_case))
                return
            if not case_already_failed:
                # We passed this case
                report_outcome(PASS, expect, "%s\n" % (current_case,))
            current_case = None
            expect = None
            case_already_failed = None
        elif line.startswith(":test:fail:"):
            cont = line[11:]
            if current_case is None:
                report_outcome(FAIL, PASS, "Fail, when no test case active\n")
                return
            if not case_already_failed:
                report_outcome(FAIL, expect, "%s: %s\n" % (current_case, cont))
                case_already_failed = True

def do_test(case):
    # Back up the cwd
    current_wd = os.getcwd()
    # Get the base name of the case (without extension)
    dir, fname = os.path.split(case)
    bcase, ext = os.path.splitext(fname)
    if ext == '.mar' or ext == '.mtc':
        case = bcase
    # Add extensions back on
    marfilename = case + ".mar"
    mtcfilename = case + ".mtc"

    mtcfile = None
    if len(dir) > 0:
        os.chdir(dir)       # May raise OSError

    if not os.path.exists(marfilename):
        raise IOError(2, "No such file or directory", marfilename)
    try:
        try:
            # MTC file is optional
            mtcfile = open(mtcfilename)
        except IOError:
            mtc = {}
        else:
            try:
                mtc = yaml.load(mtcfile.read())
            except:
                raise
            finally:
                mtcfile.close()

        sys.stdout.write("%s: " % case)
        do_test_real(marfilename, mtc)

    finally:
        os.chdir(current_wd)
        if mtcfile is not None: mtcfile.close()

def summary():
    # Printout at end
    for passed, expected, reason in outcomes:
        if passed != expected and passed != CANT_RUN:
            if PRINT_COLOR:
                sys.stdout.write('\033[%sm' % ANSI_COLOR[passed])
            if passed == COMPILE_ERROR:
                sys.stdout.write("Compile Error")
            elif passed == FAIL:
                sys.stdout.write("Fail")
            elif passed == PASS:
                sys.stdout.write("Unexpected Pass")
            else:
                sys.stdout.write("Unknown Outcome")
            if PRINT_COLOR:
                sys.stdout.write('\033[0m')
            if reason is None:
                sys.stdout.write("\n")
            else:
                sys.stdout.write(":\n")
                sys.stdout.write(reason)
    plural = lambda n: '' if n == 1 else 's'
    sys.stdout.write("%d/%d cases passed"
                     % (num_as_expected, num_total))
    if not REPORT_FAILURES:
        sys.stdout.write(" (or expected fail)")
    sys.stdout.write("\n")
    if num_pass > 0:
        sys.stdout.write("%d case%s passed unexpectedly\n"
            % (num_pass, plural(num_pass)))
    if num_skip > 0:
        sys.stdout.write("%d case%s skipped\n" % (num_skip, plural(num_skip)))
    if num_fail > 0:
        sys.stdout.write("%d case%s failed\n" % (num_fail, plural(num_fail)))

def usage():
    print >>sys.stderr, """\
Usage: test.py [OPTIONS] testcase
    testcase is a path to a test case without the .mar/.mtc extension.
    OPTIONS
        --report-fail   Report all failures, even if expected.\
"""

def main(args=None):
    global REPORT_FAILURES
    if args is None:
        args = sys.argv
    try:
        try:
            opts, args = getopt.getopt(args[1:], "", ["report-fail"])
        except getopt.error, msg:
            raise Usage(msg)
        for o, a in opts:
            if o == "--report-fail":
                REPORT_FAILURES = True
        if len(args) == 0:
            raise Usage()
        for case in args:
            try:
                do_test(case)
            except (IOError, OSError), err:
                print >>sys.stderr, '%s: %r' % (err.strerror, err.filename)
        summary()
    except Usage, err:
        if str(err):
            print >>sys.stderr, err
        usage()

if __name__ == "__main__":
    sys.exit(main())
